%{
/* 
 * (was file: fmt.c)
 * 04/88  bgray
 * file: fm.l
 * revised with (f)lex front end, 05/88  Greg Lee
 * revised for hyphenation, 04/92  Greg Lee
 * revised to show text attributes, 07/01  Greg Lee
 * revised for print, trees, unicode 10/01  Greg Lee
 *
 * This code is in the public domain, except possibly
 * for code derived from other sources (see Acknowledgments
 * in the documentation).
 */

#include "fm.h"

#ifdef HAVE_STRFTIME
#include <time.h>

static	time_t rightnow;
char today_str[20];
char day_str[20];
char date_str[20];
char time_str[20];
#endif

#define PUTTC puttc_text

#define YY_INPUT(buf,result,max_size) \
	 result = fread( buf, 1, max_size, yyin );

#define MAX_INCLUDE_DEPTH 40
static YY_BUFFER_STATE include_stack[MAX_INCLUDE_DEPTH];
static int include_stack_ptr = 0;


unsigned char
	attribute = ROMAN,	/* current character attribute */
	ipa_attribute = IPA;
static unsigned char
        nextattribute = 0,	/* attribute to restore after next
				 * word for .RB, etc. or after line
				 * for \beginsection
				 */
	previousfont = ROMAN;

static int sectionlevel = 0;
static unsigned char tbattribute = ROMAN;

static char	nopindentflag = FALSE;	/* don't do par indent */
static char	underscoring = FALSE;
char	frenchspacing = FALSE;
static unsigned char    attrstack[STACKDEPTH];	/* stack of attributes for each current ... */
static unsigned char    lrstack[STACKDEPTH];	/* stack of attributes for each current ... */
static unsigned char    l_hsize[STACKDEPTH];	/* stack of attributes for each current ... */
static int	curr_leading[STACKDEPTH];
static int	curr_leading_pt[STACKDEPTH];
static int	level = 1;		/* TeX {}-group */

/* global state of formatter */
char
        iflag  = FALSE,		/* preserving left indentation ? */
        stiflag  = FALSE,	/* user set -i flag? */
        staflag  = FALSE,	/* user set -a flag? */
        jflag  = TRUE,		/* requested no justification with -j? */
        uflag  = TRUE,		/* no request uncolored output with -u? */
        xflag  = FALSE,		/* requested use TeX commands w. -x? */
        yflag  = FALSE,		/* monospaced? */
        zflag  = FALSE,		/* debug? */
        mflag  = FALSE,		/* requested use man commands w. -m? */
        nflag  = FALSE,		/* requested no color? */
        oflag  = FALSE,		/* requested overstrike? */
        cflag  = FALSE,		/* requested format C code w. -c? */
        fflag  = FALSE,		/* requested no pagination? */
        aflag  = FALSE,		/* don't show typeface changes with -u? */
        kflag  = FALSE,		/* unicode output? */
        alwayskflag  = FALSE,	/* console always unicode output? */
        gflag  = TRUE,		/* unicode input? */
        vflag  = FALSE,		/* verbose */
        qtflag = FALSE,		/* saw odd " in processing .RB etc.? */
        ceflag = FALSE,		/* center next output line? */
        raflag = FALSE,		/* right adjust next output line? */
        laflag = FALSE,		/* left adjust next output line? */
        obflag = FALSE,		/* not concatenating? */
        obsflag = FALSE,	/* ignore user spacing? */
        liflag = FALSE,		/* not concatenating? */
        njflag = FALSE,		/* not currently justifying? */
        vbbflag = FALSE,	/* verbatim? */
        mmode = FALSE,		/* in math mode? */
        rhmode = FALSE,		/* in restricted horizontal mode? */
        rvmode = FALSE,		/* in restricted vertical mode? */
        insertionflag = FALSE,	/* processing an insertion? */
	psflag = FALSE,		/* generating PostScript? */
        lhpenalty = FALSE,	/* last line ended in hyphen? */
	treeflag = FALSE,	/* tree in progress */
	boxflag = FALSE;	/* use BOX font for box chars */

int
        parindent  = 5,
        parskip  = 0,
	baselineskip = 1,
        hsize  = -1,
        hsize_sp  = -1,
        vsize  = -1,
	pointsize = 12,
	cols = 80,
	cols_sp,
        hlevel  = 8,		/* hyphenation level ( -h ) */
        tabsize = 8,
        tabid = -1,
	outdev = 0,
	hoffset = 0,
        lskip  = 0,
        rskip  = 0,
	lmargin = 6000 /* 12 cols */,
        offset_sp  = -1,
        offset  = -1,
        bfont = 0;		/* print with ? */

static int
        indent  = 0,
        pindent  = 0,
        parskip_pt  = 3,
        pskip  = 0,
        pskip_pt  = 0,
	temptablen = 0,
	hboxsize = 0,
	tbsize = 0,
	tbsize_sp = 0,
	prev_extra = 0,
	curr_vskip = 0,
	curr_vskip_pt = 0,
	curr_vglue = 0,
	curr_badness = 0,
	true_badness = 0,
	popme = 0,
	curr_insertion = 1,
	vcenterflag = 0,
	fnflag = 0,
        tpstate = 0,		/* keep track for hanging indentation */
        hangafter = 0,		/* keep track for hanging indentation */
        tpvalue = 5,		/* amount of hang, from arg of .TP,.HP */
        tpmore = 0,		/* extra hang for \itemitem */
        tplevel = 0,		/* nesting level when started \item */
        spaces  = 0,		/* space to leave after current word */
        spaces_pp  = 0,		/* print space to leave after current word */
        tbcnt = 0,		/* tabs */
	tlen = 0,		/* target length of line for ps justification */
	amwidow = 0;		/* widow control */

static char started = 0;
static unsigned char tempa;

static char   *progname = NULL;

#define MAXGLUE 256

static int     gluecnt = 0;
static unsigned char   *glueptr[MAXGLUE];		/* glueptr[0] is unused */
static int     stretchcnt = 0;
static unsigned char   *stretchptr[MAXGLUE];
static int     shrinkcnt = 0;
static unsigned char   *shrinkptr[MAXGLUE];

static unsigned char
	oline[MAXLINE] = "\0",		/* output line buffer */
       *olp = oline,
        aline[MAXLINE] = "\0",		/* output line attribute buffer */
       *alp = aline,
        oword[2*MAXWORD] = "\0",	/* output word buffer */
       *owp = oword,
        aword[2*MAXWORD] = "\0",	/* output word attribute buffer */
       *awp = aword;
unsigned char
	*tabline[MAXTB] = {"\0"};

unsigned char glyph[128] = "\0";
unsigned char let[3][256];

extern char *optarg;		/* from getopt */
extern int  optind;

extern void	exit();

/* these are used by fmtree */
void	puttc(int ch);
void	puttc_text(int ch);
void	putts(unsigned char *s);
void	putword();
void	putline();

static void	emptyline();

static int	max(int a, int b);
static char	*basename(char *s);
char		*mypath = NULL;
char		*myfile = NULL;
char		*myname = NULL;
static void	pathname(char *s);
static void	copyindent (char *ilp);
static int	abbr(unsigned char *s, int plen);
int	endofsentence (unsigned char *p, int plen);
static void	lbreak();
static void	putwordctlspace();
static void	setindent();
static void	justifyline(int len_pp, int extra_len);
static void	skipline(int n);
static void	addparens (char *s, int a);
static void	cleartabs();
static void	setdefaulttabs();
static void	settab(int t);
static int	findtab(int t, int tb);
static int	findtabattribute(int t, int tb);
int		dotab(int c, int tb);
int		dimen(char *s, int horizontal, int cansign, int wantpoints);
static void	eat();
void	push(unsigned char attr);
void	pop();
static void	shrinkline(int *shrinkage);
static void	save_line();
static void	restore_line();
static void	parpop();
static int	num(char *s, int cansign);
static void	puts_tbuf(char *base, int count);
static void	putc_tbuf(int c);
static void	finish_tbuf();
static void	xno(char *s);
static void	putuni(int ch);
static void	stowuni(unsigned int wc);
static void	saveuni(unsigned int wc);
static void	mansym(int k1, int k2);

/* these are for tpar */

static int outercol = 0;
static int t_indent = 0;
static int minindent = 0;
static int t_level = 0;
static int buried[100];
static int t_extra = 0;
static int t_leading = 0;

static void ascend(void);
static void descend(int d_indent);

/* this is for tree */
static int tree_indent = 0;

%}

%option noyywrap

%s	T MAN PLAIN A CCODE C TAB VBB
%x	TREE PTREE

wh	[ \t]*
dim	[ \t=]*("+"|"-")?[0-9]*("."[0-9]+)?[ ]*("in"|"pt"|"em"|"ex"|"pc"|"cm")[ \t]*
num	[ \t=]*("+"|"-")?[0-9]+[ \t]*
na	[^a-z]
aln	[0-9A-Za-z]
alne	[0-9A-Za-z\*@\+\-<>\?\~\-\.\^_/!=%":;'`]
alnm	[0-9A-Za-z\*@\+\-<>\?\~\-\.\^\\_/!=%":;'`]

%%

<T,A>\\tree[ \t\n]*(-([ctuvqdLTOIJFERN]+|[bg][0-9]+)[ \t\n]*)*(\()? {
	int topbadness = -10000;
	if (spaces) {
		topbadness = 0;
	}
	tree_indent = (olp - oline) + (owp - oword) + spaces;
	tree_indent += indent + pindent /* + lrstack[level] + lskip */;
	lbreak();
	if (topbadness < 0) curr_badness -= 10000;
	true_badness = 10000;
	outercol = t_indent = 0;
	t_leading = t_level = t_extra = 0;
	minindent = 200;
	buried[0] = -1;
	puts_tbuf(yytext, yyleng);
	pindent = 0;
	push(attribute);
	if (!obflag) obflag = level;
	if (yytext[yyleng-1] == '(')
		BEGIN(PTREE);
	else BEGIN(TREE);
}
<MAN,A>^".TR"[ \n]*(\()? {
	tree_indent = 0;
	lbreak();
	outercol = t_indent = 0;
	t_leading = t_level = t_extra = 0;
	minindent = 200;
	buried[0] = -1;
	puts_tbuf("\\tree ", 6);
	pindent = 0;
	push(attribute);
	if (!obflag) obflag = level;
	if (yytext[yyleng-1] == '(') {
		putc_tbuf('(');
		BEGIN(PTREE);
	}
	else BEGIN(TREE);
}
<PTREE>{
    ^".TE"\n {
	finish_tbuf();
	pop();
	lbreak();
	if (!mflag) BEGIN(T);
	else if (!xflag) BEGIN(MAN);
	else BEGIN(A);
    }
    .*\\endtree/{na} {
    	puts_tbuf(yytext, yyleng);
	putc_tbuf('\n');
	finish_tbuf();
	pop();
	lbreak();
	eat();
	if (!mflag) BEGIN(T);
	else if (!xflag) BEGIN(MAN);
	else BEGIN(A);
    }
    .*		puts_tbuf(yytext, yyleng);
    \n		putc_tbuf('\n');
}

<TREE>{
    ^".TE"\n {
	descend(minindent);
	putc_tbuf('\n');
	finish_tbuf();
	pop();
	lbreak();
	if (!mflag) BEGIN(T);
	else if (!xflag) BEGIN(MAN);
	else BEGIN(A);
    }
    \\endtree/{na} {
	descend(minindent);
	putc_tbuf('\n');
	finish_tbuf();
	pop();
	lbreak();
	eat();
	if (!mflag) BEGIN(T);
	else if (!xflag) BEGIN(MAN);
	else BEGIN(A);
    }
    \t {
	outercol++;
	while (outercol % tabsize) outercol++;
	if (t_indent) putc_tbuf(' ');
    }
    " " {
	outercol++;
	if (t_indent) putc_tbuf(' ');
    }
    ^"-"/[ \t]+ {
	ascend();
	outercol++;
	t_indent = 0;
    }
    [ \t]"-"/[ \t]+ {
	outercol++;
	if (yytext[0] == '\t')
		while (outercol % tabsize) outercol++;
	if (t_indent) putc_tbuf(' ');
	if (t_leading) {
		t_extra++;
		putc_tbuf('(');
	}
	else {
		ascend();
		outercol++;
		t_indent = 0;
	}
    }
    [ \t]"-"[ \t]*$ {
	outercol++;
	if (yytext[0] == '\t')
		while (outercol % tabsize) outercol++;
	if (t_indent) putc_tbuf(' ');
	if (t_leading) {
		putc_tbuf('(');
		putc_tbuf(')');
	}
	else {
		ascend();
		outercol++;
	}
    }
    "-"[ \t]*$ {
	ascend();
	outercol++;
	if (t_leading) putc_tbuf('-');
    }
    \\[\)\(\-%] {
	ascend();
	outercol++;
	outercol++;
	t_leading++;
	if (yytext[1] != '-') putc_tbuf('\\');
	putc_tbuf(yytext[1]);
    }
    [\)\(] {
	ascend();
	outercol++;
	t_leading++;
	putc_tbuf('\\');
	putc_tbuf(yytext[0]);
    }
    \n	t_leading = t_indent = outercol = 0;
    "%".*	;
    . {
	ascend();
	outercol++;
	t_leading++;
	putc_tbuf(yytext[0]);
    }
}

<T,A>{
    "\\special{"[^\}\{\n]+"}" {
	int i;
	if (psflag) {
		puttc(PS_ESC);
		for (i = 8; i < yyleng - 1; i++) puttc(yytext[i]);
		puttc(' ' - 1);
		rskip -= yyleng - 7;
	}
    }

	/* match various TeX skip commands */
    \\("small"|"med"|"big")("skip"|"break")/{na} {
	int is_break = yytext[yyleng-1];
	int more_skip;

	if (yytext[1] != 's') {
	    if (yytext[1] == 'b' || !psflag) {
		skipline(1);
		if (curr_vskip > 1 && is_break == 'k') curr_vskip--;
	    }
	    else {
		lbreak();
		more_skip = 2 * ps_baselineskip10 / 40;
		if (is_break == 'k') more_skip -= curr_vskip_pt + curr_vskip * ps_baselineskip;
		if (more_skip > 0) curr_vskip_pt += more_skip;
	    }
	}
	else {
		lbreak();
		more_skip = ps_baselineskip10 / 40;
		if (is_break == 'k') more_skip -= curr_vskip_pt + curr_vskip * ps_baselineskip;
		if (psflag && more_skip > 0) curr_vskip_pt += more_skip;
	}
	if (yytext[1] == 'b') curr_badness -= pagestretch;
	else curr_badness -= pagestretch / 2;
	eat();
    }
    \\"nopagenumbers"{num} {
	nopagenumbers = num(yytext + 14, TRUE);
	eat();
    }
    \\"nopagenumbers"/{na} {
	nopagenumbers = TRUE;
	eat();
    }
    \\"pagestretch"{num} {
	pagestretch = num(yytext + 12, FALSE);
	eat();
    }
    \\pageno[ \t]+at[ \t]+bottom/{na} {
	pageattop = FALSE;
	eat();
    }
    \\"pageno"{num} {
	pageno = num(yytext + 7, TRUE);
    }
    \\"penalty"{num} {
	curr_badness += num(yytext + 8, TRUE);
    }
    \\"hangafter"{num} {
	lbreak();
	hangafter = num(yytext + 10, TRUE);
	if (hangafter < 2) tpstate = 4;
    }
    \\"hangindent"{dim}/{na} {
	tpvalue = dimen(yytext + 11, TRUE, FALSE, FALSE);
	eat();
    }
}

	/* BEGIN SHARED SECTION */

<T,A>{

  \\['`\^]({aln}|[ ]*) {
    int c;
    int dia = yytext[1];
    unsigned char some = yytext[2];
    if (some == ' ') some = 0;
    if (is_ipa(attribute) && (psflag||kflag)) {
	if (some) {
		tempa = ipa_attribute;
		ipa_attribute = IPA;
		PUTTC(some);
		ipa_attribute = tempa;
	}
	if (dia == '`') PUTTC(0);
	else if (dia == '^') PUTTC(2);
	else PUTTC(1);
    }
    else if (some) {
      int lat1 = TRUE;
	/* latin 1 accents */
      switch(some) {
	case 'A': c = 0xc1; break;
	case 'E': c = 0xc9; break;
	case 'I': c = 0xcd; break;
	case 'O': c = 0xd3; break;
	case 'U': c = 0xda; break;
	case 'Y': c = 0xdd; break;
	case 'a': c = 0xe1; break;
	case 'e': c = 0xe9; break;
	case 'i': c = 0xed; break;
	case 'o': c = 0xf3; break;
	case 'u': c = 0xfa; break;
	case 'y': if (dia == '\'') c = 0xfd;
		  else {
			c = some;
			lat1 = FALSE;
		  }
		  break;
         default: c = some; 
		  lat1 = FALSE;
		  break;
      }
      if (lat1) {
        if (dia == '`') c--;
        else if (dia == '^') c++;
	PUTTC(c);
      }
      else if (psflag && is_roman(attribute)) {
	PUTTC(c);
	c = roman_other[dia][0];
	PUTTC(c);
      }
      else if (kflag && is_roman(attribute)) {
	PUTTC(c);
	saveuni(roman_other[dia][1]);
      }
      else PUTTC(c);
    }
    else if (is_roman(attribute) && (psflag || kflag)) {
	if (psflag) PUTTC(roman_other[dia][0]);
	else saveuni(roman_other[dia][1]);
    }
    else {
	PUTTC(dia);
	if (yytext[2] == ' ') unput(' ');
    }
  }
  \\\"({aln}|[ ]*) {
    int c;
    unsigned char some = yytext[2];
    if (some == ' ') some = 0;
    if (is_ipa(attribute) && (psflag||kflag)) {
	if (some) {
		tempa = ipa_attribute;
		ipa_attribute = IPA;
		PUTTC(some);
		ipa_attribute = tempa;
	}
	PUTTC(4);
    }
    else if (some) {
	/* latin 1 dieresis */
      switch(some) {
	case 'A': c = 0xc4; break;
	case 'E': c = 0xcb; break;
	case 'I': c = 0xcf; break;
	case 'O': c = 0xd6; break;
	case 'U': c = 0xdc; break;
	case 'a': c = 0xe4; break;
	case 'e': c = 0xeb; break;
	case 'i': c = 0xef; break;
	case 'o': c = 0xf6; break;
	case 'u': c = 0xfc; break;
	case 'y': c = 0xff; break;
         default: c = some;
		  if (is_roman(attribute)) {
			PUTTC(c);
			if (psflag) PUTTC(roman_other['"'][0]);
			else if (kflag) saveuni(roman_other['"'][1]);
			else PUTTC('"');
			c = 0;
		  }
		  break;
      }
      if (c) PUTTC(c);
    }
    else if (is_roman(attribute) && (psflag || kflag)) {
	if (psflag) PUTTC(roman_other['"'][0]);
	else saveuni(roman_other['"'][1]);
    }
    else {
	PUTTC(yytext[1]);
	if (yytext[2] == ' ') unput(' ');
    }
  }
  \\\~({aln}|[ ]*) {
    int c;
    unsigned char some = yytext[2];
    if (some == ' ') some = 0;
    if (is_ipa(attribute) && (psflag||kflag)) {
	if (some) {
		tempa = ipa_attribute;
		ipa_attribute = IPA;
		PUTTC(some);
		ipa_attribute = tempa;
	}
	PUTTC(3);
    }
    else if (some) {
	/* latin 1 tilda */
      switch(some) {
	case 'n': c = 0xf1; break;
	case 'N': c = 0xd1; break;
	case 'a': c = 0xe3; break;
	case 'A': c = 0xc3; break;
         default: c = some;
		  if (is_roman(attribute)) {
			PUTTC(c);
			if (psflag) PUTTC(roman_other['~'][0]);
			else if (kflag) saveuni(roman_other['~'][1]);
			else PUTTC('~');
			c = 0;
		  }
		  break;
      }
      if (c) PUTTC(c);
    }
    else if (is_roman(attribute) && (psflag || kflag)) {
	if (psflag) PUTTC(roman_other['~'][0]);
	else saveuni(roman_other['~'][1]);
    }
    else {
	PUTTC(yytext[1]);
	if (yytext[2] == ' ') unput(' ');
    }
  }
  \\\=({aln}|[ ]*) {
    unsigned char some = yytext[2];
    if (some == ' ') some = 0;
    if (is_ipa(attribute) && (psflag||kflag)) {
	if (some) {
		tempa = ipa_attribute;
		ipa_attribute = IPA;
		PUTTC(some);
		ipa_attribute = tempa;
	}
	PUTTC(9);
    }
    else if (is_roman(attribute) && (psflag||kflag)) {
	if (some) PUTTC(some);
	if (psflag) PUTTC(roman_other['='][0]);
	else saveuni(roman_other['='][1]);
    }
    else {
	PUTTC(yytext[1]);
	if (yytext[2] == ' ') unput(' ');
    }
  }
  \\\.({aln}|[ ]*) {
    unsigned char some = yytext[2];
    if (some == ' ') some = 0;
    if (is_ipa(attribute) && (psflag||kflag)) {
	if (some) {
		tempa = ipa_attribute;
		ipa_attribute = IPA;
		PUTTC(some);
		ipa_attribute = tempa;
	}
	PUTTC(10);
    }
    else if (is_roman(attribute) && (psflag||kflag)) {
	if (some) PUTTC(some);
	if (psflag) PUTTC(roman_other['.'][0]);
	else saveuni(roman_other['.'][1]);
    }
    else {
	PUTTC(yytext[1]);
	if (yytext[2] == ' ') unput(' ');
    }
  }
  "``"  {
	if (psflag) {
		push(attribute);
		if (is_ipa(attribute)) attribute = ROMAN;
		PUTTC(PS_quotedblleft);
		pop();
	}
	else if (kflag) saveuni(0x201c);
	else PUTTC(0xab);
  }
  "''"  {
	if (psflag) {
		push(attribute);
		if (is_ipa(attribute)) attribute = ROMAN;
		PUTTC(PS_quotedblright);
		pop();
	}
	else if (kflag) saveuni(0x201d);
	else PUTTC(0xbb);
  }
  [#]+[^ \t\r\n]  {
	int i = yyleng - 1;
	unsigned char ch = yytext[i];
	i--;
	if (i > 5) i = 5;
	ch = isoprint[ch];
	if (ch < 127 && (psflag || kflag)) {
		if (attribute == MATH) {
			ch = math_other[ch][i];
			PUTTC(ch);
		}
		else if (is_ipa(attribute)) {
			ch = other[ch][i];
			PUTTC(ch);
		}
		else if (is_roman(attribute)) {
			i *= 2;
			if (i > 4) i = 4;
			if (psflag) {
				ch = roman_other[ch][i];
				PUTTC(ch);
			}
			else {
				int u;
				u = roman_other[ch][i + 1];
				saveuni(u);
			}
		}
	}
  }
  \\relax/{na}		eat();
  \\ae/{na}		PUTTC(0xe6); eat();
  \\AE/{na}  {
		if (is_ipa(attribute)) {
			push(ROMAN);
			PUTTC(0xc6);
			pop();
		} else PUTTC(0xc6);
	eat();
  }
  \\ss/{na}  {
		if (is_ipa(attribute)) {
			push(ROMAN);
			PUTTC(0xdf);
			pop();
		} else PUTTC(0xdf);
	 eat();
  }
  \\S/{na}  {
		if (is_ipa(attribute)) {
			push(ROMAN);
			PUTTC(0xa7);
			pop();
		} else PUTTC(0xa7);
	 eat();
  }
  \\P/{na}  {
		if (is_ipa(attribute)) {
			push(ROMAN);
			PUTTC(0xb6);
			pop();
		} else PUTTC(0xb6);
	 eat();
  }
  \\oslash/{na}		PUTTC(0xf8); eat();
  \\Oslash/{na}  {
		if (is_ipa(attribute)) {
			push(ROMAN);
			PUTTC(0xd8);
			pop();
		} else PUTTC(0xd8);
	 eat();
  }
  \\deg/{na}  {
		if (is_ipa(attribute)) {
			push(ROMAN);
			PUTTC(0xb0);
			pop();
		} else PUTTC(0xb0);
	 eat();
  }
  \\neg/{na}  {
		if (is_ipa(attribute)) {
			push(ROMAN);
			PUTTC(0xac);
			pop();
		} else PUTTC(0xac);
	 eat();
  }
  \\forall/{na}  {
		if (psflag) {
			push(MATH);
			PUTTC(042);
			pop();
		}
		else if (kflag) saveuni(0x2200);
		else PUTTC('A');
	 eat();
  }
  \\exists/{na}  {
		if (psflag) {
			push(MATH);
			PUTTC(044);
			pop();
		}
		else if (kflag) saveuni(0x2203);
		else PUTTC('E');
	 eat();
  }
  \\vee/{na}  {
		if (psflag) {
			push(MATH);
			PUTTC(0332);
			pop();
		}
		else if (kflag) saveuni(0x2228);
		else PUTTC('v');
	 eat();
  }
  \\wedge/{na}  {
		if (psflag) {
			push(MATH);
			PUTTC(0331);
			pop();
		}
		else if (kflag) saveuni(0x2227);
		else PUTTC('&');
	 eat();
  }
  \\supset/{na}  {
		if (psflag) {
			push(MATH);
			PUTTC(0311);
			pop();
		}
		else if (kflag) saveuni(0x2283);
		else { PUTTC('-'); PUTTC('>'); }
	 eat();
  }
  \\copyright/{na}  {
		if (is_ipa(attribute)) {
			push(ROMAN);
			PUTTC(0xa9);
			pop();
		} else PUTTC(0xa9);
	 eat();
  }
  \\aa/{na}  {
		if (is_ipa(attribute)) {
			push(ROMAN);
			PUTTC(0xe5);
			pop();
		} else PUTTC(0xe5);
	 eat();
  }
  \\AA/{na}  {
		if (is_ipa(attribute)) {
			push(ROMAN);
			PUTTC(0xc5);
			pop();
		} else PUTTC(0xc5);
	 eat();
  }
  "!`"  {
		if (is_ipa(attribute)) {
			push(ROMAN);
			PUTTC(0xa1);
			pop();
		} else PUTTC(0xa1);
  }
  "?`"  {
		if (is_ipa(attribute)) {
			push(ROMAN);
			PUTTC(0xbf);
			pop();
		} else PUTTC(0xbf);
  }
  "\\\$"	{
	if (attribute == ITALIC)  PUTTC(0xa3);
	else PUTTC('$');
  }
  \\char[0-9]+(" ")?  {
	int i = atoi(yytext + 5);
	if (i < 0x100 && is_ipa(attribute)) {
		PUTTC(i);
	}
	else stowuni(i);
  }
  \\char\'[0-9]+(" ")?  {
	int i = strtol(yytext + 6, NULL, 8);
	if (i < 0x100 && is_ipa(attribute)) {
		PUTTC(i);
	}
	else stowuni(i);
  }
  \\char\"[0-9A-Fa-f]{2,4}(" ")?  {
	int i = strtol(yytext + 6, NULL, 16);
	if (i < 0x100 && is_ipa(attribute)) {
		PUTTC(i);
	}
	else stowuni(i);
  }
  \\char\`(\\)?.(" ")?  {
	if (yytext[6] == '\\') PUTTC(yytext[7]);
	else PUTTC(yytext[6]);
  }
  \\bullet/{na}  {
	if (glyph['~']) {
		attribute += 128;
		PUTTC(glyph['~']);
		attribute -= 128;
	}
	else	stowuni(0xb7);
	eat();
  }
  \\"diamond"/{na}  {
	if (kflag) saveuni(0x25c6);
	else if (psflag) {
		push(MATH);
		PUTTC(0xa8);
		pop();
	}
	else if (glyph['`']) {
		attribute += 128;
		PUTTC(glyph['`']);
		attribute -= 128;
	}
	else	PUTTC('+');
	eat();
  }
  \\"vdash"/{na}  {
	int ix = 't';
	if (attribute == BOLD && boxflag) ix = 'T';
	if (glyph[ix]) {
		if (boxflag) push(BOX); else push(attribute + 128);
		PUTTC(glyph[ix]);
		pop();
	}
	else	PUTTC('|');
	eat();
  }
  \\"dashv"/{na}  {
	int ix = 'u';
	if (attribute == BOLD && boxflag) ix = 'U';
	if (glyph[ix]) {
		if (boxflag) push(BOX); else push(attribute + 128);
		PUTTC(glyph[ix]);
		pop();
	}
	else	PUTTC('|');
	eat();
  }
  \\"bot"/{na}  {
	int ix = 'v';
	if (attribute == BOLD && boxflag) ix = 'V';
	if (glyph[ix]) {
		if (boxflag) push(BOX); else push(attribute + 128);
		PUTTC(glyph[ix]);
		pop();
	}
	else	PUTTC('-');
	eat();
  }
  \\"top"/{na}  {
	int ix = 'w';
	if (attribute == BOLD && boxflag) ix = 'W';
	if (glyph[ix]) {
		if (boxflag) push(BOX); else push(attribute + 128);
		PUTTC(glyph[ix]);
		pop();
	}
	else	PUTTC('-');
	eat();
  }
  \\"lceil"/{na}  {
	int ix = 'l';
	if (attribute == BOLD && boxflag) ix = 'L';
	if (glyph[ix]) {
		if (boxflag) push(BOX); else push(attribute + 128);
		PUTTC(glyph[ix]);
		pop();
	}
	else	PUTTC(' ');
	eat();
  }
  \\"rceil"/{na}  {
	int ix = 'k';
	if (attribute == BOLD && boxflag) ix = 'K';
	if (glyph[ix]) {
		if (boxflag) push(BOX); else push(attribute + 128);
		PUTTC(glyph[ix]);
		pop();
	}
	else	PUTTC(' ');
	eat();
  }
  \\"lfloor"/{na}  {
	int ix = 'm';
	if (attribute == BOLD && boxflag) ix = 'M';
	if (glyph[ix]) {
		if (boxflag) push(BOX); else push(attribute + 128);
		PUTTC(glyph[ix]);
		pop();
	}
	else	PUTTC(' ');
	eat();
  }
  \\"rfloor"/{na}  {
	int ix = 'j';
	if (attribute == BOLD && boxflag) ix = 'J';
	if (glyph[ix]) {
		if (boxflag) push(BOX); else push(attribute + 128);
		PUTTC(glyph[ix]);
		pop();
	}
	else	PUTTC(' ');
	eat();
  }
  \\"|"  {
	int ix = 'n';
	if (attribute == BOLD && boxflag) ix = 'N';
	if (glyph[ix] && attribute != TTYPE) {
		if (boxflag) push(BOX); else push(attribute + 128);
		PUTTC(glyph[ix]);
		pop();
	}
	else	PUTTC('|');
  }
  "|"  {
	int ix = 'x';
	if (attribute == BOLD && boxflag) ix = 'X';
	if (glyph[ix] && attribute != TTYPE &&
		!is_ipa(attribute) && attribute != MUSIX && !liflag) {
		if (boxflag) push(BOX); else push(attribute + 128);
		PUTTC(glyph[ix]);
		pop();
	}
	else	PUTTC('|');
  }

  \\("l")?"dots"/{na}  {
		if (psflag) {
			PUTTC(PS_ellipsis);
		} else {
			PUTTC('.'); PUTTC('.'); PUTTC('.');
		}
	eat();
  }
  \\"cdots"/{na}  {
		if (psflag) {
			PUTTC(PS_ellipsis);
		} else {
		stowuni(0xb7); /* bullets */
		stowuni(0xb7);
		stowuni(0xb7);
		}
	eat();
  }
  \\uparrow/{na}  {
		if (psflag) {
			push(MATH); PUTTC(0xad); pop();
		}
		else if (kflag) stowuni(0x2191);
		else if (glyph['-']) {
			attribute += 128;
			PUTTC(glyph['-']);
			attribute -= 128;
		}
		else {
			PUTTC('^');
		}
	eat();
  }
  \\downarrow/{na}  {
		if (psflag) {
			push(MATH); PUTTC(0xaf); pop();
		}
		else if (kflag) stowuni(0x2193);
		else if (glyph['.']) {
			attribute += 128;
			PUTTC(glyph['.']);
			attribute -= 128;
		}
		else {
			PUTTC('v');
		}
	eat();
  }
  \\leftarrow/{na}  {
		if (psflag) {
			push(MATH); PUTTC(0xac); pop();
		}
		else if (kflag) stowuni(0x2190);
		else if (glyph[',']) {
			attribute += 128;
			PUTTC(glyph[',']);
			attribute -= 128;
		}
		else {
			PUTTC('<');
		}
	eat();
  }
  \\(rightarrow|to)/{na}  {
		if (psflag) {
			push(MATH); PUTTC(0xae); pop();
		}
		else if (kflag) stowuni(0x2192);
		else if (glyph['+']) {
			attribute += 128;
			PUTTC(glyph['+']);
			attribute -= 128;
		}
		else {
			PUTTC('>');
		}
	eat();
  }

  \\[ijLl]/{na}  {
    int dia = yytext[1];
    if ( (psflag || kflag) && is_ipa(attribute)) {
	switch(dia) {
	    case 'i': PUTTC(25); break;
	    case 'j': PUTTC(26); break;
	    case 'l': PUTTC(170); break;
	    default: PUTTC(dia); break;
	}
    }
    else if ( (psflag || kflag) && is_roman(attribute)) {
	if (psflag) PUTTC(roman_other[dia][0]);
	else saveuni(roman_other[dia][1]);
    }
    else PUTTC(dia);
    eat();
  }
  \\[vcuHrkdb][ \t]+{aln}  {
    unsigned char some = yytext[yyleng-1];
    int dia = yytext[1];
    if ( (psflag || kflag) && is_ipa(attribute)) {
		tempa = ipa_attribute;
		ipa_attribute = IPA;
		PUTTC(some);
		ipa_attribute = tempa;
		switch(dia) {
			case 'v': PUTTC(7); break;
			case 'c': PUTTC(11); break;
			case 'u': PUTTC(8); break;
			case 'H': PUTTC(5); break;
			case 'r': PUTTC(6); break;
			case 'k': PUTTC(12); break;
			case 'd': ipa_attribute = IPAUNDER; PUTTC(10); break;
			case 'b': ipa_attribute = IPAUNDER; PUTTC(9); break;
		}
    }
    else if (dia == 'v') {
	    switch (some) {
		case 's': PUTTC(PS_scaron); break;
		case 'S': PUTTC(PS_Scaron); break;
		case 'z': PUTTC(PS_zcaron); break;
		case 'Z': PUTTC(PS_Zcaron); break;
		default: PUTTC(some);
			if ( (psflag || kflag) && is_roman(attribute)) {
				if (psflag) PUTTC(roman_other[dia][0]);
				else saveuni(roman_other[dia][1]);
			}
			break;
	    }
    }
    else if (dia == 'c') {
	    switch (some) {
		case 'c': PUTTC(0xe7); break;
		case 'C': PUTTC(0xc7); break;
		default: PUTTC(some);
			if ( (psflag || kflag) && is_roman(attribute)) {
				if (psflag) PUTTC(roman_other[dia][0]);
				else saveuni(roman_other[dia][1]);
			}
			break;
	    }
    }
    else if ( (psflag || kflag) && is_roman(attribute)) {
	PUTTC(some);
	if (psflag) PUTTC(roman_other[dia][0]);
	else saveuni(roman_other[dia][1]);
    }
    else PUTTC(some);
  }
  \\[vcuHrkdb]/{na}  {
    int dia = yytext[1];
    if ( (psflag || kflag) && is_ipa(attribute)) {
		switch(dia) {
			case 'v': PUTTC(7); break;
			case 'c': PUTTC(11); break;
			case 'u': PUTTC(8); break;
			case 'H': PUTTC(5); break;
			case 'r': PUTTC(6); break;
			case 'k': PUTTC(12); break;
			case 'd': ipa_attribute = IPAUNDER; PUTTC(10); break;
			case 'b': ipa_attribute = IPAUNDER; PUTTC(9); break;
		}
    }
    else if ( (psflag || kflag) && is_roman(attribute)) {
	if (psflag) PUTTC(roman_other[dia][0]);
	else saveuni(roman_other[dia][1]);
    }
    eat();
  }
  \\prime/{na}  {
	if (psflag) {
		push(MATH);
		PUTTC(0xa2);
		pop();
	}
	else if (kflag) saveuni(0x2032);
	else PUTTC('\'');
	eat();
  }

  "--"("-")?  {
	int ix = 'q';
	if ( (attribute == BOLD || attribute == BITALIC) && boxflag) ix = 'Q';
	if (psflag && ix == 'q') {
		if (yyleng > 2) PUTTC(PS_emdash);
		else PUTTC(PS_endash);
	}
	else if (glyph[ix]) {
		if (boxflag) push(BOX); else push(attribute + 128);
		PUTTC(glyph[ix]);
		if (yyleng > 2) PUTTC(glyph[ix]);
		pop();
	}
	else {
		PUTTC('-');
		if (yyleng > 2) PUTTC('-');
	}
  }

  \^\{		if (is_ipa(attribute)) push(IPASUPER); else push(SUPER);
  \^{alne} {
			if (is_ipa(attribute)) push(IPASUPER); else push(SUPER);
			PUTTC(yytext[1]);
			pop();
  }
  _\{		if (is_ipa(attribute)) push(IPASUB); else push(SUB);
  _{alne} {
			if (is_ipa(attribute)) push(IPASUB); else push(SUB);
			PUTTC(yytext[1]);
			pop();
  }
  \^/(\\|#)	ipa_attribute = IPAOVER;
  _/(\\|#)	ipa_attribute = IPAUNDER;

  \\("day"|"time"|"date"|"today")/{na} {
    #ifdef HAVE_STRFTIME
	int i;
	char *rs = NULL;
	int which = yytext[3];
	switch (which) {
		case 't': rs = date_str; break;
		case 'y': rs = day_str; break;
		case 'm': rs = time_str; break;
		case 'd':
		default: rs = today_str; break;
	}
	if (rs[0] == ' ') rs++;
	for (i = 0; i < 20 && rs[i]; i++) PUTTC(rs[i]);
	eat();
    #endif
  }
}


\` {
	if (is_roman(attribute) && kflag && !psflag) saveuni(0x2018);
	else PUTTC('`');
}
\' {
	if (is_roman(attribute) && kflag && !psflag) saveuni(0x2019);
	else PUTTC('\'');
}
	/* END SHARED SECTION */

    /* blank line in input gives blank line in output */
\n{wh}(\r)?$  {
		if (!xflag || liflag || obflag)
			skipline(1);
		else if (xflag) {
			if (!underscoring) nopindentflag = FALSE;
			lbreak();
		}
		else skipline(1);
		if (vbbflag && !yflag) {
			vbbflag = FALSE;
			liflag = FALSE;
			if (mflag && xflag) BEGIN(A);
			else if (xflag) BEGIN(T);
		}
}

    /* traditional */
<PLAIN>^\..* {
			lbreak();
			putts(yytext);
			putword();
			putline();
}

<T,A>\\v("b"|"erbatim"|"bb")/{na} {
			nopindentflag = TRUE;
			lbreak();
			liflag = TRUE;
			pindent = 0;
			true_badness = 500;
			curr_badness -= 500;
			if (yytext[3] == 'b') {
				vbbflag = TRUE;
				BEGIN(VBB);
			}
	eat();
}

<T,A>"%".*\n{wh}$ {
			lbreak();
}
<T,A>"%".*\n{wh} {
			if (xflag && nextattribute && level > sectionlevel) ; else
			if (obflag || liflag || nextattribute) { putword(); putline(); }
			if (nextattribute && level <= sectionlevel) {
				attribute = nextattribute;
				sectionlevel = nextattribute = 0;
			}
}

    /* escaped special characters -- \\ for \ is not like real TeX */
<T,A>\\[#{}&%\^_~"\\]		puttc(yytext[1]);
<TAB>\\[#{}&%\^_~"\\]  {
	if (psflag) {
		temptablen += COL(pwidth(yytext + 1, attribute, 1));
	}
	else temptablen += 1;
}
<T,A,TAB>\\\/		;

    /* some font-changing commands */

<MAN,A>{
    "\\f"(I|2)		previousfont = attribute; attribute = ITALIC;
    "\\fH"     		previousfont = attribute; attribute = SLANTY;
    "\\f"(B|3)    	previousfont = attribute; attribute = BOLD;
    "\\f"(R|1)		previousfont = attribute; attribute = ROMAN;
    "\\f"(G|4)		previousfont = attribute; attribute = IPA;
    "\\fP"		attribute = previousfont;
    "\\fC"     		previousfont = attribute; attribute = TTYPE;
    "\\fS"     		previousfont = attribute; attribute = MATH;
}

<T,A,TAB>{
    "\\it"/{na}		attribute = ITALIC; eat();
    "\\sl"/{na}		attribute = SLANTY; eat();
    "\\bf"/{na}		attribute = BOLD; eat();
    "\\title"/{na}	if (psflag) attribute = TITLE; else attribute = BOLD; eat();
    "\\rm"/{na}		attribute = ROMAN; eat();
    "\\tt"/{na}		attribute = TTYPE; eat();
    "\\bi"/{na}		attribute = BITALIC; eat();
    "\\sa"/{na}		attribute = SANS; eat();
    "\\ipa"/{na}	attribute = IPA; eat();
    "\\p"/{na}		attribute = IPA; eat();
    "\\mu"/{na}  {
    #ifdef USEMUSIX
			attribute = MUSIX;
    #else
			attribute = TTYPE;
    #endif
	eat();
    }
    "\\bx"/{na}		attribute = BOX; eat();
    "\\zd"/{na}		attribute = ZD; eat();

    [\$]+  {
		if (mmode) {
			pop();
			mmode = FALSE;
			if (yyleng > 1) {
				skipline(1);
				nopindentflag = TRUE;
				lbreak();
			}
		}
		else {
			push(MATH);
			mmode = TRUE;
			if (yyleng > 1) {
				lbreak();
				ceflag = TRUE;
				skipline(1);
			}
		}
    }
    "\\TeX"/{na}  {
		puttc('T');
		push(SUB);
		puttc('E');
		pop();
		puttc('X');
	eat();
    }
}

<T,A>{
    \\"centerline"/{na}	lbreak(); ceflag = TRUE; eat();
    \\"rightline"/{na}	lbreak(); raflag = TRUE; eat();
    \\"leftline"/{na}	lbreak(); laflag = TRUE; eat();
    \\"raggedright"/{na}	lbreak(); if (!njflag) njflag = level; eat();
    \\"obeylines"/{na}	lbreak(); if (!obflag) obflag = level; eat();
    \\"obeyspaces"/{na}	lbreak(); if (!obsflag) obsflag = level; eat();

    \\"enskip"{wh}  {
		if (tbcnt) puttc(' ');
		else {
			putword();
	/* I guess the normal word space is already there
			if (spaces) spaces++;
	 */
		}
    }
    \\("no")?"frenchspacing"/{na} {
	frenchspacing = (yytext[1] == 'f');
	eat();
    }
    \\"noindent"/{na}	pindent = 0; eat();
    \\"narrower"/{na}  {
	lrstack[level] += 5;
	eat();
    }
    \\"parindent"{dim}/{na}  {
		lbreak();
		parindent = dimen(yytext+10, TRUE, FALSE, FALSE);
		tpvalue = parindent;
		lbreak();
	eat();
    }
    \\"baselineskip"{dim}/{na}  {
	if (!started && level == 1) {
		ps_baselineskip = dimen(yytext+13, FALSE, FALSE, TRUE);
		ps_baselineskip10 = ps_baselineskip * 10;
		calculate_ps_page_length();
		baselineskip = dimen(yytext+13, FALSE, FALSE, FALSE);
	}
	else {
	    lbreak();
	    if (!psflag) {
		curr_leading[level] = dimen(yytext+13, FALSE, FALSE, FALSE);
		if (curr_leading[level] > 0) curr_leading[level]--;
	    }
	    else {
		curr_leading_pt[level] = dimen(yytext+13, FALSE, FALSE, TRUE) - ps_baselineskip;
		if (curr_leading_pt[level] < 0) curr_leading_pt[level] = 0;
	    }
	}
	eat();
    }
    \\"parskip"{dim}/{na}  {
		lbreak();
		parskip = dimen(yytext+8, FALSE, FALSE, FALSE);
		parskip_pt = dimen(yytext+8, FALSE, FALSE, TRUE);
		lbreak();
	eat();
    }
    \\"report"/{na}  {
	fprintf(stderr,"\nlevel %d:", level);
	if (psflag) {
	  fprintf(stderr,"baselineskip %d points\n", ps_baselineskip + curr_leading_pt[level]);
	}
	else {
	  fprintf(stderr,"baselineskip %d\n", baselineskip + curr_leading[level]);
	}
	eat();
    }
    \\"vskip"{dim}/{na}  {
		int i;
		i = dimen(yytext+6, FALSE, FALSE, FALSE);
		if (i) skipline(i);
		else {
		    lbreak();
		    if (psflag) {
			i = dimen(yytext+6, FALSE, FALSE, TRUE);
			curr_vskip_pt += i;
		    }
		}
	eat();
    }
    \\"vglue"{dim}/{na}  {
		lbreak();
		curr_vglue = dimen(yytext+6, FALSE, FALSE, FALSE);
	eat();
    }
    \\"tenrm"/{na}  {
		pointsize = 10;
	ps_pointsize10 = 100;
	ps_pointsize = 10.0;
	ps_pitch_sp = 720000 / ps_pointsize10;
	ps_pitch = 144.0 / ps_pointsize;
	calculate_offset();
	eat();
    }
    \\"hoffset"{dim}/{na}  {
		hoffset = dimen_sp(yytext+8, TRUE, TRUE);
		calculate_offset();
	eat();
    }
    \\"voffset"{dim}/{na}  {
		ps_voffset = dimen(yytext+8, FALSE, TRUE, TRUE);
	eat();
    }
    \\"hsize"{dim}/{na}  {
		int hsz;
		hsz = dimen_sp(yytext+6, TRUE, FALSE);
		l_hsize[level] = COL(hsz);
		if (level == 1) {
			hsize = COL(hsz);
			hsize_sp = hsz;
			calculate_offset();
		}
	eat();
    }
    \\"vsize"{dim}/{na}  {
		vsize = dimen(yytext+6, FALSE, FALSE, FALSE);
		devinit();
	eat();
    }
    \\("hskip"|"kern"){dim}/{na}  {
	int i = 6, w, v;
	if (yytext[1] == 'k') i--;
	w = dimen(yytext+i, TRUE, TRUE, FALSE);
	if (psflag && (i < 6 || w < 0)) {
		v = dimen(yytext+i, TRUE, TRUE, TRUE);
		if (v < -5000) v = -5000;
		if (v > 5000) v = 5000;
		v += 10000;
		puttc(KERN_ESC);
		puttc(v/100 + 1);
		puttc(v % 100 + 1);
		rskip += w;
	}
	else {
		if (w < 0) {
			while (owp > oword && w < 0) {
				w++;
				owp--;
				awp--;
			}
			while (olp > oline && w < 0) {
				w++;
				olp--;
				alp--;
			}
			lskip += w;
		}
		else {
			if (tbcnt || yytext[1]=='k')
				 for (i = 0; i < w; i++) puttc(' ');
			else {
		  		putword();
		  		spaces += w - 1;
			}
		}
	}
	eat();
    }
    \\"indent"/{na}  {
		int i;
		if (tbcnt) for (i = 0; i < parindent; i++)
			puttc(' ');
		else {
			putword();
			if (spaces && parindent) spaces += parindent - 1;
		}
	eat();
    }
    \\"quad"/{na}  {
		if (tbcnt) putts("  ");
		else {
			putword();
			if (spaces) spaces += 1;
		}
	eat();
    }
    \\"qquad"/{na}  {
		if (tbcnt) putts("    ");
		else {
			putword();
			if (spaces) spaces += 3;
		}
	eat();
    }
    \\"footnote"([^#\{\\]|"#"+[^ #\t\r\n]|"{"[^\{\}\\\n]*"}"|"\\fn"([aAbcdefghiIlnops]|("-"?[0-9]+)))/"{"  {
		int i, t = 0;
		unsigned char tag = yytext[9], intag = 0;
		tempa = attribute; attribute = SUPER;
		if (tag == '{') {
		    for (i = 10; yytext[i] != '}'; i++)
			PUTTC(yytext[i]);
		}
		else if (tag == '\\') {
			xno(yytext + 9);
		}
		else if (tag == '#') {
			for (t = 0; yytext[t + 10] == '#'; t++) ;
			intag = isoprint[ (unsigned char)yytext[t + 10] ];
			t *= 2;
			if (t > 4) t = 4;
			if (psflag) PUTTC(roman_other[intag][t]);
			else saveuni(roman_other[intag][t + 1]);
		}
		else {
			PUTTC(tag);
		}
		attribute = tempa;
		putword();
		save_line();
		if (insertionflag) {
			/* this is an error */
			pop();
			curr_insertion++;
		}
		insertionflag = FOOTNOTE;
		fnflag = level;
		pindent = 0;
		if (glyph['q'])
		    for (i = 0; i < hsize/3; i++) {
			*owp++ = glyph['q'];
			if (boxflag) *awp++ = BOX;
			else *awp++ = ROMAN + 128;
		    }
		else for (i = 0; i < hsize/3; i++) {
			puttc('-');
		}
		lbreak();
		pindent = 0;
		attribute = SUPER;
		if (tag == '{') {
		    for (i = 10; yytext[i] != '}'; i++)
			puttc(yytext[i]);
		}
		else if (tag == '\\') {
			yytext[9+3] = 'l';
			xno(yytext + 9);
		}
		else if (tag == '#') {
			if (psflag) PUTTC(roman_other[intag][t]);
			else saveuni(roman_other[intag][t + 1]);
		}
		else {
			PUTTC(tag);
		}
		attribute = tempa;
		putword();
		spaces = 0;
    }
    \\"vbox"{wh}/"{"  {
		lbreak();
		if (insertionflag) {
			/* this is an error */
			pop();
			curr_insertion++;
		}
		insertionflag = VBOX;
		vcenterflag = level;
		rvmode = TRUE;
    }
    \\"vcenter"{wh}/"{"  {
		lbreak();
		if (insertionflag) {
			/* this is an error */
			pop();
			curr_insertion++;
		}
		insertionflag = VCENTER;
		vcenterflag = level;
		rvmode = TRUE;
    }
    \\("mid"|"top"|"page")"insert"/{na}  {
		skipline(1);
		if (insertionflag) {
			/* this is an error */
			pop();
			curr_insertion++;
		}
		if (yytext[1] == 'm') insertionflag = MIDINSERTION;
		else if (yytext[1] == 't') insertionflag = TOPINSERTION;
		else if (yytext[1] == 'p') insertionflag = PAGEINSERTION;
		push(attribute);
	eat();
    }
    \\"endinsert"/{na}  {
		lbreak();
		pop();
		insertionflag = FALSE;
		curr_insertion++;
	eat();
    }
    \\"eject"/{na}  {
		lbreak();
		if (!fflag) curr_vskip += vsize;
		emptyline();
	eat();
    }
    \\"bye"/{na}  {
		lbreak();
    		if (!fflag) curr_vskip += 10000;
		/*emptyline();*/
		return 0;
    }
}

<MAN,A>{
    ^".sp"				skipline(1);
    ^"."("sp"|"ne")" "[0-9]+		skipline(atoi(yytext+4));
    ^".nf"				lbreak(); obflag = TRUE; obsflag = TRUE;
    ^".fi"				lbreak(); obflag = FALSE; obsflag = FALSE;
    ^".na"				njflag = TRUE;
    ^".ad"(" b"|" n")?			njflag = FALSE;
    ^".hy "[0-9]			hlevel = atoi(yytext+3);
    ^".nh"				hlevel = 0;
    ^".ig".*\n([^\.].*\n)*".."		;
}

<T,A>{
    \\[cefgs]n([aAbcdefghiIlnops]|("-"?[0-9]+)) {
	xno(yytext);
    }
    \\"hbox to"{dim}{wh}"{"  {
	/* if something now in word buffer, kill spaces after putword() call */
	int oldlen = owp - oword;
	if (!rhmode) {
		putword();
		if (oldlen > 0) spaces_pp = spaces = 0;
		push(ROMAN);
		rhmode = level;
		hboxsize = dimen(yytext+8, TRUE, FALSE, FALSE);
	}
    }
    \\"hbox{"  {
		push(ROMAN);
		if (!rhmode) rhmode = level;
		hboxsize = 0;
    }
    "{"  {
		push(attribute);
    }

    "}"  {
		pop();
    }
}

^[ \t]+ {
			putword();
			if (iflag || obsflag || liflag) {
				putline();
				if (iflag && !liflag) curr_badness -= pagestretch / 2;
				tempa = attribute;
				/* for comments with indentation */
				if (cflag) attribute = ROMAN;
				copyindent(yytext);
				attribute = tempa;
			}
			else if (mflag)
				putline();
}

<T,A>"\\"/\n |
<T,A>"\\"[ \t]	putwordctlspace();

<MAN,A,CCODE,C>\n  {	/* for C code reverse all of multiline comments */
			if (!cflag) attribute = ROMAN;
			nextattribute = qtflag = FALSE;
			/* signal end of tag for hanging indent */
			if (tpstate == 1) tpstate = 2;
			putword();
			if (obflag || liflag) putline();
}

    /* more font-changing */

<MAN,A>{
    ^\.("B"|"SB")[ \n]		previousfont = attribute; attribute = BOLD;
    ^\.(I|C|"Nm")[ \n] 		previousfont = attribute; attribute = ITALIC;
    ^\.("R"|"SM")[ \n]		previousfont = attribute; attribute = ROMAN;
    ^".S"[ \n]			previousfont = attribute; attribute = MATH;
    ^\.H[ \n]			previousfont = attribute; attribute = SLANTY;
    ^\.G[ \n]			previousfont = attribute; attribute = IPA;
    ^".ft B"\n			previousfont = attribute; attribute = BOLD;
    ^".ft I"\n			previousfont = attribute; attribute = ITALIC;
    ^".ft"(" R")?\n		previousfont = attribute; attribute = ROMAN;
    [#]+[^ \t\r\n]  {
	int i = yyleng - 1;
	int ch = yytext[i];
	i--;
	if (i > 5) i = 5;
	ch = isoprint[ch];
	if (ch < 127) {
		if (psflag || kflag) {
			if (attribute == MATH) ch = math_other[ch][i];
			else ch = other[ch][i];
		}
		PUTTC(ch);
	}
    }

    \\[\|\^&!{}\nacdeprtuz]		;

    ^".if n "				;
    ^".if t ".*\n				;
    ^".de ".*\n".."			;
    ^".de ".*\n.*\n".."			;
    ^".de ".*\n.*\n.*\n".."		;
    ^".de ".*\n.*\n.*\n.*\n".."		;

    \\.				puttc(yytext[1]);

    ^".ti +"[0-9]+  {
	int i;
	lbreak();
	for (i = 0; i < atoi(yytext+5) - 1; i++) puttc(' ');
    }

    ^\.br 				lbreak();
    ^\.\n 				lbreak();
}

<T,A>\\"par"/{na}	{
	if (!rhmode) {
		nopindentflag = FALSE;
		if (xflag && nextattribute && level > sectionlevel) ;
		else lbreak();
		if (nextattribute && level <= sectionlevel) {
			attribute = nextattribute;
			sectionlevel = 	nextattribute = 0;
		}
	}
	eat();
}
<MAN,A>^\.(P|L|"")P 			skipline(1); indent = 5;

    /* hanging indent stuff (awfully complicated) */

<MAN,A>^\.(T|H)"P "[0-9]+\n  {
			skipline(1);
			tplevel = level;
			if (yytext[1] == 'H')
				tpstate = 4;
			else	tpstate = 1;
			tpvalue = atoi(yytext+4);
			indent = 5;
}

<T,A>\\"item"("item")?/{na}  |
<MAN,A>^".HP ".*\n	|
<MAN,A>^".TP".*\n	|
<MAN,A>^".IP"[ \t]*([0-9]*\n)?  {
			if (xflag) lbreak(); else skipline(1);
			tplevel = level;
			if (yytext[1] == 'I' && yytext[3] == '\n')
				tpstate = 4;
			else	tpstate = 1;
			if (!xflag) {
				indent = 5;
				tpvalue = 5;
			}
			if (yytext[0] == '\\') {
				if (yytext[5] == 'i') {
					tpmore = tpvalue;
				}
				eat();
			}
}

    /* various begin-end commands to set off blocks of text */

<MAN,A>^"."(R|D)(S|E)  {
			if (yytext[1] == 'D' && yytext[2] == 'S') {
				skipline(1);
				obflag = TRUE;
				obsflag = TRUE;
			}
			else	lbreak();
			indent = 10;
			if (yytext[2] == 'E') {
				indent = 5;
				obflag = FALSE;
				obsflag = FALSE;
			}
}

<MAN,A>^".E"(X|E)  {
			skipline(1);
			if (yytext[2] == 'E') {
				indent = 5;
				obflag = FALSE;
				obsflag = FALSE;
			}
			else {	indent = 10;
				obflag = TRUE;
				obsflag = TRUE;
			}
}


<MAN,A>^".N"(T|E).*  {
			skipline(1);
			if (yytext[2] == 'E') {
				indent = 5;
				/* hsize += 5; */
			}
			else {	indent = 10;
				/* hsize -= 5; */
				tempa = attribute;
				attribute = ITALIC;
				ceflag = TRUE;
				if (yyleng < 5) putts("NOTE");
				else putts(yytext+4);
				attribute = tempa;
				skipline(1);
			}
}

    /* those curious -man commands to join words, alternating fonts */

<MAN,A>{
    ^\.("IR"|"PN")" "	nextattribute = ROMAN; previousfont = attribute; attribute = ITALIC;
    ^\."BR "		nextattribute = ROMAN; previousfont = attribute; attribute = BOLD;
    ^\."RI "		nextattribute = ITALIC; previousfont = attribute; attribute = ROMAN;
    ^\."BI "		nextattribute = ITALIC; previousfont = attribute; attribute = BOLD;
    ^\."RB "		nextattribute = BOLD; previousfont = attribute; attribute = ROMAN;
    ^\."IB "		nextattribute = BOLD; previousfont = attribute; attribute = ITALIC;
}

<T,A>\\"proclaim"[^\.\n]+"."  {
		int i;
		skipline(1);
		curr_badness -= pagestretch / 2;
		for (i = 9; yytext[i] == ' '; i++) ;
		for (; yytext[i] != '.'; i++) {
			*owp++ = yytext[i];
			*awp++ = BOLD;
		}
		*owp++ = '.';
		*awp++ = BOLD;
		*owp++ = ' ';
		*awp++ = ROMAN;
		pindent = 0;
		popme = TRUE;
		push(SLANTY);
}
    /* section headings */

<T,A>\\("begin"|"sub"|"subsub")?"section"/{na}  {
			if (yytext[5] == 'u') skipline(1);
			else skipline(2);
			curr_badness -= pagestretch;
			laflag = TRUE;
			sectionlevel = level;
			nextattribute = attribute;
			if (yytext[2] == 'e') {
				curr_vskip += 10000;
				attribute = BOLD;
			}
			else if (yytext[5] == 'y') attribute = SLANTY;
			else attribute = BITALIC;
			underscoring = TRUE;
			nopindentflag = TRUE;
	eat();
}

<MAN,A>^\."S"(H|S)(" "|\n).* {
			skipline(1);
			curr_badness -= pagestretch;
			obflag = FALSE;
			obsflag = FALSE;
			if (yytext[2] != 'S') indent = 0;
			attribute = BOLD;
			underscoring = TRUE;
			putts(yytext+4);
			lbreak();
			indent = 5;
			attribute = ROMAN;
}

<MAN,A>^\."TH ".* {
			lbreak();
			indent = 0;
			addparens(yytext+4, BOLD);
			raflag = TRUE;
			skipline(2);
			indent = 5;
}

<MAN,A>^\."Dd ".* {
			lbreak();
			indent = 0;
			putts(yytext+4);
			raflag = TRUE;
}

    /* "Manual Section" reference */
<MAN,A>^".MS ".*	addparens(yytext+4, ITALIC);

<MAN,A>^".CT ".  {
			if (uflag || oflag) {
				tempa = attribute;
				attribute = TTYPE;
				puttc(toupper(yytext[4]));
				attribute = tempa;
			}
			else {
				putts("<CTRL/");
				puttc(yytext[4]);
				puttc('>');
			}
}

<MAN,A>^".V"(S|E).* |
<MAN,A>^".UC"(..)?  |
<MAN,A>^".IX".*     |
<MAN,A>^".fn "      |
<MAN,A>^\.([ ]+)?\\\".*	;

    /* for join-words commands, words can be quoted */
<MAN,A>\"  {
		if (nextattribute || tpstate == 1)  {
			qtflag = !qtflag;
		}
		else puttc('"');
}

<MAN,A,CCODE,C>" "  {
		if (nextattribute)  {
			if (qtflag) puttc(' ');
			else  {
				tempa = attribute;
				attribute = nextattribute;
				nextattribute = tempa;
			}
		}
		else {
			putword();
			if (obsflag || liflag) copyindent(yytext);
		}
}

<T,A,CCODE,C>\n  {
		putword();
			if (xflag && nextattribute && level > sectionlevel) ; else
		if (obflag || liflag || nextattribute) putline();
		if (nextattribute && level <= sectionlevel) {
			attribute = nextattribute;
			sectionlevel = nextattribute = 0;
		}
}

<T,A>"~"		puttc(' ');

<T,A>"&"[ \t\n]*  {
		if (tbcnt) {
			tbcnt++;
			/* attribute = tbattribute;*/
			putword();
			spaces_pp = spaces = 0;
			tbsize_sp = tbsize = 0;
			puttc('\t');
			attribute = findtabattribute(tbcnt, tabid);
		}
		else puttc('&'); /* ?? */
}

[ \t]+  {
		putword();
		if (obsflag || liflag) copyindent(yytext);
}

[\n\r\f]		putword(); if (vbbflag || obflag) putline();

<T,A>\\"backslash"/{na}	puttc('\\'); eat();

<T,A>\\"settabs"{wh}[0-9]+{wh}\\"columns"/{na}  {
		int col, w, i, hs, r;
		cleartabs();
		hs = l_hsize[level] - 2*lrstack[level];
		col = atoi(yytext + 8);
		if (col > 1 && col < hs / 2 ) {
			w = hs / col;
			r = hs - col * w;
			for (i = 0; i < col - 1; i++) {
				if (r) { w++; r--; }
				/* tabline[tabid][w+1] = '&'; */
				tabline[tabid][w+1] = attribute;
				w += hs / col;
			}
			/* tabline[tabid][hs] = '&'; */
			tabline[tabid][hs] = attribute;
		}
	eat();
}
<T,A>\\"settabs"{wh}\\"+"("%".*)?  {
		cleartabs();
		BEGIN(TAB);
}
<TAB>\\"cr"/{na}  {
		if (mflag) BEGIN(A);
		else BEGIN(T);
		temptablen = 0;
	eat();
}
<T,A>^\\"+"  {
		nopindentflag = TRUE;
		lbreak();
		tbattribute = attribute;
		tbcnt = 1;
		attribute = findtabattribute(tbcnt, tabid);
		tbsize_sp = tbsize = 0;
		temptablen = 0;
}
<T,A>\\"cr"/{na}  {
		nopindentflag = TRUE;
		if (tbcnt) {
			attribute = tbattribute;
			if (!curr_vskip) curr_badness += 4 * pagestretch;
		}
		tbcnt = 0;
		tbsize_sp = tbsize = 0;
		lbreak();
	eat();
}

<TAB>{
    "%".*\n		;
    [^\&\\%\n]*"&"[ \t\n]*  {
		int i;
		for (i = 0; yytext[i] != '&'; i++) ;
		if (psflag) {
		#ifdef TABDEBUG
			fprintf(stderr,"tab to: len %d, pwidth %d, cols=%d\n",
			i, pwidth(yytext, attribute, i),
			COL(pwidth(yytext, attribute, i)) );
		#endif
			i = COL(pwidth(yytext, attribute, i));
		}
		else if (gflag) {
			i = pwidth(yytext, attribute, i);
		}
		settab(i + temptablen + 1);
		#ifdef TABDEBUG
			fprintf(stderr,"tabset col %d (i=%d temptablen=%d)\n",
			i + temptablen + 1, i, temptablen);
		#endif
		temptablen = 0;
    }
    \\enskip/{na}   temptablen += 1; eat();
    \\quad/{na}     temptablen += 2; eat();
    \\qquad/{na}    temptablen += 4; eat();
    \\indent/{na}   temptablen += parindent; eat();
    \\hskip{dim}/{na}  {
		int w;
		w = dimen(yytext+6, TRUE, FALSE, FALSE);
		temptablen += w;
	eat();
    }
    [^\&\\\n]+      {
	#ifdef TABDEBUG
		fprintf(stderr," temptablen = %d", temptablen);
	#endif
	if (psflag) {
		temptablen += COL(pwidth(yytext, attribute, yyleng));
	}
	else if (gflag) {
		temptablen += pwidth(yytext, attribute, yyleng);
	}
	else temptablen += yyleng;
	#ifdef TABDEBUG
		fprintf(stderr," + %d = %d\n", yyleng, temptablen);
	#endif
    }
}

<T,A>{
    \\let[ \t]+[^ \t\r\n]=(\"[0-9a-fA-F]+|\'[0-7]+|[0-9]{2,}|[^ \t\r\n]) {
	int i, newc;
	unsigned char oldc;
	for (i = 4; yytext[i] == ' ' || yytext[i] == '\t'; i++) ;
	oldc = yytext[i];
	i += 2;
	if (!yytext[i+1]) newc = (unsigned char)yytext[i];
	else {
		if (yytext[i] == '"') newc = strtol(yytext+i+1, 0, 16);
		else if (yytext[i] == '\'') newc = strtol(yytext+i+1, 0, 8);
		else newc = strtol(yytext+i, 0, 10);
	}
	if (newc >= 0 && newc < 256) {
		if (is_ipa(attribute)) let[2][oldc] = newc;
		else if (attribute == MATH) let[1][oldc] = newc;
		else if (is_roman(attribute)) let[0][oldc] = newc;
	}
    }
    \\vfill/{na}		lbreak(); if (!fflag) curr_vskip += 10000; eat();

    \\(h|dot)fill{wh}[^\&\\\}\{\n]*(\\(h|dot)fill{wh})?  {
		int w = 0, wl, i = 6, j, ti, c, room, confined = FALSE;
		int leaderlen, sp_len, postlen = 0, dots = FALSE, moredots = FALSE;
		 /* if there's stuff in the word buffer, there was no word break before
		  * "\hfill", so the rule should abut what is to the left
		  */
		int oldlen = owp - oword;
		putword();
		if (rhmode || tbcnt) confined = TRUE;
		if (oldlen != 0) {
			if (rhmode) hboxsize += spaces;
			if (tbcnt) tbsize -= spaces;
			if (tbcnt) tbsize_sp -= SP(spaces);
			spaces_pp = spaces = 0;
		}
		if (yytext[1] == 'd') {
			i = 8;
			dots = TRUE;
		}

		while (yytext[i] &&
		       (yytext[i] == ' '|| yytext[i] == '\t') ) i++;
		ti = i;
		room = l_hsize[level] - pindent - indent -
				2*lrstack[level] - (olp - oline);
		while (yytext[i] && yytext[i] != '\\') {
			if (w >= room) break;
			i++;
			w++;
		}

		wl = w;
		if (w && boxflag)
			w = pwidth(yytext + ti, attribute, w);

		if (rhmode && hboxsize) {
			leaderlen = hboxsize;
			hboxsize -= leaderlen;
			if (hboxsize < 0) hboxsize = 0;
		}
		else if (tbcnt)
			leaderlen = findtab(tbcnt, tabid) -
			   findtab(tbcnt-1, tabid);
		else
			leaderlen = room;

		if (psflag) {
			leaderlen = SP(leaderlen);
			if (tbcnt) leaderlen -= tbsize_sp;
		}
		else {
			if (tbcnt) leaderlen -= tbsize;
		}

		if (yytext[i] == '\\') {
			postlen = (leaderlen - w) / 2;
			leaderlen -= postlen;
			if (yytext[i+1] == 'd') moredots = TRUE;
		}
		if (psflag) {
			sp_len = leaderlen - w;
			*owp++ = HRULE_ESC;
			*awp++ = ROMAN;
			if (dots) *owp++ = '.';
			else *owp++ = ' ';
			*awp++ = ROMAN;
			*owp++ = sp_len >> 8;
			*awp++ = ROMAN;
			*owp++ = sp_len & 0xff;
			*awp++ = ROMAN;
		}
		else for (j = 0; j < leaderlen - w; j++) {
			if (dots) *owp++ = '.';
			else *owp++ = ' ';
			*awp++ = ROMAN;
		}
		if (!confined) {
			putword();
			spaces_pp = spaces = 0;
		}
		while (wl > 0 && (c=yytext[ti]) ) {
		    if (c == '|' && glyph['x'] && attribute != TTYPE && !liflag) {
			    *owp++ = glyph['x'];
			    if (boxflag) *awp++ = BOX;
			    else *awp++ = ROMAN;
		    }
		    else {
			    if (c == '~') c = ' ';
			    *owp++ = c;
			    *awp++ = attribute;
		    }
		    wl--;
		    ti++;
		}
	    if (!tbcnt || moredots) {
		if (psflag) {
			sp_len = postlen;
			*owp++ = HRULE_ESC;
			*awp++ = ROMAN;
			if (moredots) *owp++ = '.';
			else *owp++ = ' ';
			*awp++ = ROMAN;
			*owp++ = sp_len >> 8;
			*awp++ = ROMAN;
			*owp++ = sp_len & 0xff;
			*awp++ = ROMAN;
			
		}
		else for (j = 0; j < postlen; j++) {
			if (moredots) *owp++ = '.';
			else *owp++ = ' ';
			*awp++ = ROMAN;
		}
	    }
    }

    \\"hrule"("fill")?{wh}[^\&\\\}\{\n]*  {
		int w = 0, i = 6, ti, c, sp_len;
		int leaderlen, isfill = FALSE;
		int ix = 'q';
 			/* if there's stuff in the word buffer, there was no word break before
  			 * "\hrule", so the rule should abut what is to the left
  			 */
		int oldlen = owp - oword;

		if (attribute == BOLD && boxflag) ix = 'Q';
		if (yytext[6] == 'f') {
			i = 10;
			isfill = TRUE;
		}
		if (tbcnt || rhmode) isfill = TRUE;

		if (isfill) putword();
		else lbreak();

		if (oldlen != 0) {
			#ifdef HRULEDEBUG
			if (tbcnt) fprintf(stderr,"tbcnt=%d, oldlen=%d, tbsize=%d - spaces =%d\n",
			tbcnt, oldlen, tbsize, tbsize - spaces);
			#endif
			if (rhmode) hboxsize += spaces;
			if (tbcnt) tbsize -= spaces;
			if (tbcnt) tbsize_sp -= SP(spaces);
			spaces_pp = spaces = 0;
		}
		while (yytext[i] &&
		       (yytext[i] == ' '|| yytext[i] == '\t') ) i++;
		ti = i;
		while (yytext[i]) { i++; w++; }

		if (w && boxflag) w = pwidth(yytext + ti, attribute, w);

		if (rhmode && hboxsize) {
			#ifdef HRULEDEBUG
			fprintf(stderr,"hboxsize=%d, w=%d, leaderlen=%d\n",
			hboxsize, w, hboxsize - w);
			#endif
			leaderlen = hboxsize - w;
			hboxsize -= leaderlen;
			if (hboxsize < 0) hboxsize = 0;
		}
		else if (tbcnt) {
			leaderlen = findtab(tbcnt, tabid) -
			   findtab(tbcnt-1, tabid) - w - tbsize;
			/*if (tbcnt > 1) leaderlen++;*/
		}
		else {
			if (!isfill) {
				pindent = 0;
				indent = 0;
			}
			leaderlen = l_hsize[level] - pindent - indent -
				 2*lrstack[level] - (olp - oline) - w;
		}
		/* if (tbcnt) leaderlen -= olp - tempolp - 1; */
		if (psflag) {
			sp_len = SP(leaderlen);
			*owp++ = HRULE_ESC;
			*awp++ = ROMAN;
			*owp++ = '-';
			*awp++ = ROMAN;
			*owp++ = sp_len >> 8;
			*awp++ = ROMAN;
			*owp++ = sp_len & 0xff;
			*awp++ = ROMAN;
		}
		else if (glyph[ix])
		    for (i = 0; i < leaderlen; i++) {
			*owp++ = glyph[ix];
			*awp++ = boxflag? BOX : ROMAN + 128;
		    }
		else
		    for (i = 0; i < leaderlen; i++) {
			*owp++ = '-';
			*awp++ = ROMAN;
		    }

		while (w > 0 && yytext[ti]) {
		    c = (unsigned char)yytext[ti];
		    if (c == '|' && glyph['x'] && attribute != TTYPE && !liflag) {
			    *owp++ = glyph['x'];
			    if (boxflag) *awp++ = BOX;
			    else *awp++ = ROMAN + 128;
		    }
		    else {
			    if (c == '~') *owp++ = ' ';
			    else *owp++ = c;
			    *awp++ = attribute;
		    }
		    w--;
		    ti++;
		}
		if (!isfill) lbreak();
    }

    \\"input"{wh}[^ \t\n\{\}\\%]+ {
	char *fname;
	int i = 6;
	FILE *save_yyin;

	while (isspace(yytext[i])) i++;

        if ( include_stack_ptr >= MAX_INCLUDE_DEPTH ) {
		windup();
		fprintf(stderr, "Probable input loop\n");
		exit(1);
        }
        include_stack[include_stack_ptr++] =
                YY_CURRENT_BUFFER;

	if (mypath) {
		fname =
		 strcpy( (char *)malloc(strlen(yytext+i)+5+strlen(mypath)), mypath);
		strcat(fname, yytext + i);
	}
	else fname = strcpy( (char *)malloc(strlen(yytext+i)+5), yytext+i );

	save_yyin = yyin;
	yyin = NULL;

	yyin = fopen(fname, "r");

	if (yyin == NULL) {
		if (strncmp(fname+strlen(fname)-4,".tex",4)) {
			strcat(fname, ".tex");
			yyin = fopen(fname, "r");
		}
	}
	if (yyin == NULL) {
		windup();
		fprintf(stderr, "Source file %s not found.\n", fname);
		include_stack_ptr--;
		exit(1);
	}

	if (yyin == NULL) yyin = save_yyin;
	else {
	    if (vflag) fprintf(stderr," (%s ", fname);
            yy_switch_to_buffer(
                yy_create_buffer( yyin, YY_BUF_SIZE ) );
	}
    }  

    <<EOF>> {
        if ( --include_stack_ptr < 0 ) {
                yyterminate();
        }
        else {
		if (yyin) fclose(yyin);
	        if (vflag) fprintf(stderr,") ");
                yy_delete_buffer( YY_CURRENT_BUFFER );
                yy_switch_to_buffer(
                     include_stack[include_stack_ptr] );
        }
    }
}

<MAN,A>{

    \\"*(rq"	mansym('r','q');
    \\"*(lq"	mansym('l','q');
    \\"*R"	mansym('r','g'); /* the registered sign */
    \\"*(Tm"	mansym('t','m'); /* the trademark sign */
    \\"("{alnm}{alnm}	mansym(yytext[2], yytext[3]);

    \\"s"("-"|"+")?[0-9] |
    "\\*S"		|
    ^".DT"		|
    ^".ta ".*	|
    ^".PD".*	|
    ^".NX".*			;
}

<T,A>\\beginC/{na}	{
		BEGIN(C);
		pindent = 0;
		obflag = TRUE;
		obsflag = TRUE;
		cflag = TRUE;
		xflag = FALSE;
}
<C>\\endC/{na}  {
		obflag = FALSE;
		obsflag = FALSE;
		cflag = FALSE;
		xflag = TRUE;
		if (mflag) BEGIN(A);
		else BEGIN(T);
}

<CCODE,C>{

    "//"	{
	if (!qtflag && attribute != TTYPE) {
		int ch = input();
		attribute = TTYPE;
		puttc('/');
		puttc('/');
		while (ch != '\n' && ch != EOF) {
			puttc(ch);
			ch = input();
		}
		putword(); putline();
		attribute = ROMAN;
	}
	else {
		puttc('/');
		puttc('/');
	}
    }
    \"	{	if (attribute != TTYPE) qtflag = !qtflag;
			puttc('"');
    }

    "/*"	{	puttc('/');
			if (!qtflag) attribute = TTYPE;
			puttc('*');
    }

    "*/"	{	puttc('*');
			if (!qtflag) attribute = ROMAN;
			puttc('/');
    }

    /* control flow keywords are bold */
    "continue"|"default"|"do"|"goto"|"return"  |
    "if"|"else"|"while"|"break"|"switch"|"case"|"for"  {
			if (attribute != TTYPE && !qtflag) {
				attribute = BOLD;
				putts(yytext);
				attribute = ROMAN;
			} else putts(yytext);
    }

    /* storage class keywords are underlined */
    ("auto"|"char"|"double"|"enum"|"extern"|"float")  |
    ("int"|"long"|"register"|"short"|"sizeof"|"static") |
    ("struct"|"typedef"|"union"|"unsigned"|"void") {

			if (attribute != TTYPE && !qtflag) {
				attribute = ITALIC;
				putts(yytext);
				attribute = ROMAN;
			} else putts(yytext);
    }

    /* preprocessor keywords are blinking */
    "#"[ \t]*("define"|"undef"|"include"|"if"|"ifdef")  |
    "#"[ \t]*("ifndef"|"else"|"endif"|"line")  {

			if (attribute != TTYPE && !qtflag) {
				attribute = SLANTY;
				putts(yytext);
				attribute = ROMAN;
			} else putts(yytext);
    }

    /* identifiers that happen to contain keywords are plain */
    [a-zA-Z_]+	putts(yytext);

}

.  {

	unsigned char ch = yytext[0];

	if (ch < 0x80  || !gflag) {
	    if (is_ipa(attribute)) ch = let[2][ch];
	    else if (attribute == MATH) ch = let[1][ch];
	    else if (is_roman(attribute)) ch = let[0][ch];
	    PUTTC(ch);
	}
	else putuni(ch);
}


%%

static void eat()
{
	int c;
	while( ( (c = input()) == ' ' || c == '\t' ) && c != EOF ) ;
	unput(c);
}

static int     old_gluecnt = 0;
static unsigned char   *old_glueptr[MAXGLUE];		/* glueptr[0] is unused */
static int     old_stretchcnt = 0;
static unsigned char   *old_stretchptr[MAXGLUE];
static int     old_shrinkcnt = 0;
static unsigned char   *old_shrinkptr[MAXGLUE];

static unsigned char
	old_oline[MAXLINE] = "\0",	/* save output line buffer */
       *old_olp = old_oline,
        old_aline[MAXLINE] = "\0",	/* save output line attribute buffer */
       *old_alp = old_aline;
static int old_spaces, old_curr_vskip, old_curr_vskip_pt, old_curr_vglue;
static int old_pindent;
static int old_underscoring, old_curr_badness, old_true_badness;
static int old_lskip, old_rskip, old_hangafter, old_tpstate, old_indent;

static void save_line() {
	int i;
	unsigned char *p = oline, *q = aline, *m = old_oline, *n = old_aline;
	while (p < olp) {
		*m++ = *p++;
		*n++ = *q++;
	}
	old_olp = olp;
	old_alp = alp;
	old_spaces = spaces;
	old_pindent = pindent;

	old_underscoring = underscoring;
	underscoring = FALSE;
	old_curr_vskip = curr_vskip;
	old_curr_vskip_pt = curr_vskip_pt;
	old_curr_vglue = curr_vglue;
	old_curr_badness = curr_badness;
	old_true_badness = true_badness;
	old_lskip = lskip;
	old_rskip = rskip;

	olp = oline;
	alp = aline;
	curr_vskip = curr_vskip_pt = 0;
	curr_vglue = 0;
	true_badness = curr_badness = 0;
	lskip = rskip = 0;
	spaces_pp = spaces = 0;
	pindent = 0;

	for (i = 0; i < MAXGLUE; i++) {
		old_glueptr[i] = glueptr[i];
		old_stretchptr[i] = stretchptr[i];
		old_shrinkptr[i] = shrinkptr[i];
	}
	old_gluecnt = gluecnt;
	gluecnt = 0;
	old_stretchcnt = stretchcnt;
	stretchcnt = 0;
	old_shrinkcnt = shrinkcnt;
	shrinkcnt = 0;

	old_hangafter = hangafter;
	hangafter = 0;
	old_tpstate = tpstate;
	tpstate = 0;
	old_indent = indent;
	indent = 0;
}
static void restore_line() {
	int i;
	unsigned char *p = oline, *q = aline, *m = old_oline, *n = old_aline;
	while (p < old_olp) {
		*p++ = *m++;
		*q++ = *n++;
	}
	olp = old_olp;
	alp = old_alp;
	spaces = old_spaces;
	/* spaces_pp = SP(spaces); */
	pindent = old_pindent;

	for (i = 0; i < MAXGLUE; i++) {
		glueptr[i] = old_glueptr[i];
		stretchptr[i] = old_stretchptr[i];
		shrinkptr[i] = old_shrinkptr[i];
	}
	gluecnt = old_gluecnt;
	stretchcnt = old_stretchcnt;
	shrinkcnt = old_shrinkcnt;

	underscoring = old_underscoring;
	curr_vskip = old_curr_vskip;
	curr_vskip_pt = old_curr_vskip_pt;
	curr_vglue = old_curr_vglue;
	curr_badness = old_curr_badness;
	true_badness = old_true_badness;
	lskip = old_lskip;
	rskip = old_rskip;
	hangafter = old_hangafter;
	tpstate = old_tpstate;
	indent = old_indent;
}

void push(unsigned char attr)
{
		if (level > STACKDEPTH - 2) {
			windup();
			fprintf(stderr,"Too much group nesting.\n");
			exit(1);
		}
		attrstack[level++] = attribute; /* push attr. onto stack */
		lrstack[level] = lrstack[level-1];
		l_hsize[level] = l_hsize[level-1];
		curr_leading[level] = curr_leading[level-1];
		curr_leading_pt[level] = curr_leading_pt[level-1];
		attribute = attr;
}

void pop()
{
		/* if emerging from {} in which \obeylines or
		 * \raggedright occurred, cancel  (not a very general
		 * strategy)
		 */
		if (obflag == level) obflag = FALSE;
		if (obsflag == level) obsflag = FALSE;
		if (njflag == level) njflag = FALSE;
		if (rhmode == level) {
			rhmode = FALSE;
			hboxsize = 0;
		}

		/* pop attribute off stack */
		if (level > 1) attribute = attrstack[--level];

		/* } after \item signals end of tag; will not
		 * work right with {} inside tag */
		if (tpstate == 1 && tplevel == level)  {
			tpstate = 2;
			putword();
			if (olp == oline) {
				tpstate = 4;
				indent = tpvalue;
			}
		}
		if (vcenterflag == level) {
			lbreak();
			insertionflag = FALSE;
			curr_insertion++;
			vcenterflag = FALSE;
			rvmode = FALSE;
		}
		if (fnflag == level) {
			lbreak();
			insertionflag = FALSE;
			curr_insertion++;
			fnflag = FALSE;
			restore_line();
		}
}

static void parpop() {
	if (popme) pop();
	popme = FALSE;
}

#define MAXTREE 2048
static char *treebuf = NULL;
static int tree_total = 0;

/* call puts_tbuf before this */
static void putc_tbuf(int c) {
	if (tree_total + 1 >= MAXTREE) {
		windup();
		fprintf(stderr,"Tree too big.\n");
		exit(1);
	}
	treebuf[tree_total++] = c;
}
static void puts_tbuf(char *base, int count) {
	if (treebuf == NULL) {
		treebuf = (char *)malloc(MAXTREE);
		if (treebuf == NULL) {
			windup();
			fprintf(stderr,"Could not get memory for tree.\n");
			exit(1);
		}
	}
	if (tree_total + count >= MAXTREE) {
		windup();
		fprintf(stderr,"Tree too big.\n");
		exit(1);
	}

	memcpy(treebuf + tree_total, base, count);
	tree_total += count;
}

static void finish_tbuf() {
	treeflag = TRUE;
	scantree(treebuf, tree_total, tree_indent);
	lbreak();
	treeflag = FALSE;
	tree_total = 0;
}

/* tpar stuff */

static void descend(int d_indent)
{	int i;

	if (t_extra) do putc_tbuf(')'); while (--t_extra);
	if (d_indent > buried[t_level]) {
		buried[++t_level] = d_indent;
	}
	else {
		putc_tbuf(')');
		while (t_level >= 0 && d_indent < buried[t_level]) {
			t_level--;
				/* optional pretty print */
				putc_tbuf('\n');
				for (i = 1; i < buried[t_level]; i++)
					putc_tbuf(' ');
			putc_tbuf(')');
		}
	}
	putc_tbuf('\n');
}

static void ascend()
{	int i;

	if (t_indent) return;
	descend(t_indent = outercol+1);
	if (t_indent < minindent) minindent = t_indent;
	for (i = 1; i < t_indent; i++) putc_tbuf(' ');
	putc_tbuf('(');
}

static void xno(char *s) {
	static int exno = 0, sxno = 0, cxno = 0, fxno = 0, gxno = 0;
	static int cexno = 0, csxno = 0, ccxno = 0, cfxno = 0, cgxno = 0;
	char tpa, tpb;
	char rs[50];
	int n = 0, m, caps = FALSE;

	tpa = s[1];
	tpb = s[3];

	switch (tpa) {
		case 'e': n = exno; caps = cexno; break;
		case 'f': n = fxno; caps = cfxno; break;
		case 'g': n = gxno; caps = cgxno; break;
		case 'c': n = cxno; caps = ccxno; break;
		case 's': n = sxno; caps = csxno; break;
	}
	if (tpb == 'o') {
		if (n < 0) n--;
		else n++;
	}
	if (n < 0) m = -n;
	else m = n;

	switch (tpb) {
		case 'a': m = n = 0; caps = 3; break;
		case 'A': m = n = 0; caps = 4; break;
		case 'b': m -= 5; break;
		case 'c': m -= 4; break;
		case 'd': m -= 3; break;
		case 'e': m -= 2; break;
		case 'p': m -= 1; break;
		case 'l': break;
		case 'o': break;
		case 'n': m += 1; break;
		case 's': m += 2; break;
		case 'f': m += 3; break;
		case 'g': m += 4; break;
		case 'h': m += 5; break;
		case 'i': m = n = -1; caps = FALSE; break;
		case 'I': m = n = -1; caps = TRUE; break;
		default: m = n = atoi(s + 3);
			 if (n >= 0) caps = FALSE;
			 break;
	}

	if (n < 0 && m >= 0) m = -m;
	if (caps > TRUE && m < 0) m = 0;

	switch (tpa) {
		case 'e': exno = n; cexno = caps; break;
		case 'f': fxno = n; cfxno = caps; break;
		case 'g': gxno = n; cgxno = caps; break;
		case 'c': cxno = n; ccxno = caps; break;
		case 's': sxno = n; csxno = caps; break;
	}

	if (m < 0) {
		romanize(m, rs);
		if (caps == TRUE) for (n = 0; rs[n]; n++)
			rs[n] = toupper(rs[n]);
		sprintf(owp, "%s", rs);
	}
	else if (caps > TRUE) {
		n = m % 26;
		m /= 26;
		if (m) {
			m--;
			m = m % 26;
			if (caps == 3) *owp++ = m + 'a';
			else *owp++ = m + 'A';
			*awp++ = attribute;
		}
		if (caps == 3) *owp++ = n + 'a';
		else *owp++ = n + 'A';
		*awp++ = attribute;
		*owp = '\0';
	}
	else sprintf(owp, "%d", m);

	while (*owp) {
		owp++;
		*awp++ = attribute;
	}
}


static void usage(char *prog) {
  fprintf(stderr, "Usage: %s [options] [files]\n", prog);
  fprintf(stderr, "options = -a\tascii display toggle\n");
  fprintf(stderr, "          -b\tbackspace and overstrike for bold, etc.\n");
  fprintf(stderr, "          -c\tC source\n");
  fprintf(stderr, "          -d<n>\tdevice for non-PS output at pointsize <n>\n");
  fprintf(stderr, "          -e\tescaped ascii display\n");
  fprintf(stderr, "          -f<n>\tfont <n=0-4> for PS output[0]\n");
  fprintf(stderr, "          -F\tfilter -- do not paginate\n");
  fprintf(stderr, "          -h<n>\thyphenation level <n=0-9>[8]\n");
  fprintf(stderr, "          -i\tindent as in source file toggle\n");
  fprintf(stderr, "          -j\tjustification off\n");
  fprintf(stderr, "          -l<n>\tleft margin <n> columns\n");
  fprintf(stderr, "          -m\tman macro source\n");
  fprintf(stderr, "          -n\tno color display\n");
  fprintf(stderr, "          -N\tno formatting\n");
  fprintf(stderr, "          -p<n>\tpointsize <n> for PS output\n");
  fprintf(stderr, "          -#<n>\tdo <n> copies for PS output\n");
  fprintf(stderr, "          -q\tquiet about pages output\n");
  fprintf(stderr, "          -t<n>\ttab width is <n> columns[8]\n");
  fprintf(stderr, "          -u\tunicode output\n");
  fprintf(stderr, "          -U\tno unicode output[default]\n");
  fprintf(stderr, "          -v\tversion display and exit\n");
  fprintf(stderr, "          -w<n>\twidth of line <n> columns[78]\n");
  fprintf(stderr, "          -x\tTeX source\n");
  fprintf(stderr, "          -y\tno formatting, monospaced print\n");
  fprintf(stderr, "          -z\tdebug line penalties\n");
  fprintf(stderr, "          -1\tLatin-1 input -- not unicode\n");
}

/*
 * Characters, along with the currently active "font"= screen
 * attribute, are collected in a word buffer until we hit a
 * word break (white space), then moved to a line buffer.
 * Words are collected in the line buffer until we hit a line
 * break (end of paragraph) or the line is long enough, then
 * the line is put into a page buffer.  When the page is long
 * enough, it is output.
 */
int main(int argc, char **argv)
{
	int c, hsize_set = FALSE;

	myname = strdup( getenv("USER") );

#ifdef HAVE_STRFTIME
	time(&rightnow);

	/* \date */
	strftime(date_str, 20, "%D", localtime( &rightnow ));
	/* \today */
	strftime(today_str, 20, "%B %e, %Y", localtime( &rightnow ));
	/* \day */
	strftime(day_str, 20, "%A", localtime( &rightnow ));
	/* \time */
	strftime(time_str, 20, "%l:%M %P", localtime( &rightnow ));
#endif

	{	int i, j;
		for (i = 0; i < 3; i++)
		    for (j = 0; j < 256; j++)
			let[i][j] = j;
	}

	progname = basename(argv[0]);

	if (!strcmp(progname,"fan")) {
		mflag = TRUE;
		oflag = TRUE;
		uflag = FALSE;
		jflag = TRUE;
	}
	else if (!strcmp(progname,"fex")) {
		xflag = TRUE;
		jflag = TRUE;
	}

	fmcfg();

	while ((c = getopt (argc, argv, "1abceFijkmnNquUvxyzf:d:l:h:p:t:w:#:")) != EOF)
		switch (c) {
			case 'i': stiflag = TRUE; break;
			case 'j': jflag = FALSE; break;
			case 'c': cflag = TRUE; break;
			case 'F': fflag = TRUE; vflag = FALSE; break;
			case 'x': xflag = TRUE; break;
			case 'y': yflag = TRUE; break;
			case 'z': zflag = TRUE; break;
			case 'm': mflag = TRUE; break;
			case 'n': nflag = TRUE; break;
			case 'b': oflag = TRUE; uflag = FALSE; break;
			case 'a': staflag = TRUE; break;
			case 'e': uflag = oflag = aflag = kflag = FALSE; break;
			case 'w': hsize = max(0, atoi(optarg)); break;
			case 'h': hlevel = max(0, atoi(optarg)); break;
			case 'p': ps_pointsize = atof(optarg);
				  vflag = psflag = TRUE; break;
			case 'l': lmargin = SP(atoi(optarg)); break;
			case 't': tabsize = max(1, atoi(optarg)); break;
			case 'd': outdev = max(0, atoi(optarg));
				  vflag = TRUE;
				  ps_pointsize = (float)max(1,outdev); break;
			case 'u': kflag = TRUE; break;
			case 'U': kflag = FALSE; break;
			case '1': gflag = FALSE; break;
			case 'N': liflag = obflag = obsflag = -1; jflag = FALSE; hlevel = 0; break;
			case 'q': vflag = FALSE; break;
			case 'f': bfont = max(0, atoi(optarg)); break;
			case '#': numcopies = max(1, atoi(optarg)); break;
			case 'v': printf("fm 1.0\n");
				  exit(0);
			default:  usage(progname);
				  exit (1);
		}

	if (optind < argc) {
		char *file = argv[optind];
		int l = strlen(file);
		if (!strncmp(file+l-4,".tex",4)) xflag = TRUE;
		else if (!strncmp(file+l-4,".man",4)) mflag = TRUE;
		else if (!strncmp(file+l-2,".c",2)) cflag = TRUE;
		else if (!strncmp(file+l-2,".h",2)) cflag = TRUE;
		else if ( *(file+l-2) == '.' &&
			  *(file+l-1) >= '1' &&
			  *(file+l-1) <= '9'
			 ) mflag = TRUE;
		pathname(file);
	}

	/* make sure point size is reasonable */
	if (ps_pointsize < 1.0) ps_pointsize = 12.0;
	ps_pointsize10 = (int)(10.0 * ps_pointsize + 0.5);
	ps_pitch_sp = 720000 / ps_pointsize10;
	ps_pitch = 144.0 / ps_pointsize;
	if (ps_baselineskip < 0.0)  { /* may have been set in cfg */
		if (ps_pointsize10 < 15)
			 ps_baselineskip10 = ps_pointsize10 + 10;
		else ps_baselineskip10 = ps_pointsize10 + 20;
	}
	ps_vpitch_sp = 360000 / ps_baselineskip10;

	ps_baselineskip = (ps_baselineskip10 + 5) / 10;
	baselineskip = 1;

	for (c = 0; c < MAXTL; c++) treeline[c].status = 0;


	if (hlevel > 9) hlevel = 9;
	cleartabs();
	setdefaulttabs();
	lrstack[level] = 0;

/* set lex start state */
	if (yflag) {
		staflag = TRUE;
		xflag = mflag = cflag = FALSE;
		vbbflag = TRUE;
		attribute = TTYPE;
	}

	if (xflag && mflag) {
		if (stiflag) iflag = TRUE;
		if (staflag) { uflag = FALSE; aflag = TRUE; }
		pindent = parindent;
		BEGIN (A);
	}
	else if (xflag) {
		if (stiflag) iflag = TRUE;
		if (staflag) { uflag = FALSE; aflag = TRUE; }
		pindent = parindent;
		BEGIN (T);
	}
	else if (mflag) {
		if (stiflag) iflag = TRUE;
		if (staflag) { uflag = FALSE; aflag = TRUE; }
		parindent = 0;
		indent = 5;
		BEGIN (MAN);
	}
	else if (cflag) {
		if (stiflag) iflag = TRUE;
		if (staflag) { uflag = FALSE; aflag = TRUE; }
		parindent = 0;
		obflag = TRUE;
		obsflag = TRUE;
		BEGIN (CCODE);
	}
	else {
		if (!stiflag) iflag = TRUE;
		if (!staflag || yflag) { staflag = TRUE; uflag = FALSE; aflag = TRUE; }
		BEGIN (PLAIN);
	}

	if (!uflag && !oflag && !aflag) kflag = FALSE;

	ttyinit();

	if (uflag && (out_is_tty || !psflag)) terminit();

	if (out_is_tty || !in_is_tty) vflag = FALSE;
	if (vflag) fprintf(stderr,"fm 1.0");

	if (hsize < 0) {
		if (pointsize == 12 && !mflag) hsize = 78; /* 6 1/2" */
		else /* 10 */ hsize = 72;
	}
	else hsize_set = TRUE;

	if ((out_is_tty || !hsize_set) && hsize > cols) hsize = cols;

	hsize_sp = SP(hsize);

	/* page dimensions will not have been set correctly ... */
/* should the next 3 lines come right after ttyinit() ? */
	if (out_is_tty) psflag = FALSE;
	if (kflag || psflag) boxflag = TRUE;
	if (psflag) kflag = FALSE;

	/* init page device */
	devinit();

	if (vsize < 0) vsize = 76;
	l_hsize[level] = hsize;

	curr_leading[level] = 0;
	curr_leading_pt[level] = 0;

/* call lex scanner */
	if (optind >= argc) {
		if (in_is_tty) {
			windup();
			usage(progname);
			exit(1);
		}
		yylex();
		lbreak();
	} else for (; (optind < argc); optind++) {
		char *file = argv[optind];
		yyin = fopen(file, "r");
		if (vflag && yyin != NULL) fprintf(stderr,"\n%s ", file);
		if (yyin == NULL) {
			char *tfname = (char *)malloc(strlen(file)+5);
			strcpy(tfname, file);
			strcat(tfname, ".tex");
			if (!xflag || ( (yyin = fopen (tfname, "r")) == NULL)) {
				windup();
				fprintf(stderr,
			    	"Couldn't open file: %s\n", argv[optind]);
				exit (1);
			}
			if (vflag) fprintf(stderr,"\n%s ", tfname);
		}
		yy_init = 1;
		yylex();
		lbreak();
		fclose(yyin);
		yyin = NULL;
	}
	windup();

	if (rhmode) fprintf(stderr,"** Ended inside an hbox.\n");
	if (tbcnt) fprintf(stderr,"** Ended inside a tab line.\n");
	if (level > 1) fprintf(stderr,"** Ended inside a group.\n");
	return 0;
}				/* main */

/* preserve spacing due to spaces or tabs */
static void	copyindent (char *ilp)
{
	int     col;
	unsigned char whitetype = attribute;

	if (olp == oline) whitetype = ROMAN;
	else whitetype = *(alp - 1);

	col = (olp - oline) + 1;

	for (; (isspace (*ilp)); col++) {
		if (*ilp++ == '\t')
			for (; (col % tabsize); col++) {
				*olp++ = ' ';
				*alp++ = whitetype;
			}
		*olp++ = ' ';
		*alp++ = whitetype;
	}
}				/* copyindent */

/* one character goes into word buffer */
void puttc(int ch) {
	if (ch < 0 || ch > 255) {
		windup();
		fprintf(stderr,"puttc: ch=%d is out of bounds\n", ch);
		exit(1);
	}
	if (owp - oword >= MAXWORD) return;
	if (ch != LIT_ESC && attribute == IPA) attribute = ipa_attribute;
	*owp++ = ch;
	*awp++ = attribute;
	if (attribute == IPAOVER || attribute == IPAUNDER)
	  ipa_attribute = attribute = IPA;
	else if (attribute == TITLE && psflag && !curr_vglue) curr_vglue = 1;
}				/* puttc */

void puttc_text(int ch) {
	if (is_ctl(ch)) puttc(LIT_ESC);
	puttc(ch);
}

/* put string in word buffer (might have spaces) */
void putts(unsigned char *s) {
	while (*s)
		puttc (*s++);
}				/* putts */

/* finish pending word and line and put it out (no justification) */
static void lbreak() {
	char saveflag = nopindentflag;
	putword();
	putline();
	nopindentflag = saveflag;
	if (xflag && !obflag && !liflag && !iflag && !nopindentflag &&
	    !ceflag && !laflag && !raflag) {
		pindent = parindent;
		pskip = parskip;
		pskip_pt = parskip_pt;
	}
	else {
		if (xflag || mflag || iflag) pindent = 0;
		else pindent = parindent;
		pskip = 0;
		pskip_pt = 0;
	}
	hangafter = tpstate = tpmore = 0;
	if (xflag) indent = 0;
	tplevel = level;
	liflag = FALSE;
	true_badness = 0;
	lhpenalty = FALSE;
	parpop();
}				/* lbreak */

static int ctlspaceflag = 0;
static void putwordctlspace() {
	ctlspaceflag = 1;
	putword();
}

static void setindent() {
    if (tpstate) {
	if (hangafter < 0) pindent = tpvalue + tpmore;
	else pindent = tpmore;
    }
    if (hangafter > 0) pindent = 0;
}

/* append pending word to line buffer */
void putword () {
    int plen, hyph_len, curr_len, curr_indent, curr_room;
    int plen_pp, hyph_len_pp, curr_len_pp, curr_room_pp;
    int future_room;
    int	extra_len, end_len = 0, stretch = 0;
    int	i, hyph_ok = TRUE;
    unsigned char *p, *q;
    unsigned char thisatt, prevatt;


    /* just return if no characters have been accumulated yet */
    if (owp == oword) {
	/* this is for the special case in which a file
	 * starts with \n, which should count as  a "blank line"
	 */
	if (!started) {
		if (!xflag) curr_vskip = 1;
		started = 1;
	}
	return;
    }
    started = 1;

    curr_len = olp - oline;
    curr_len_pp = pnwidth(oline, aline, curr_len);

    setindent();
    curr_indent = indent + pindent + lrstack[level] + lskip;
    curr_room = l_hsize[level] - curr_len -
		curr_indent - lrstack[level] - rskip;
    curr_room_pp =
      SP((l_hsize[level] - curr_indent - lrstack[level] - rskip)) - curr_len_pp;

    plen = owp - oword;
    /* puttc ('\0'); */
    p = oword;
    q = aword;
    /* plen = strlen (p); */
/*
    if (psflag) plen_pp = pnwidth(p, q, plen);
    else plen_pp = SP(plen);
*/
    plen_pp = pnwidth(p, q, plen);

    if (treeflag) hyph_ok = FALSE;
    else for (i = 0; i < plen; i++) {
	if (p[i] <= ' ') hyph_ok = FALSE;
	if (is_ipa(q[i])) hyph_ok = FALSE;
	if (q[i] == MUSIX || q[i] == BOX) hyph_ok = FALSE;
    }

    if (curr_len <= 0) prevatt = ROMAN;
    else prevatt = *(alp-1);

    if (uflag || oflag || aflag) extra_len = 0;
    else {
	if (prevatt == *q) extra_len = 0;
	else if (prevatt == ROMAN) extra_len = 1;
	else if (*q == ROMAN) extra_len = 1;
	else extra_len = 2;
	if ( *(q + plen - 1) != ROMAN) end_len = 1;
    }

	/* ugly way to ignore "0.3i" arg of .IP */
    if (tpstate == 2 && *p == '0' && p[plen - 1] == 'i') {
	spaces = plen = plen_pp = 0;
	*p = '\0';
    }

    /* finish pending line if pending word won't fit */
    hyph_len = plen;  /* assume no hyphenation */
    hyph_len_pp = plen_pp;  /* assume no hyphenation */

    if (!spaces) spaces_pp = 0;
    else if (!psflag) spaces_pp = SP(spaces);

    future_room = curr_room_pp - spaces_pp - SP((extra_len + end_len + prev_extra));

    /* no fill when obeylines, literal, tabbing line, inside hbox */
    if (!obflag && !liflag && !tbcnt && !rhmode) {

	int bad, underfull;
	int shrinkage = 0;
	int realshrink = shrinkcnt;

	bad = SP(hlevel);
	/* how much room will be left for this word? */

	if (shrinkcnt && shrinkptr[shrinkcnt] >= olp) realshrink--;

	underfull = future_room + spaces_pp;
	/* if room is small, don't do it */
	if (future_room < SP(3) && gluecnt > 2) bad = 0;
	else if (bad < SP(9)) {
		bad += underfull - SP((l_hsize[level]/8));
		if (underfull > SP(5))
		  bad += underfull - SP(5);
	}

	if (bad > SP(7) && future_room < plen_pp && future_room + SP(realshrink) >= plen_pp) {
		shrinkage = COL(plen_pp) - COL(future_room);
		shrinkline(&shrinkage);
		if (shrinkage) {
			future_room += SP(shrinkage) + spaces_pp;
			curr_len_pp -= SP(shrinkage);
		}
	}

	if (lhpenalty) bad -= SP(3);

	/* if short of room, try hyphenating */
	if (hlevel && bad > SP(7) && future_room < plen_pp && hyph_ok) {
		hyph_len = hyphenate(oword, aword, plen, COL(plen_pp), COL(future_room));
		if (hyph_len < plen) {
			hyph_len_pp = pnwidth(oword, aword, hyph_len - 1);
			hyph_len_pp += pnwidth("-", aword + hyph_len -1, 1);
		}
	}
	/* does it help that much? */
	if (bad < SP(5) && hyph_len_pp < SP(4)) {
		hyph_len = plen;
		hyph_len_pp = plen_pp;
	}

	/* if word is too long for line, make arbitrary break */
	if (hyph_len_pp > SP((l_hsize[level] - curr_indent)) ) {
		 hyph_len = COL(future_room);
		 hyph_len_pp = future_room;
	}

	/* when doesn't fit, start a new line */
	if (future_room < hyph_len_pp) {
		hyph_len = plen;
		hyph_len_pp = plen_pp;
		lhpenalty = FALSE;
		if ( *(alp - 1) == ROMAN || aflag) end_len = 0;
		else end_len = 1;
		if (uflag || aflag) extra_len = prev_extra = end_len = 0;
		if (jflag && !njflag)
			justifyline(curr_len_pp, prev_extra + end_len);
		amwidow = -1;
		putline();
		prevatt = ROMAN;
		extra_len = 0;
		if (*q != prevatt && !oflag && !uflag)
			prev_extra = 1;
	/* Shouldn't I set spaces = 0 here?? No, putline did. */
	}
    }

    if (uflag || oflag || aflag) extra_len = 0;

    /* append spaces which were previously decided to go after
     * last word put in line buffer */
    if (spaces) {
	char spcatt;
	if (!tbcnt && /* prevatt != MATH && prevatt != TTYPE &&*/ !rhmode) {
	  /* don't use space inside tag of hanging indent to
	   * right justify */
	  if (!tpstate || tpstate > 3)
		glueptr[++gluecnt] = olp;
	}
	/* if tag already padded out, permit justification
	 * of space after next word */
	if (tpstate == 3)
		tpstate = 4;
	if (prevatt != *q || *q == ROMAN) spcatt = ROMAN;
	else spcatt = prevatt;
	curr_len_pp += spaces_pp;
	for (; (spaces > 0); spaces--) {
		*olp++ = ' ';
		*alp++ = spcatt;
	}
	if (prevatt != spcatt) extra_len = 1;
	else extra_len = 0;
	if (spcatt != *q) extra_len++;
    }

    /* append first part of hyphenated word to line */
    if (hyph_len < plen) {
	int j;
	thisatt = *q;
	curr_len_pp += hyph_len_pp;
	for (j = 1; j < hyph_len; j++) {
		if (thisatt != *q) {
			extra_len++;
			thisatt = *q;
		}
		*olp++ = *p++;
		*alp++ = *q++;
		plen--;
	}
	/* for hyphenation break after '-', include it */
	if (*p == '-') {
		*olp++ = *p++;
		*alp++ = *q++;
		plen--;
	}
	/* add hyphen */
	if (*(olp-1) != '-' && *(olp-1) != ' ') {
		*olp++ = '-';
		*alp++ = *(q - 1);
	}
	else curr_len_pp -= SP(1);
	if (*(alp - 1) == ROMAN || aflag) end_len = 0;
		else end_len = 1;
	/* line is now complete, so print and start a new one */
	if (uflag || aflag) extra_len = prev_extra = end_len = 0;
	if (jflag && !njflag)
		justifyline(curr_len_pp, extra_len + prev_extra + end_len);
	amwidow = -1;
	putline();
	lhpenalty = TRUE;
	prevatt = ROMAN;
	extra_len = 0;
	if (*q != prevatt && !oflag && !uflag)
		prev_extra = 1;
	/* when remnant of word is too long for next line, move it back
         * to beginning of word buffer and recurse */
	if (plen > olp - oline) {
		unsigned char *owp_orig = owp;
		owp = oword;
		awp = aword;
		while (p < owp_orig) {
			*owp++ = *p++;
			*awp++ = *q++;
		}
		putword ();
		return;
	}
    }

    if (rhmode && hboxsize) {
	hboxsize -= plen + spaces;
    }
    if (tbcnt) {
/*
        if (psflag) plen_pp = pnwidth(p, q, owp - p);
        else plen_pp = SP((owp - p));
*/
        plen_pp = pnwidth(p, q, owp - p);
	tbsize += plen + spaces;
	tbsize_sp += plen_pp + spaces_pp;
    }

    thisatt = *q;
    stretch = 0;
    if (!tpstate || tpstate > 3) {
	if (is_roman(thisatt)) stretch = endofsentence(p, plen);
    }

    /* append pending word in line buffer */
    while (p < owp) {
	if (thisatt != *q) {
		extra_len++;
		thisatt = *q;
	}
	if (!psflag || thisatt != *(q+1) || thisatt == TTYPE ||
		 thisatt == MATH || thisatt == MUSIX) {
		*olp++ = *p++; *alp++ = *q++;
		continue;
	}
    /* ligatures */
      if (p + 1 < owp) {
	if (is_ipa(thisatt)) {
	    if (*p == 34 && *(p + 1) == 34) {
		p++;
		*olp++ = 0x7f; *alp++ = *q++;
		p++;  q++;
		continue;
	    }
	    if (*p == 124 && *(p + 1) == 124) {
		p++;
		*olp++ = 0x7b; *alp++ = *q++;
		p++;  q++;
		continue;
	    }
	}
	else if (*p == 'f' && *(p + 1) == 'i') {
		p++;
		*olp++ = PS_fi; *alp++ = *q++;
		p++;  q++;
		continue;
	}
	else if (*p == 'f' && *(p + 1) == 'l') {
		p++;
		*olp++ = PS_fl; *alp++ = *q++;
		p++;  q++;
		continue;
	}
      }
	*olp++ = *p++;
	*alp++ = *q++;
    }

    /* figure spaces which should go after pending word which
     * is just about to be appended in line buffer -- remember
     * so they can be put before next word (cf. just above) */
    if (!obsflag && !liflag) {
	spaces = 1;
	spaces_pp = WORDSPACE;
	if (stretch/2 && !ctlspaceflag && !tbcnt && (psflag ||!frenchspacing)) {
		spaces += stretch / 2;
		spaces_pp = SENTENCESPACE;
		shrinkptr[++shrinkcnt] = olp;
	}
	if (stretch && !tbcnt && !rhmode /* && prevatt != TTYPE && prevatt != MATH*/) {
		stretchptr[++stretchcnt] = olp;
		if (stretch == 1) spaces_pp = COMMASPACE;
	}
	if (!psflag) spaces_pp = SP(spaces);
    }
    ctlspaceflag = 0;

    if (!oflag && !uflag &&!aflag) prev_extra += extra_len;

    /* part of tortuous way of handling hanging indent */
	/* tpstate = 0: no hanging indent
	 *	     1: scanning tag to go in left margin
	 *	     2: have seen end of tag, time to
	 *		pad out tag to hanging indent value
	 *	     3: finished padding
	 *	     4: finished first line of hanging par.;
	 *		subsequent lines should be indented
	 *		to hanging indent value
	 */
    if (tpstate == 2) {	/* time to pad tag? */
	curr_len = olp - oline;
/*
	if (psflag) curr_len_pp = pnwidth(oline, aline, curr_len);
*/
	curr_len_pp = pnwidth(oline, aline, curr_len);
	tpstate = 3;	/* signal padding done */
	if (psflag) {
	    if (curr_len_pp + spaces_pp > SP(tpvalue)) { /* will tag fit? */
		amwidow = -1;
		putline();	/* no it won't, so start new line */
	    }
	    else {		/* yes it will, so pad it out */
		*olp++ = COL_ESC; *alp++ = ROMAN;
		*olp++ = tpvalue + 1; *alp++ = ROMAN;
		spaces_pp = spaces = 0;
	    }
	}
	else {
	    if (curr_len + spaces > tpvalue) { /* will tag fit? */
		amwidow = -1;
		putline();	/* no it won't, so start new line */
	    }
	    else {		/* yes it will, so pad it out */
		while (curr_len + spaces < tpvalue)
			spaces++;
	    }
	}
    }

    /* purge word buffer -- get ready for next word */
    owp = oword;
    awp = aword;
}				/* putword */

/* decrease inter-word spacings for right justification */
static void shrinkline(int *shrinkage) {
	unsigned char *fp, *tp, *fa, *ta;
	int len, k;

	if (shrinkcnt && shrinkptr[shrinkcnt] >= olp)
		shrinkcnt--;
	if (shrinkcnt < *shrinkage) { *shrinkage = 0; return; }

	while (shrinkage--) {
		if (!shrinkcnt) break;
		tp = shrinkptr[shrinkcnt];
		len = tp - oline;
		ta = aline + len;
		fp = tp + 1;
		fa = ta + 1;
		for (k = 1; k <= gluecnt; k++) {
			if (glueptr[k] > tp) glueptr[k]--;
		}
		for (k = 1; k <= stretchcnt; k++) {
			if (stretchptr[k] > tp) stretchptr[k]--;
		}
		while (fp < olp) {
			*tp++ = *fp++;
			*ta++ = *fa++;
		}
		olp--;
		alp--;
		shrinkcnt--;
	}
}

/* increase inter-word spacings for right justification */
static int stretchline(int stretch) {
	unsigned char *fp, *tp, *fa, *ta;
	int ret = TRUE, k;

/*
fprintf(stderr,"\tstretch=%d stretchcnt=%d gluecnt=%d\n",
stretch,stretchcnt,gluecnt);
*/
	if (stretchcnt && stretchptr[stretchcnt] >= olp)
		stretchcnt--;

	if (!stretchcnt) ret = FALSE;
	if (!ret && gluecnt) return ret;

    if (stretchcnt) {
	if (stretch <= 0) return TRUE;
	if (stretch > stretchcnt) {
		stretch = stretchcnt;
		ret = FALSE;
	}
	else ret = TRUE;

	while (stretch--) {
		fp = olp - 1;
		fa = alp - 1;
		tp = olp;
		ta = alp;
		while (fp >= stretchptr[stretchcnt]) {
			*tp-- = *fp--;
			*ta-- = *fa--;
		}
		if (!ret) for (k = 1; k <= gluecnt; k++) {
			if (glueptr[k] >= stretchptr[stretchcnt])
				glueptr[k]++;
		}
		*tp-- = ' ';
		*ta-- = ta[1];
		olp++;
		alp++;
		stretchcnt--;
	}
	if (ret) return TRUE;
    }

    return FALSE;
}

/* increase inter-word spacings for right justification */
static void justifyline(int len_pp, int extra_len) {
	int currlen = l_hsize[level] - indent - pindent
		 - 2*lrstack[level] - extra_len - lskip - rskip;
	int n, target;
	unsigned char *fp, *tp, *fa, *ta;
	static int dir = FALSE;

	if (psflag && gluecnt) {
		stretchline(currlen - COL(len_pp));
		n = hsize_sp;
		if (hsize != l_hsize[level]) n = SP(l_hsize[level]);
		if (lrstack[level]) n -= SP((2*lrstack[level]));
		n -= SP((indent + pindent + lskip + rskip));
		tlen = n;
		return;
	}
	if (psflag) return;

	target = currlen + olp - oline - COL(len_pp);

/*
fprintf(stderr, "at [%c%c ...] currlen=%d target=%d slen=%d c len_pp=%d\n",
oline[0],oline[1],currlen,target,olp-oline, COL(len_pp));
*/
	if (stretchline(currlen - COL(len_pp))) return;

/* this works out unfortunately for unicode or escapes */
	if (!gluecnt) for (fp = oline + 1, fa = aline + 1; fp < olp; fp++, fa++) {
		glueptr[++gluecnt] = fp;
		if (*fp > 0x7f || *fa != ROMAN) {
			gluecnt = 0;
			break;
		}
	}

	if (!gluecnt) return;

	dir = (!(dir));
	fp = olp - 1;
	fa = alp - 1;
	olp = &oline[target];
	alp = &aline[target];
	tp = olp - 1;
	ta = alp - 1;
	/* really ought to distinguish stretchabilities here */
	while (tp > fp) {
		while (fp >= glueptr[gluecnt]) {
			*tp-- = *fp--;
			*ta-- = *fa--;
		}
		if (dir)
			n = ((tp - fp) - 1) / gluecnt + 1;
		else
			n = (tp - fp) / gluecnt;
		while (n--) {
			*tp-- = ' ';
			*ta-- = ta[1];
		}
		gluecnt--;
	}
}				/* justifyline */


/* put out pending line in line buffer */
void putline() {
	int vskip = 0, vskip_pt = 0, badness = 0;
	int curr_len = olp - oline;
	int curr_len_pp, hsz, spaces_sp, narrow_sp;

	if (!curr_len) {
		if (treeflag) curr_vskip++;
		return;
	}

	setindent();

	curr_len_pp = pnwidth(oline, aline, curr_len);

	if (l_hsize[level] == hsize) hsz = hsize_sp;
	else hsz = SP(l_hsize[level]);

	narrow_sp = SP(lrstack[level]);

	if (ceflag) {
		spaces_sp = max((hsz - curr_len_pp - 2*narrow_sp) / 2, 1);
		ceflag = FALSE;
	}
	else if (raflag) {
		spaces_sp = max((hsz - curr_len_pp - 2*narrow_sp), 1);
		raflag = FALSE;
	}
	else if (laflag) {
		spaces_sp = 0;
		laflag = FALSE;
	}
	else  {
		spaces_sp = SP((indent + pindent));
		if (pindent) {
			if (psflag) vskip_pt = pskip_pt;
			else vskip = pskip;
			badness -= 50;
		}
		pskip = 0;
	}
	spaces_sp += narrow_sp;

	prev_extra = 0;

	*olp = '\0';

	if (!underscoring) nopindentflag = FALSE;

/*
fprintf(stderr,"spaces_sp = %d (%d col); SP(lskip) = %d (%d lskip)\n",
spaces_sp, COL(spaces_sp), SP(lskip), lskip);
*/
	spaces_sp += SP(lskip);

	if (amwidow == -1) amwidow = 1;
	else if (amwidow == 1) {
		amwidow = 0;
		if (badness >= 0) badness += 100;
	}

	shipline(oline, aline, curr_len, tlen, spaces_sp, vskip + curr_vskip + curr_leading[level],
		 curr_vglue, vskip_pt + curr_vskip_pt + curr_leading_pt[level], underscoring,
		 badness + curr_badness + true_badness,
		(insertionflag)? curr_insertion*16+insertionflag : 0);

	tlen = 0;
	if (underscoring) curr_vskip = 1;
	else curr_vskip = 0;
	curr_vskip_pt = 0;
	curr_vglue = 0;
	if (underscoring) curr_badness = pagestretch / 2;
	else curr_badness = 0;
	lskip = rskip = 0;
	underscoring = FALSE;

	/* purge line buffer -- get ready for new line */
	*oline = '\0';
	olp = oline;
	alp = aline;
	spaces_pp = spaces = 0;
	gluecnt = 0;
	stretchcnt = 0;
	shrinkcnt = 0;
	pindent = 0;

	/* if finished with first line of hanging par., increase
	 * left indent for subsequent lines */
	if (hangafter > 0) {
		hangafter--;
		if (!hangafter) tpstate = 4;
	}
	if (tpstate > 2 || hangafter < 0) {
		if (xflag) indent = tpvalue;
		else indent = tpvalue + 5;
		tpstate = 4;
		if (hangafter < 0) {
			indent = 0;
			hangafter++;
			if (!hangafter) tpstate = 0;
		}
	}
}				/* putline */


void emptyline() {

	shipline(NULL, NULL, 0/*curr_len*/, 0/*tlen*/, 0/*spaces_sp*/, curr_vskip,
		 curr_vglue, curr_vskip_pt, underscoring,
		 0 /*curr_badness + true_badness*/,
		(insertionflag)? curr_insertion*16+insertionflag : 0);
	curr_vskip = 0;
	curr_vskip_pt = 0;
	curr_vglue = 0;

}

/*
 Tab in 4; c = 3; findtab returns 5
*/
static int lasttab = 0;
int dotab(int c, int tb) {
	int t;
	if (c < 0) {
		lasttab = 0;
		return 0;
	}
	lasttab++;
	t = findtab(lasttab, tb);
	if (!t) return c;
	t--;
	if (psflag) return t;
	while (c < t) {
		putchar(' ');
		c++;
	}
/* this erases text before empty columns unnecessarily
	while (c > t) {
		putchar(8);
		c--;
	}
*/
	return c;
}

/* put blank lines */
static void skipline (int n) {
	nopindentflag = FALSE;
	lbreak ();
	curr_vskip += n;
}				/* skipline */

/* decide how many extra spaces go after word */
int  endofsentence(unsigned char *p, int plen) {
	int c, d;

	if (plen < 3) return 0;

	c = *(p + plen - 1);
	d = *(p + plen - 2);

	/* here should look past ), ", 0xbb = right guillemot, etc. */
	if (kflag && plen > 2 && *(p + plen - 3) == UNI_ESC) {
		if (plen < 4) return 0;
		/* right double or single quote */
		if (c == 0x20 && (d == 0x1d || d == 0x19)) {
			c = d = *(p + plen - 4);
			if (plen > 4) d = *(p + plen - 5);
		}
		else return 0;
	}
	else if ( (psflag && c == PS_quotedblright) || (!gflag && c == 0xbb) ) {
		c = d;
		if (plen > 2) d = *(p + plen - 3);
	}
	else if (c == ')' || c == '"' || c == '\'') {
		c = d;
		if (plen > 2) d = *(p + plen - 3);
	}

	if (!gflag && c == 0xa1) return 2; /* inverted ! */
	if (!gflag && c == 0xbf) return 2; /* inverted ? */

	if (strchr(",;", c)) return 1;

	if (!strchr(".:?!", c)) return 0;

	if (c != '.') return 2;

	if (d >= 'A' && d <= 'Z') return 0;

	if (abbr(p, plen)) return 0;

	return 2;
}				/* endofsentence */


/* detect "." which is not real end of sentence */
static int abbr(unsigned char *s, int plen) {
	unsigned char   *p, *q, *r;
	int len;

	while (plen && *s == '(') { s++; plen--; }

	if (!plen) return FALSE;

	q = ".i.e.g.dr.mr.mrs.st.";
	for (; (*q); q++) {
		p = q;
		r = s;
		len = plen;
		while (len && (*p++ == (*r++ | 0x20))) len--;
		if (!len)
			return TRUE;
	}
	return FALSE;
}				/* abbr */

static char  *basename(char *s) {
	char   *p;

	if ( (p = strrchr (s, '/')) )
		return (++p);
	else
		return (s);
}				/* basename */

static  void pathname(char *s) {
	char   *p;

	if (!s) return;
	if (!strlen(s)) return;
	myfile = (char *)strdup(s);
	if ( !(p = strrchr (s, '/')) ) return;
	mypath = (char *)strdup(s);
	if ( (p = strrchr (mypath, '/')) )
		*(p + 1) = '\0';
}				/* pathname */

static int max(int a, int b) {
	return ((a > b) ? a : b);
}				/* max */

/* put parentheses around number of manual section */
void addparens(char *s, int a) {
	nextattribute = attribute;
	attribute = a;
	while (*s && *s != ' ')
		puttc (*s++);
	if (*s == ' ')
		s++, puttc ('(');
	while (*s && *s != ' ')
		puttc (*s++);
	if (*s == ' ')
		s++;
	puttc (')');
	attribute = nextattribute;
	nextattribute = 0;
	if (*s && isalpha (*s))
		puttc (' ');
	while (*s)
		puttc (*s++);
}				/* addparens */

static void cleartabs()
{	static int must_alloc = TRUE;
	int i;
	if (tabid == -1) tabid = 0;
		else tabid++;
	if (tabid == MAXTB) {
		must_alloc = FALSE;
		tabid = 0;
	}
	if (must_alloc || tabline[tabid] == NULL) {
		tabline[tabid] = (unsigned char *)malloc(MAXLINE);
		if (tabline[tabid] == NULL) {
			windup();
			fprintf(stderr,"Couldn't get memory.\n");
			exit(1);
		}
	}
	for (i = 0; i < MAXLINE; i++) tabline[tabid][i] = '\0';
}

static void setdefaulttabs()
{	int i;
	for (i = 0; i < MAXLINE; i++) {
		if ((i+1) % tabsize == 0) tabline[tabid][i] = ROMAN;
		else tabline[tabid][i] = '\0';
	}
}

static void settab(int t)
{	int i;
	if (t < 0) t = 0;
	for (i = MAXLINE-1; i > 0; i--) if (tabline[tabid][i]) break;
	/*if (i > 0) i += 1;
	tabline[tabid][i+t-1] = '&';
	*/
	/* tabline[tabid][i+t] = '&'; */
	if (i+t < MAXLINE) tabline[tabid][i+t] = attribute;
	else {
		windup();
		fprintf(stderr,"illegal tab set -- exiting\n");
		exit(1);
	}
/*
fprintf(stderr,"\nTABSET %d from %d\n", i+t-1, t);
*/
}

static int findtab(int t, int tb)
{	int i = 0;
	if (!t) i = 1;
	while (t) {
		while (i < MAXLINE && !tabline[tb][i]) i++;
		if (i == MAXLINE) break;
		i++;
		t--;
	}
	if (i == MAXLINE) i = 0;
	return i;
}


static int findtabattribute(int t, int tb)
{	int i = 0;
	if (!t) i = 1;
	while (t) {
		while (i < MAXLINE && !tabline[tb][i]) i++;
		if (i == MAXLINE) break;
		i++;
		t--;
	}
	if (i == MAXLINE) return ROMAN;
	i = tabline[tb][i-1];
	if (!i) return ROMAN;
	return i;
}

static int num(char *s, int cansign)
{	int i;
	for (i = 0; s[i] && s[i] != '='; i++) ;
	if (s[i] == '=') s += i + 1;
	i = atoi(s);
	if (!cansign && i < 0) return 0;
	return i;
}

int dimen(char *s, int horizontal, int cansign, int wantpoints)
{	float num, pitch, vpitch;
	int i, scale;

	for (i = 0; s[i] && s[i] != '='; i++) ;
	if (s[i] == '=') s += i + 1;
	num = atof(s);
	for (i = 0; s[i] != 'i' && s[i] != 'p' &&
		 s[i] != 'e' && s[i] != 'c'; i++) ;
    /* picas? */
	if (s[i] == 'p' && s[i+1] == 'c') scale = 0;
    /* ems or exs? */
	else if (s[i] == 'e') scale = 1;
    /* inches? */
	else if (s[i] == 'i') scale = 2;
    /* centimeters? */
	else if (s[i] == 'c') scale = 3;
    /* points */
	else scale = 4;

	vpitch = 6.0;
	if (psflag || outdev) {
		pitch = 144.0 / ps_pointsize; /* assuming Courier */
		vpitch = 72.0 / ps_baselineskip;
	}
	else if (pointsize == 10) pitch = 14.4;
	else pitch = 12.0;

	if (horizontal) {
		if (scale == 0) num *= (pitch / 6.0);
		else if (scale == 1) num *= 2.0;
		else if (scale == 2) num *= pitch;
		else if (scale == 3) num *= pitch / 2.54;
		else num /= (72.0 / pitch);  /* 6.0 at 12 pitch */
	}
	else {
		if (scale == 0) num /= (12.0 / vpitch);
		else if (scale == 1) num /= 2.0;
		else if (scale == 2) num *= vpitch;
		else if (scale == 3) num *= vpitch / 2.54;
		else num /= (72.0 / vpitch);
	}
/* printf("%s gave %f = %d\n", s, num, (int)(num + 0.5));*/
	if (wantpoints) {
		if (horizontal) num *= (72.0 / pitch);
		else num *= (72.0 / vpitch);
	}
	if (!cansign && num < 0.5) return 0;
	if (num < 0) num -= 0.5;
	else num += 0.5;
	return (int)num;
}
int dimen_sp(char *s, int horizontal, int cansign)
{	int num;
	float fnum;
	int i, scale;

	for (i = 0; s[i] && s[i] != '='; i++) ;
	if (s[i] == '=') s += i + 1;
	fnum = atof(s);
	for (i = 0; s[i] != 'i' && s[i] != 'p' &&
		 s[i] != 'e' && s[i] != 'c'; i++) ;
    /* picas? */
	if (s[i] == 'p' && s[i+1] == 'c') scale = 0;
    /* ems or exs? */
	else if (s[i] == 'e') scale = 1;
    /* inches? */
	else if (s[i] == 'i') scale = 2;
    /* centimeters? */
	else if (s[i] == 'c') scale = 3;
    /* points */
	else scale = 4;

	if (horizontal) {
		if (scale == 0) num = (int)(fnum * ps_pitch_sp / 6.0);
		else if (scale == 1) num = (int)(fnum * 1000);
		else if (scale == 2) num = (int)(fnum * ps_pitch_sp);
		else if (scale == 3) num = (int)(fnum * ps_pitch_sp / 2.54);
		else num = (int)(fnum * ps_pitch_sp / 72.0);
	}
	else {
		if (scale == 0) num = (int)( fnum * ps_vpitch_sp / 6.0 );
		else if (scale == 1) num = (int)( fnum * 250.0 );
		else if (scale == 2) num = (int)( fnum * ps_vpitch_sp );
		else if (scale == 3) num = (int)( fnum * ps_vpitch_sp / 2.54 );
		else num = (int)( fnum  * ps_vpitch_sp / 72.0 );
	}
	if (!cansign && num < 0) return 0;
	return num;
}

static void putuni(int ch) {
	unsigned char c;
	unsigned int wc;
	int count;

        c = (unsigned char)ch;
	if (c < 0x80) {
		PUTTC(c);
		return;
	}
	if (!gflag && (!xflag || !is_ipa(attribute))) {
		/* this is a Latin-1 character */
		PUTTC(c);
		return;
	}
	if (!gflag) {
		wc = (unsigned int)ch;
	}
	else {
            if (c < 0xC0) return;

            if (c < 0xE0) {
              wc = (unsigned int)(c & 0x1F) << 6;
              count = 1;
              if (c < 0xC2) return;
            }
            else if (c < 0xF0) {
              wc = (unsigned int)(c & 0x0F) << 12;
              count = 2;
            }
            else if (c < 0xF8) {
              wc = (unsigned int)(c & 0x07) << 18;
              count = 3;
            }
	    else if (c < 0xFC) {
              wc = (unsigned int)(c & 0x03) << 24;
              count = 4;
            }
	    else if (c < 0xFE) {
              wc = (unsigned int)(c & 0x01) << 30;
              count = 5;
            }
            else return;

	    while (count) {

	      if ( (ch = input()) == EOF ) return;
              c = (unsigned char) ch ^ 0x80;
              if (!(c < 0x40)) return;
              wc |= (unsigned int) c << (6 * --count);

	    }
	}

	if (wc >= 0x100 && !gflag) return;

	if (wc < 0x100) {
	    if (is_ipa(attribute)) wc = let[2][wc];
	    else if (attribute == MATH) wc = let[1][wc];
	    else if (is_roman(attribute)) wc = let[0][wc];
	}

	stowuni(wc);
}

extern int uni_zd[256];

static void saveuni(unsigned int wc) {
	if (kflag) {
		puttc(UNI_ESC);
		puttc( wc & 0xff );
		puttc( (wc >> 8) & 0xff);
	}
}

static void stowuni(unsigned int wc) {
	int h, i = 0, u = -1, m = -1, ipac, diac, font, tfont;

	h = (int)wc;

	if (wc < 0x80 || (wc < 0x100 && !is_ipa(attribute))) {
		PUTTC(h);
		return;
	}

#if 0
	/* split some ipa chars into two */
	if (!psflag && is_ipa(attribute)) switch(wc) {
		case 163: stowuni(227); u = 0x0309; break; /* hooktop rt d = rt d with hook */
		case 139: stowuni('p'); u = 0x031b; break; /* left hook p = p with horn? */
		case 243: stowuni('r'); u = 0x0322; break; /* rt r = r with ret. hook */
		default: break;
	}
	if (u == -1)
#endif
	if (is_roman(attribute)) tfont = ROMAN;
	else if (is_ipa(attribute)) tfont = IPA;
	else tfont = attribute;

	for (i = 0; ; i++) {
		if ( (u=revtab[i][0]) == -1) break;
		font = revtab[i][1];
		if (h == u) m = i; 
		if (h == u && font == tfont) break; 
	}
	if (u == -1 && m != -1) {
		i = m;
		u = h;
	}

	if (u == -1) for (i = 0; i < 256 ; i++) {
		if (h == uni_zd[i]) {
			push(ZD);
			PUTTC(i);
			pop();
			return;
		}
	}
	if (u == -1) {
		push(ROMAN);
		if (wc < 0x100 && !kflag) {
			PUTTC(wc);
			pop();
			return;
		}
		if (psflag) switch(wc) {
			case 0x2013: u = PS_endash; break;
			case 0x2014: u = PS_emdash; break;
			case 0x201C: u = PS_quotedblleft; break;
			case 0x201D: u = PS_quotedblright; break;
			case 0x2026: u = PS_ellipsis; break;
			case 0x02DD: u = PS_hungarumlaut; break;
			case 0x02C6: u = PS_circumflex; break;
			case 0x02C7: u = PS_caron; break;
			case 0x02D8: u = PS_breve; break;
			case 0x02D9: u = PS_dotaccent; break;
			case 0x02DB: u = PS_ogonek; break;
			default: break;
		}
		if (u != -1) {
			PUTTC(u);
		}
/*
		else if (!psflag && wc >= 0x7f && wc < 0xa0 && !is_ipa(attribute)) {
			if (boxflag) attribute = BOX; else attribute += 128;
			PUTTC(glyph['h']);
		}
*/
		else if (wc < 0x10000 && kflag && !psflag) saveuni(wc);
		else {
			if (boxflag) attribute = BOX; else attribute += 128;
			PUTTC(glyph['h']);
		}
		pop();
		return;
	}

	font = revtab[i][1];
	ipac = revtab[i][2];
	diac = revtab[i][3];

	if (font == ROMAN && psflag) {
		if (tfont != ROMAN) push(ROMAN);
		PUTTC(ipac);
		if (diac >= 0) PUTTC(diac);
		if (tfont != ROMAN) pop();
		return;
	}
	/* This is for stranded diacritics. */
	if (tfont == ROMAN && diac == -1 && wc >= 0x300 && wc <= 0x348) {
		diac = ipac;
		ipac = -1;
	}
	/* Try to translate an IPA diacritic into a PS one. */
	if (tfont == ROMAN && diac != -1) {
		int rx = 1;
		if (psflag) rx = 0;
		if (font == IPAUNDER) rx += 2;
		if (!psflag || font != ROMAN)
		    switch(diac) {
			case 0: diac = roman_other['`'][rx]; break;
			case 1: diac = roman_other['\''][rx]; break;
			case 2: diac = roman_other['^'][rx]; break;
			case 3: diac = roman_other['~'][rx]; break;
			case 4: diac = roman_other['"'][rx]; break;
			case 5: diac = roman_other['H'][rx]; break;
			case 6: diac = roman_other['r'][rx]; break;
			case 7: diac = roman_other['v'][rx]; break;
			case 8: diac = roman_other['u'][rx]; break;
			case 9: diac = roman_other['='][rx]; break;
			case 10: diac = roman_other['.'][rx]; break;
			case 11: diac = roman_other['c'][rx]; break;
			case 12: diac = roman_other['k'][rx]; break;
			case 14: diac = roman_other['`'][rx+2]; break;
			default: rx = -1; break;
		    }
		if (rx != -1) {
			if (ipac >= 0) PUTTC(ipac);
			if (psflag) PUTTC(diac);
			else saveuni(diac);
			return;
		}
		if (!psflag) {
			saveuni(wc);
			return;
		}
	}
	if (ipac == -1) {
		ipac = diac;
		diac = -1;
	}


#ifdef STANDALONETREE

	if (tex_opt && diac >= 0 && diac <= 11 && isalnum(ipac)) {
		int j;
		PUTTC('\\');
		for (j = 0; pre[diac][j]; j++) PUTTC(pre[diac][j]);
	}
	if (tex_opt) {
		PUTTC(ipac);
		return;
	}

#endif

	if (diac == -1 && (font == IPAUNDER || font == IPAOVER)) ipa_attribute = font;
	if (font == IPAUNDER || font == IPAOVER) push(IPA);
	else if (font == BOX) {
		if (boxflag) push(BOX); else push(attribute + 128);
	}
	else push(font);

	if (font == BOX) {
		PUTTC(glyph[ipac]);
	}
	else PUTTC(ipac);

	if (diac != -1) {
		if (font == IPAUNDER || font == IPAOVER) ipa_attribute = font;
		PUTTC(diac);
	}
#ifndef FMTREE
#ifndef STANDALONETREE
	/* if this is a diacritic, go back and make the last char IPA */
	else if (is_ipa(font) && !ipa_width[ipac] && awp >= aword + 2) {
		*(awp - 2) = IPA;
		if (is_ctl(ipac) && awp >= aword + 3)
			*(awp - 3) = IPA;
	}
#endif
#endif
	pop();
}

static int man_symbol[][3] = {
{'d', 'q', 0x0022},
{'s', 'h', 0x0023},
{'D', 'o', 0x0024},
{'a', 'q', 0x0027},
{'p', 'l', 0x002B},
{'s', 'l', 0x002F},
{'e', 'q', 0x003D},
{'a', 't', 0x0040},
{'l', 'B', 0x005B},
{'r', 's', 0x005C},
{'r', 'B', 0x005D},
{'a', '^', 0x005E},
{'h', 'a', 0x005E},
{'r', 'u', 0x005F},
{'u', 'l', 0x005F},
{'\\', '`', 0x0060},
{'g', 'a', 0x0060},
{'l', 'C', 0x007B},
{'b', 'a', 0x007C},
{'o', 'r', 0x007C},
{'r', 'C', 0x007D},
{'a', '~', 0x007E},
{'t', 'i', 0x007E},
{'r', '!', 0x00A1},
{'c', 't', 0x00A2},
{'P', 'o', 0x00A3},
{'C', 's', 0x00A4},
{'Y', 'e', 0x00A5},
{'b', 'b', 0x00A6},
{'s', 'c', 0x00A7},
{'a', 'd', 0x00A8},
{'c', 'o', 0x00A9},
{'O', 'f', 0x00AA},
{'F', 'o', 0x00AB},
{'n', 'o', 0x00AC},
{'r', 'g', 0x00AE},
{'a', '-', 0x00AF},
{'d', 'e', 0x00B0},
{'+', '-', 0x00B1},
{'S', '2', 0x00B2},
{'S', '3', 0x00B3},
{'a', 'a', 0x00B4},
{'p', 's', 0x00B6},
{'p', 'c', 0x00B7},
{'a', 'c', 0x00B8},
{'S', '1', 0x00B9},
{'O', 'm', 0x00BA},
{'F', 'c', 0x00BB},
{'1', '4', 0x00BC},
{'1', '2', 0x00BD},
{'3', '4', 0x00BE},
{'r', '?', 0x00BF},
{'`', 'A', 0x00C0},
{'\'', 'A', 0x00C1},
{'^', 'A', 0x00C2},
{'~', 'A', 0x00C3},
{':', 'A', 0x00C4},
{'o', 'A', 0x00C5},
{'A', 'E', 0x00C6},
{'`', 'E', 0x00C8},
{'\'', 'E', 0x00C9},
{'^', 'E', 0x00CA},
{':', 'E', 0x00CB},
{'`', 'I', 0x00CC},
{'\'', 'I', 0x00CD},
{'^', 'I', 0x00CE},
{':', 'I', 0x00CF},
{'-', 'D', 0x00D0},
{'~', 'N', 0x00D1},
{'`', 'O', 0x00D2},
{'\'', 'O', 0x00D3},
{'^', 'O', 0x00D4},
{'~', 'O', 0x00D5},
{':', 'O', 0x00D6},
{'m', 'u', 0x00D7},
{'/', 'O', 0x00D8},
{'`', 'U', 0x00D9},
{'\'', 'U', 0x00DA},
{'^', 'U', 0x00DB},
{':', 'U', 0x00DC},
{'\'', 'Y', 0x00DD},
{'T', 'P', 0x00DE},
{'s', 's', 0x00DF},
{'`', 'a', 0x00E0},
{'\'', 'a', 0x00E1},
{'^', 'a', 0x00E2},
{'~', 'a', 0x00E3},
{':', 'a', 0x00E4},
{'o', 'a', 0x00E5},
{'a', 'e', 0x00E6},
{'`', 'e', 0x00E8},
{'\'', 'e', 0x00E9},
{'^', 'e', 0x00EA},
{':', 'e', 0x00EB},
{'`', 'i', 0x00EC},
{'\'', 'i', 0x00ED},
{'^', 'i', 0x00EE},
{':', 'i', 0x00EF},
{'S', 'd', 0x00F0},
{'~', 'n', 0x00F1},
{'`', 'o', 0x00F2},
{'\'', 'o', 0x00F3},
{'^', 'o', 0x00F4},
{'~', 'o', 0x00F5},
{':', 'o', 0x00F6},
{'d', 'i', 0x00F7},
{'/', 'o', 0x00F8},
{'`', 'u', 0x00F9},
{'\'', 'u', 0x00FA},
{'^', 'u', 0x00FB},
{':', 'u', 0x00FC},
{'\'', 'y', 0x00FD},
{'T', 'p', 0x00FE},
{':', 'y', 0x00FF},
{'.', 'i', 0x0131},
{'/', 'L', 0x0141},
{'/', 'l', 0x0142},
{'O', 'E', 0x0152},
{'o', 'e', 0x0153},
{'F', 'n', 0x0192},
{'a', 'h', 0x02C7},
{'a', 'b', 0x02D8},
{'a', '.', 0x02D9},
{'a', 'o', 0x02DA},
{'h', 'o', 0x02DB},
{'a', '"', 0x02DD},
{'*', 'A', 0x0391},
{'*', 'B', 0x0392},
{'*', 'G', 0x0393},
{'*', 'D', 0x0394},
{'*', 'E', 0x0395},
{'*', 'Z', 0x0396},
{'*', 'Y', 0x0397},
{'*', 'H', 0x0398},
{'*', 'I', 0x0399},
{'*', 'K', 0x039A},
{'*', 'L', 0x039B},
{'*', 'M', 0x039C},
{'*', 'N', 0x039D},
{'*', 'C', 0x039E},
{'*', 'O', 0x039F},
{'*', 'P', 0x03A0},
{'*', 'R', 0x03A1},
{'*', 'S', 0x03A3},
{'*', 'T', 0x03A4},
{'*', 'U', 0x03A5},
{'*', 'F', 0x03A6},
{'*', 'X', 0x03A7},
{'*', 'Q', 0x03A8},
{'*', 'W', 0x03A9},
{'*', 'a', 0x03B1},
{'*', 'b', 0x03B2},
{'*', 'g', 0x03B3},
{'*', 'd', 0x03B4},
{'*', 'e', 0x03B5},
{'*', 'z', 0x03B6},
{'*', 'y', 0x03B7},
{'*', 'h', 0x03B8},
{'*', 'i', 0x03B9},
{'*', 'k', 0x03BA},
{'*', 'l', 0x03BB},
{'*', 'm', 0x03BC},
{'*', 'n', 0x03BD},
{'*', 'c', 0x03BE},
{'*', 'o', 0x03BF},
{'*', 'p', 0x03C0},
{'*', 'r', 0x03C1},
{'t', 's', 0x03C2},
{'*', 's', 0x03C3},
{'*', 't', 0x03C4},
{'*', 'u', 0x03C5},
{'*', 'f', 0x03C6},
{'*', 'x', 0x03C7},
{'*', 'q', 0x03C8},
{'*', 'w', 0x03C9},
{'+', 'h', 0x03D1},
{'+', 'f', 0x03D5},
{'+', 'p', 0x03D6},
{'h', 'y', 0x2010},
{'e', 'n', 0x2013},
{'e', 'm', 0x2014},
{'o', 'q', 0x2018},
{'c', 'q', 0x2019},
{'b', 'q', 0x201A},
{'l', 'q', 0x201C},
{'r', 'q', 0x201D},
{'B', 'q', 0x201E},
{'d', 'g', 0x2020},
{'d', 'd', 0x2021},
{'b', 'u', 0x2022},
{'%', '0', 0x2030},
{'f', 'm', 0x2032},
{'s', 'd', 0x2033},
{'f', 'o', 0x2039},
{'f', 'c', 0x203A},
{'r', 'n', 0x203E},
{'f', '/', 0x2044},
{'I', 'm', 0x2111},
{'w', 'p', 0x2118},
{'R', 'e', 0x211C},
{'t', 'm', 0x2122},
{'A', 'h', 0x2135},
{'<', '-', 0x2190},
{'u', 'a', 0x2191},
{'-', '>', 0x2192},
{'d', 'a', 0x2193},
{'<', '>', 0x2194},
{'l', 'A', 0x21D0},
{'u', 'A', 0x21D1},
{'r', 'A', 0x21D2},
{'d', 'A', 0x21D3},
{'h', 'A', 0x21D4},
{'f', 'a', 0x2200},
{'p', 'd', 0x2202},
{'t', 'e', 0x2203},
{'e', 's', 0x2205},
{'g', 'r', 0x2207},
{'m', 'o', 0x2208},
{'n', 'm', 0x2209},
{'s', 't', 0x220B},
{'\\', '-', 0x2212},
{'m', 'i', 0x2212},
{'*', '*', 0x2217},
{'s', 'r', 0x221A},
{'p', 't', 0x221D},
{'i', 'f', 0x221E},
{'/', '_', 0x2220},
{'A', 'N', 0x2227},
{'O', 'R', 0x2228},
{'c', 'a', 0x2229},
{'c', 'u', 0x222A},
{'i', 's', 0x222B},
{'t', 'f', 0x2234},
{'t', 'i', 0x223C},
{'a', 'p', 0x223C},
{'=', '~', 0x2245},
{'~', '~', 0x2248},
{'~', '=', 0x2248},
{'!', '=', 0x2260},
{'=', '=', 0x2261},
{'<', '=', 0x2264},
{'>', '=', 0x2265},
{'s', 'b', 0x2282},
{'s', 'p', 0x2283},
{'i', 'b', 0x2286},
{'i', 'p', 0x2287},
{'c', '+', 0x2295},
{'c', '*', 0x2297},
{'p', 'p', 0x22A5},
{'p', 'c', 0x22C5},
{'l', 'c', 0x2308},
{'r', 'c', 0x2309},
{'l', 'f', 0x230A},
{'r', 'f', 0x230B},
{'l', 'a', 0x2329},
{'r', 'a', 0x232A},
{'C', 'R', 0x240D},
{'a', 'n', 0x2500},
{'b', 'r', 0x2502},
{'b', 'v', 0x2502},
{'r', 'k', 0x251D},
{'l', 'k', 0x2525},
{'l', 't', 0x256D},
{'r', 't', 0x256E},
{'r', 'b', 0x256F},
{'l', 'b', 0x2570},
{'s', 'q', 0x25A1},
{'l', 'z', 0x25CA},
{'c', 'i', 0x25EF},
{'l', 'h', 0x261C},
{'r', 'h', 0x261E},
{'f', 'f', 0xFB00},
{'f', 'i', 0xFB01},
{'f', 'l', 0xFB02},
{'F', 'i', 0xFB03},
{'F', 'l', 0xFB04},
{0, 0, 0}
};

static void mansym(int k1, int k2) {
	int i, u;
	for (i = 0; man_symbol[i][0] && (man_symbol[i][0] != k1 ||
		man_symbol[i][1] != k2); i++) ;
	if (man_symbol[i][0]) {
		u = man_symbol[i][2];
		if (psflag) stowuni(u);
		else saveuni(u);
	}
}
