/*
 * unicode.c
 * output unicode utf-8
 * Greg Lee, 10/01
 */
#include <stdio.h>
#include "fm.h"
#include "revtab.h"

extern int ipa_width[256];

static int ipa_uni[256] = {
0x0300,	/* "*`" C 0	accent grave	  : COMBINING GRAVE ACCENT */
0x0301,	/* "*'" C 1	accent acute	  : COMBINING ACUTE ACCENT */
0x0302,	/* "*^" C 2	accent circumflex  : COMBINING CIRCUMFLEX ACCENT */
0x0303,	/* "*~" C 3	accent tilde	  : COMBINING TILDE */
0x0308,	/* "*"" C 4	accent umlaut	  : COMBINING DIAERESIS */
0x030B,	/* C 5	accent double acute  : COMBINING DOUBLE ACUTE ACCENT */
0x030A,	/* C 6	accent ring	  : COMBINING RING ABOVE */
0x030C,	/* C 7	accent wedge	  : COMBINING CARON */
0x0306,	/* C 8	accent breve	  : COMBINING BREVE */
0x0304,	/* "*=" C 9	accent macron	  : COMBINING MACRON */
0x0307,	/* "*." C 10	accent dot	  : COMBINING DOT ABOVE */
0x0327,	/* C 11	accent cedilla  : COMBINING CEDILLA */
0x0328,	/* C 12	accent ogonek  : COMBINING OGONEK */
0x030F,	/* C 13	accent double grave		  : COMBINING DOUBLE GRAVE ACCENT */
0x0316,	/* C 14	accent subscript grave  : COMBINING GRAVE ACCENT BELOW */
0x0317,	/* C 15	accent subscript acute  : COMBINING ACUTE ACCENT BELOW */
0x0311,	/* C 16	accent round cap / subscript arch  : COMBINING INVERTED BREVE BELOW cf. 0x32F COMBINING INVERTED BREVE */
0x0346,	/* C 17	accent subscript bridge	 : COMBINING BRIDGE BELOW  cf. 0x32A COMBINING BRIDGE ABOVE */
0x033A,	/* C 18	accent inverted subscript bridge  : COMBINING INVERTED BRIDGE BELOW */
0x033B,	/* C 19	accent subscript square		  : COMBINING SQUARE BELOW */
0x0339,	/* C 20	accent subscript right half-ring  : COMBINING RIGHT HALF RING BELOW */
0x031C,	/* C 21	accent subscript left half-ring  : COMBINING LEFT HALF RING BELOW */
0x02B7,	/* (modifier small w) C 22	accent subscript w		  */
0x033C,	/* C 23	accent seagull  : COMBINING SEAGULL BELOW */
0x033D,	/* C 24	accent over-cross		  : COMBINING X ABOVE */
0x0131,	/* 'i.' ı C 25	dotless i  : SMALL LETTER DOTLESS I */
0x025F,	/* (dotless with stoke ) C 26	dotless j  */
0x031F,	/* C 27	accent subscript plus		  : COMBINING PLUS SIGN BELOW */
0x031D,	/* C 28	accent raising sign		  : COMBINING UP TACK BELOW */
0x031E,	/* C 29	accent lowering sign		  : COMBINING DOWN TACK BELOW */
0x0319,	/* C 30	accent advancing sign  : COMBINING RIGHT TACK BELOW */
0x0318,	/* C 31	accent retracting sign  : COMBINING LEFT TACK BELOW */
32,	/* C 32 space (changed) */
0x01C3,	/* C 33	exclamation mark  : LATIN LETTER RETROFLEX CLICK; EXCLAMATION MARK */
0x0329,	/* C 34	syllabicity mark  : COMBINING VERTICAL LINE BELOW */
0x0309,	/* C 35	accent hooktop  : COMBINING HOOK ABOVE */
0x0322,	/* C 36	accent right hook  : COMBINING RETROFLEX HOOK BELOW */
0x0321,	/* C 37	accent palatalization hook  : COMBINING PALATALIZED HOOK BELOW */
0x0334,	/* C 38	accent superimposed tilde  : COMBINING TILDE OVERLAY */
0x02BC,	/* C 39	apostrophe (ejective)  : MODIFIER LETTER APOSTROPHE */
'(',	/* C 40	left paren  */
')',	/* C 41	right paren  */
'*',	/* C 42	asterisk (raised)  */
'+',	/* C 43	plus  */
',',	/* C 44	comma  */
'-',	/* C 45	hyphen  */
'.',	/* C 46	period (syllable break)  */
'/',	/* C 47	slash  */
0x0289,	/* "u/" "%0" ʉ C 48	barred u: SMALL LETTER U BAR */
0x0268,	/* "i/" "%1" ɨ C 49	barred i: SMALL LETTER I WITH STROKE */
0x028C,	/* "`v" "%2" ʌ C 50	turned v: SMALL LETTER TURNED V */
0x025C,	/* "'E" "%3" ɜ C 51	reversed epsilon: SMALL LETTER REVERSED OPEN E */
0x0265,	/* "`h" "%4" ɥ C 52	turned h: SMALL LETTER TURNED H */
0x0250,	/* "`a" "%5" ɐ C 53	turned a: SMALL LETTER TURNED A */
0x0252,	/* "`A" "%6" ɒ C 54	turned script a: SMALL LETTER TURNED ALPHA ... SCRIPT A */
0x0264,	/* "rh" "%7" ɤ C 55	rams horns: SMALL LETTER RAMS HORN */
0x0275,	/* "o/" "%8" ɵ C 56	barred o: SMALL LETTER BARRED O */
0x0258,	/* "'e" "%9" ɘ C 57	reversed e: SMALL LETTER REVERSED E */
0x02D0,	/* "::" ː C 58	length mark: MODIFIER LETTER TRIANGULAR COLON */
0x02D1,	/* ":," ∵ C 59	half-length mark: MODIFIER LETTER HALF TRIANGULAR COLON */
0x02CD,	/* (modifier extra low tone bar) C 60	accent ??  bottom tie bar */
'=',	/* C 61	ascii '='  */
0x02C9,	/* (modifier extra high tone bar) C 62	accent top tie bar  */
'?',	/* C 63	question mark  */
0x0259,	/* "%@" ə C 64	@ = schwa: SMALL LETTER SCHWA */
0x0251,	/* "%A" ɑ C 65	A = script a: LATIN SMALL LETTER ALPHA; SCRIPT A */
0x03B2,	/* "%B" β C 66	B = beta: GREEK SMALL LETTER BETA */
0x0255,	/* "%C" ɕ C 67	C = curly-tail c: SMALL LETTER C WITH CURL */
0x00F0,	/* "%D" ð C 68	D = eth, [\dh]: SMALL LETTER ETH */
0x025B,	/* "%E" ɛ C 69	E = epsilon: SMALL LETTER OPEN E */
0x0278,	/* "%F" ɸ C 70	F = phi: SMALL LETTER PHI */
0x0263,	/* "%G" ɣ C 71	G = gamma: SMALL LETTER GAMMA */
0x0266,	/* "%H" "h2" ɦ ɦ C 72	H =hooktop h: SMALL LETTER H WITH HOOK */
0x026A,	/* "%I" "iI" ɪ ɪ C 73	I =small capital i: LETTER SMALL CAPITAL I */
'j',	/* C 74	J = curly-tail j  */
0x0281,	/* "%K" ʁ C 75	K = inverted small capital r: LETTER SMALL CAPITAL INVERTED R */
0x028E,	/* "%L" "`y" ʎ ʎ C 76	L =turned y: SMALL LETTER TURNED Y */
0x0271,	/* "%M" ɱ C 77	M =left-tail m: LATIN SMALL LETTER M WITH HOOK */
0x014B,	/* "%N" ŋ C 78	N =eng, [\ng]: SMALL LETTER ENG */
0x0254,	/* "%O" ɔ C 79	O =open o: SMALL LETTER OPEN O */
0x0294,	/* "%P" ʔ C 80	P =glottal stop: LETTER GLOTTAL STOP */
0x0295,	/* "%Q" C 81	Q =reversed glottal stop: LATIN LETTER PHARYNGEAL VOICED FRICATIVE; REVERSED GLOTTAL STOP */
0x027E,	/* "%R" C 82	R =fish-hook r: SMALL LETTER R WITH FISHHOOK */
0x0283,	/* "%S" C 83	S =esh: SMALL LETTER ESH */
0x03B8,	/* "%T" C 84	T =theta: GREEK SMALL LETTER THETA */
0x028A,	/* "%U" C 85	U =upsilon: LATIN SMALL LETTER UPSILON */
0x028B,	/* "%V" C 86	V =script v: LATIN SMALL LETTER V WITH HOOK; SCRIPT V */
0x026F,	/* "%W" "`m" C 87	W =turned m: SMALL LETTER TURNED M */
0x03C7,	/* "%X" C 88	X = chi: GREEK SMALL LETTER CHI */
0x028F,	/* "%Y" "yY" C 89	Y =small capital y: LETTER SMALL CAPITAL Y */
0x021D,	/* "%Z" C 90	Z = yogh: SMALL LETTER EZH, cf. 0x0292 SMALL LETTER YOGH */
'[',	/* C 91	left bracket  */
0x02BD,	/* C 92	reversed apostrophe (aspirated)  */
']',	/* C 93	right bracket  */
0x02ED,	/* C 94	corner  : MODIFIER LETTER UNASPIRATED */
0x02D6,	/* (modifier +?) C 95	open corner  */
0x2018,	/* C 96	turned apostrophe  */
'a',	/* C 97	a  */
'b',	/* C 98	b  */
'c',	/* C 99	c  */
'd',	/* C 100	d  */
'e',	/* C 101	e  */
'f',	/* C 102	f  */
0x0261,	/* "%g" C 103	script g  : SMALL LETTER SCRIPT G */
'h',	/* C 104	h  */
'i',	/* C 105	i  */
'j',	/* C 106	j  */
'k',	/* C 107	k  */
'l',	/* C 108	l  */
'm',	/* C 109	m  */
'n',	/* C 110	n  */
'o',	/* C 111	o  */
'p',	/* C 112	p  */
'q',	/* C 113	q  */
'r',	/* C 114	r  */
's',	/* C 115	s  */
't',	/* C 116	t  */
'u',	/* C 117	u  */
'v',	/* C 118	v  */
'w',	/* C 119	w  */
'x',	/* C 120	x  */
'y',	/* C 121	y  */
'z',	/* C 122	z  */
0x01C1,	/* "%{" ǁ C 123	double pipe: LATIN LETTER LATERAL CLICK; DOUBLE PIPE */
0x01C0,	/* C 124 | =pipe: LATIN LETTER DENTAL CLICK; PIPE */
0x01C2,	/* "%}" ǂ C 125	=double-barred pipe: LATIN LETTER ALVEOLAR CLICK; PIPE DOUBLE BAR */
0x02B5,	/* "r2" ʵ C 126	=rhoticity: MODIFIER LETTER SMALL TURNED R WITH HOOK */
0x2379,	/* (APL omega underbar) C 127	=inverted omega */
0x02C8,	/* "%'" ˈ C 128 =primary stress, CHANGED: MODIFIER LETTER VERTICAL LINE */
0x02CC,	/* "%," ˌ C 129 secondary stress, CHANGED: MODIFIER LETTER LOW VERTICAL LINE */
0x02B9,	/* "cp" ʹ C 130  =celtic palatalization mark: MODIFIER LETTER PRIME */

0,	/* C 131 rthooklong */
0x0321,	/* (=37) C 132 palhooklong */
0x0321,	/* (=37) C 133 palhookvar */
0,	/* C 134 retractingvar */
0,	/* C 135 reversed Polish hook (?? htrtaild, Hook-top Right-Tail D) */
0,	/* C 136 rtailhth, Right-Tail Hooktop H */
0,	/* C 137 heng */
0,	/* C 138 hmlig, H_M Ligature */
0,	/* C 139 lhookp */
0x01AA,	/* C 140 looptoprevesh */
0,	/* C 141 invscripta */
0,	/* C 142 ctinvglotstop, Curly-Tail Inverted Glottal Stop */
0x027F,	/* C 143 lhti, left-hooktop i  : latin small letter reversed r with fishhook */
0x019E,	/* C 144 nrleg, n, right leg, Long-Leg N  */
0,	/* C 145 htbardotlessjvar, Hooktop Barred Dotless J (a variety) */

0x007c,	/* C 146	vertical line  */
0x2016,	/* C 147	double vertical line  */
0x2193,	/* '-v' ↓ C 148	down arrow (downstep)  : DOWNWARDS ARROW */
0x2191,	/* '-!' ↑ C 149	up arrow (upstep)  : UPWARDS ARROW */
0x2197,	/* C 150	upward diagonal arrow  : NORTH EAST ARROW */
0x2198,	/* C 151	downward diagonal arrow  : SOUTH EAST ARROW */
0,	/* C 152	accent macron acute ligature  */
0,	/* C 153	accent grave macron ligature  */
0,	/* C 154	accent grave wedge ligature  */
0,	/* C 155	accent circumflex acute ligature  */
0x030D,	/* C 156	accent vertical bar  : COMBINING VERTICAL LINE ABOVE */
0x030E,	/* (0x0348 is below) C 157 accent double vertical bar: COMBINING DOUBLE VERTICAL LINE ABOVE */
0,	/* C 158	accent grave dot  */
0,	/* C 159	accent dot acute  */
0x0180,	/* "b/" ƀ C 160	crossed b: SMALL LETTER B WITH STROKE */
0x0111,	/* 'd/' đ C 161	crossed d  */
0,	/* C 162	curly-tail d  */
0,	/* C 163	htrtaild	  */
'E',	/* C 164	small capital e	  */
'g',	/* C 165	text g  */
/* 0x027f,	 C 165 SEE C 143 ABOVE @D left-hooktop i  : latin small letter reversed r with fishhook */
0,	/* C 166	left-hooktop long i  */
0,	/* C 167	viby i  */
'J',	/* C 168	small capital j	  */
0x029E,	/* "`k" ʞ C 169	turned k: SMALL LETTER TURNED K */
0x019A,	/* 'l/' ł C 170	barred l: SMALL LETTER L WITH STROKE cf. 0x0142 SMALL LETTER L WITH BAR */
0x03BB,	/* 'l*' λ C 171	lambda: GREEK SMALL LETTER LAMDA */
0x0196,	/* C 172	@A crlambda Crossed Lambda  */
0x026E,	/* (same as new) C 173	old l-yogh ligature  */
0,	/* C 174	curly-tail n  */
/* 0x019E,	 C 174 SEE C 144 ABOVE, @D n, right leg  */
0,	/* C 175	turned c-e ligature  */
0x03C9,	/* C 176	omega  : GREEK SMALL LETTER OMEGA */
0x03A9,	/* (Greek cap omega) C 177	small capital omega  */
/* 0,	GONE C 177	small capital q	  */
0x0286,	/* ",S" C 178	curly-tail esh: SMALL LETTER ESH WITH CURL */
0x01AB,	/* ",t" C 179	left-hook t (palatalized t) */
0,	/* C 180	curly-tail t */
0x02A6,	/* "ts" ʦ C 181	t-s ligature: SMALL LETTER TS DIGRAPH */
0x1EF7,	/* C 182 @A htlongy Left-hooktop Long Y (Does 1EF7 have hook at left?) */
0,	/* C 183	viby y  */
0x01BA,	/* C 184	curly-tail yogh  : SMALL LETTER EZH WITH TAIL */
0x01B9,	/* C 185	reversed yogh  : SMALL LETTER EZH REVERSED */
0x044C,	/* '%\' C 186	soft sign  : CYRILLIC SMALL LETTER SOFT SIGN */
0x044A,	/* '=\' C 187	hard sign  : CYRILLIC SMALL LETTER HARD SIGN */
0x02C0,	/* C 188	raised glottal stop  : MODIFIER LETTER GLOTTAL STOP */
0x02C2,	/* C 189	left pointer  : MODIFIER LETTER LEFT ARROWHEAD */
0x02C3,	/* C 190	right pointer  : MODIFIER LETTER RIGHT ARROWHEAD */
0x0362,	/* C 191	combining double (should be single) right arrow below  */
'A',	/* C 192	small capital a	  */
0x0188,	/* "c2" ƈ C 193	hooktop c: SMALL LETTER C WITH HOOK */
0x0297,	/* "%c" ʗ C 194	stretched c: LETTER STRETCHED C */
0x02A4,	/* "dZ" ʤ C 195	d-yogh ligature: SMALL LETTER DEZH DIGRAPH */
0x025A,	/* "@2" ɚ C 196	right-hook schwa: SMALL LETTER SCHWA WITH HOOK */
0x029A,	/* C 197	closed epsilon: SMALL LETTER CLOSED OPEN E */
0x025E,	/* C 198	closed reversed epsilon: SMALL LETTER CLOSED REVERSED OPEN E */
0x025D,	/* "e2" ɝ C 199	right-hook reversed epsilon: SMALL LETTER REVERSED OPEN E WITH HOOK */
0x02E0,	/* C 200	baby gamma: MODIFIER LETTER SMALL GAMMA */
0x029B,	/* "G2" ʛ C 201	hooktop small capital g: LETTER SMALL CAPITAL G WITH HOOK */
0x0267,	/* "H2" ɧ C 202	hooktop heng: SMALL LETTER HENG WITH HOOK */
0x029C,	/* "hH" ʜ C 203	small capital h: LETTER SMALL CAPITAL H */
0x0269,	/* ι C 204	iota: SMALL LETTER IOTA */
0x029D,	/* C 205	old barred dotless j: LATIN SMALL LETTER J WITH CROSSED-TAIL */
0x0199,	/* "k2" ƙ C 206	hooktop k: SMALL LETTER K WITH HOOK */
0x029F,	/* "lL" ʟ C 207	small capital l: LETTER SMALL CAPITAL L */
0x026E,	/* "lZ" C 208	l-yogh ligature: SMALL LETTER LEZH */
0x0277,	/* C 209	closed omega: SMALL LETTER CLOSED OMEGA */
0x01A5,	/* "p2" ƥ C 210	hooktop p: SMALL LETTER P WITH HOOK */
0x02A0,	/* "q2" ʠ C 211	hooktop q: SMALL LETTER Q WITH HOOK */
0x027C,	/* C 212	long-leg r: SMALL LETTER R WITH LONG LEG */
0x027A,	/* C 213	turned long-leg r: SMALL LETTER TURNED R WITH LONG LEG */
0x01AD,	/* "t2" ƭ C 214	hooktop t: SMALL LETTER T WITH HOOK */
0x0276,	/* C 215	small capital o-e ligature: LETTER SMALL CAPITAL OE */
0x0287,	/* "`t" ʇ C 216	turned t: SMALL LETTER TURNED T */
0x02A7,	/* "tS" ʧ C 217	t-esh ligature: SMALL LETTER TESH DIGRAPH */
'U',	/* C 218	small capital u	*/
0x0296,	/* "`?" ʖ C 219	inverted glottal stop: LETTER INVERTED GLOTTAL STOP */
0x02A1,	/* "?/" ʡ C 220	barred glottal stop: LETTER GLOTTAL STOP WITH STROKE */
0x02A2,	/* "/?" ʢ C 221	barred reversed glottal stop: LETTER REVERSED GLOTTAL STOP WITH STROKE */
0,	/* C 222	comma-tail z	*/
0x01BF,	/* C 223	wynn: LETTER WYNN */
0x0299,	/* "bB" ʙ C 224	small capital b: LETTER SMALL CAPITAL B */
0x0253,	/* "b2" ɓ C 225	hooktop b: SMALL LETTER B WITH HOOK */
0x0257,	/* "d2" ɗ C 226	hooktop d	  : SMALL LETTER D WITH HOOK */
0x0256,	/* ">d" ɖ C 227	right-tail d	  : LATIN SMALL LETTER D WITH TAIL; RETROFLEX HOOK */
0x0260,	/* "g2" ɠ C 228	hooktop g	  : SMALL LETTER G WITH HOOK */
0x0262,	/* "gG" ɢ C 229	small capital g	  : LETTER SMALL CAPITAL G */
0x00E6,	/* æ C 230	ash		  : SMALL LETTER AE */
0x00E7,	/* C 231	c cedilla: SMALL LETTER C WITH CEDILLA */
0x0127,	/* 'h/' ħ C 232	crossed h	  : SMALL LETTER H WITH STROKE */
0x025F,	/* "j/" ɟ C 233	barred dotless j: SMALL LETTER DOTLESS J WITH STROKE */
0x0284,	/* "j2" ʄ C 234	hooktop barred dotless j: SMALL LETTER DOTLESS J WITH STROKE AND HOOK */
0x026B,	/* "l~" ɫ C 235	l with tilde: SMALL LETTER L WITH MIDDLE TILDE */
0x026C,	/* "%l" ɬ C 236	belted l	  : SMALL LETTER L WITH BELT */
0x026D,	/* ">l" ɭ C 237	right-tail l	  : SMALL LETTER L WITH RETROFLEX HOOK */
0x0270,	/* C 238	turned m - right leg: SMALL LETTER TURNED M WITH LONG LEG */
0x0273,	/* ">n" ɳ C 239	right-tail n	  : LATIN SMALL LETTER N WITH RETROFLEX HOOK */
0x0274,	/* "nN" ɴ C 240	small capital n: LETTER SMALL CAPITAL N */
0x0272,	/* ",n" ɳ C 241	left-tail n - at left: LATIN SMALL LETTER N WITH LEFT HOOK */
0x0298,	/* C 242	bulls eye	  : LATIN LETTER BILABIAL CLICK; BULLSEYE */
0,	/* C 243	right-tail r	  */
0x0279,	/* "`r" ɹ C 244	turned r: SMALL LETTER TURNED R */
0x027B,	/* C 245	turned r, right tail  : LATIN SMALL LETTER TURNED R WITH HOOK */
0x0280,	/* "rR" ʀ C 246	small capital r: LETTER SMALL CAPITAL R */
0x0153,	/* C 247	o-e ligature, \oe  : SMALL LIGATURE OE */
0x00F8,	/* "o/" ø C 248	slashed o, \o  : SMALL LETTER O WITH STROKE */
0x0282,	/* ">s" ʂ C 249	right-tail s - at left - retroflex: LATIN SMALL LETTER S WITH HOOK */
0x0288,	/* ">t" ʈ C 250	right-tail t: SMALL LETTER T WITH RETROFLEX HOOK */
0x028D,	/* "`w" ʍ C 251	turned w: SMALL LETTER TURNED W */
0x0290,	/* ">z" ʐ C 252	right-tail z: LATIN SMALL LETTER Z WITH RETROFLEX HOOK */
0x0291,	/* ",z" ʑ C 253	curly-tail z: SMALL LETTER Z WITH CURL */
0x00FE,	/* C 254	thorn, [\th]  : SMALL LETTER THORN */
0x0195		/* C 255	h-v ligature: SMALL LETTER HV */
};

static int accents[][27] = {
/* 0 = GRAVE */
{
0x00E0, /* SMALL LETTER A WITH GRAVE */
-1,-1,-1, /* b c d */
0x00E8, /* SMALL LETTER E WITH GRAVE */
-1,-1,-1, /* f g h */
0x00EC, /* SMALL LETTER I WITH GRAVE */
-1,-1,-1,-1, /* j k l m */
0x01F9, /* SMALL LETTER N WITH GRAVE */
0x00F2, /* SMALL LETTER O WITH GRAVE */
-1,-1,-1,-1,-1, /* p q r s t */
0x00F9, /* SMALL LETTER U WITH GRAVE */
-1, /* v */
0x1E81, /* SMALL LETTER W WITH GRAVE */
-1, /* x */
0x1EF3, /* SMALL LETTER Y WITH GRAVE */
-1, /* z */
0 /* OVER */
},{
/* 1 = ACUTE */
0x00E1, /* SMALL LETTER A WITH ACUTE */
-1, /* b */
0x0107, /* SMALL LETTER C WITH ACUTE */
-1, /* d */
0x00E9, /* SMALL LETTER E WITH ACUTE */
-1, /* f */
0x01F5, /* SMALL LETTER G WITH ACUTE */
-1, /* h */
0x00ED, /* SMALL LETTER I WITH ACUTE */
-1, /* j */
0x1E31, /* SMALL LETTER K WITH ACUTE */
0x013A, /* SMALL LETTER L WITH ACUTE */
0x1E3F, /* SMALL LETTER M WITH ACUTE */
0x0144, /* SMALL LETTER N WITH ACUTE */
0x00F3, /* SMALL LETTER O WITH ACUTE */
0x1E55, /* SMALL LETTER P WITH ACUTE */
-1, /* q */
0x0155, /* SMALL LETTER R WITH ACUTE */
0x015B, /* SMALL LETTER S WITH ACUTE */
-1, /* t */
0x00FA, /* SMALL LETTER U WITH ACUTE */
-1, /* v */
0x1E83, /* SMALL LETTER W WITH ACUTE */
-1, /* x */
0x00FD, /* SMALL LETTER Y WITH ACUTE */
0x017A, /* SMALL LETTER Z WITH ACUTE */
0 /* OVER */
},{
/* 2 = CIRCUMFLEX */
0x00E2, /* SMALL LETTER A WITH CIRCUMFLEX */
-1, /* b */
0x0109, /* SMALL LETTER C WITH CIRCUMFLEX */
-1, /* d */
0x00EA, /* SMALL LETTER E WITH CIRCUMFLEX */
-1, /* f */
0x011D, /* SMALL LETTER G WITH CIRCUMFLEX */
0x0125, /* SMALL LETTER H WITH CIRCUMFLEX */
0x00EE, /* SMALL LETTER I WITH CIRCUMFLEX */
0x0135, /* SMALL LETTER J WITH CIRCUMFLEX */
-1,-1,-1,-1, /* k l m n */
0x00F4, /* SMALL LETTER O WITH CIRCUMFLEX */
-1,-1,-1, /* p q r */
0x015D, /* SMALL LETTER S WITH CIRCUMFLEX */
-1, /* t */
0x00FB, /* SMALL LETTER U WITH CIRCUMFLEX */
-1, /* v */
0x0175, /* SMALL LETTER W WITH CIRCUMFLEX */
-1, /* x */
0x0177, /* SMALL LETTER Y WITH CIRCUMFLEX */
0x1E91, /* SMALL LETTER Z WITH CIRCUMFLEX */
0 /* OVER */
},{
/* 3 = TILDE */
0x00E3, /* SMALL LETTER A WITH TILDE */
-1,-1,-1, /* b c d */
0x1EBD, /* SMALL LETTER E WITH TILDE */
-1,-1,-1, /* f g h */
0x0129, /* SMALL LETTER I WITH TILDE */
-1,-1,-1,-1, /* j k l m */
0x00F1, /* SMALL LETTER N WITH TILDE */
0x00F5, /* SMALL LETTER O WITH TILDE */
-1,-1,-1,-1,-1, /* p q r s t */
0x0169, /* SMALL LETTER U WITH TILDE */
0x1E7D, /* SMALL LETTER V WITH TILDE */
-1,-1, /* w x */
0x1EF9, /* SMALL LETTER Y WITH TILDE */
-1, /* z */
0 /* OVER */
},{
/* 4 = UMLAUT */
0x00E4, /* SMALL LETTER A WITH DIAERESIS */
-1,-1,-1, /* b c d */
0x00EB, /* SMALL LETTER E WITH DIAERESIS */
-1,-1, /* f g */
0x1E27, /* SMALL LETTER H WITH DIAERESIS */
0x00EF, /* SMALL LETTER I WITH DIAERESIS */
-1,-1,-1,-1,-1, /* j k l m n */
0x00F6, /* SMALL LETTER O WITH DIAERESIS */
-1,-1,-1,-1, /* p q r s */
0x1E97, /* SMALL LETTER T WITH DIAERESIS */
0x00FC, /* SMALL LETTER U WITH DIAERESIS */
-1, /* v */
0x1E85, /* SMALL LETTER W WITH DIAERESIS */
0x1E8D, /* SMALL LETTER X WITH DIAERESIS */
0x00FF, /* SMALL LETTER Y WITH DIAERESIS */
-1, /* z */
0 /* OVER */
},{
/* 5 = DOUBLE ACUTE */
-1, /* a */
-1,-1,-1, /* b c d */
-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /* e f g h i j k l m n */
0x0151, /* SMALL LETTER O WITH DOUBLE ACUTE */
-1,-1,-1,-1,-1, /* p q r s t */
0x0171, /* SMALL LETTER U WITH DOUBLE ACUTE */
-1, -1, -1, -1, -1, /* v w x y z */
0 /* OVER */
},{
/* 6 = RING */
0x00E5, /* SMALL LETTER A WITH RING ABOVE */
-1,-1,-1, /* b c d */
-1, /* e */
-1,-1,-1, /* f g h */
-1, /* i */
-1,-1,-1,-1,-1, /* j k l m n */
-1, /* o */
-1,-1,-1,-1,-1, /* p q r s t */
0x016F, /* SMALL LETTER U WITH RING ABOVE */
-1, /* v */
0x1E98, /* SMALL LETTER W WITH RING ABOVE */
-1, /* x */
0x1E99, /* SMALL LETTER Y WITH RING ABOVE */
-1, /* z */
0 /* OVER */
},{
/* 7 = CARON */
0x01CE, /* SMALL LETTER A WITH CARON */
-1, /* b */
0x010D, /* SMALL LETTER C WITH CARON */
0x010F, /* SMALL LETTER D WITH CARON */
0x011B, /* SMALL LETTER E WITH CARON */
-1, /* f */
0x01E7, /* SMALL LETTER G WITH CARON */
0x021F, /* SMALL LETTER H WITH CARON */
0x01D0, /* SMALL LETTER I WITH CARON */
0x01F0, /* SMALL LETTER J WITH CARON */
0x01E9, /* SMALL LETTER K WITH CARON */
0x013E, /* SMALL LETTER L WITH CARON */
-1, /* m */
0x0148, /* SMALL LETTER N WITH CARON */
0x01D2, /* SMALL LETTER O WITH CARON */
-1, /* p */
-1, /* q */
0x0159, /* SMALL LETTER R WITH CARON */
0x0161, /* SMALL LETTER S WITH CARON */
0x0165, /* SMALL LETTER T WITH CARON */
0x01D4, /* SMALL LETTER U WITH CARON */
-1, /* v */
-1, /* w */
-1, /* x */
-1, /* y */
0x017E, /* SMALL LETTER Z WITH CARON */
0 /* OVER */
},{
/* 8 = BREVE */
0x0103, /* SMALL LETTER A WITH BREVE */
-1,-1,-1, /* b c d */
0x0115, /* SMALL LETTER E WITH BREVE */
-1, /* f */
0x011F, /* SMALL LETTER G WITH BREVE */
-1, /* h */
0x012D, /* SMALL LETTER I WITH BREVE */
-1,-1,-1,-1,-1, /* j k l m n */
0x014F, /* SMALL LETTER O WITH BREVE */
-1,-1,-1,-1,-1, /* p q r s t */
0x016D, /* SMALL LETTER U WITH BREVE */
-1, -1, -1, -1, -1, /* v w x y z */
0 /* OVER */
},{
/* 9 = MACRON */
0x0101, /* SMALL LETTER A WITH MACRON */
-1,-1,-1, /* b c d */
0x0113, /* SMALL LETTER E WITH MACRON */
-1, /* f */
0x1E21, /* SMALL LETTER G WITH MACRON */
-1, /* h */
0x012B, /* SMALL LETTER I WITH MACRON */
-1,-1,-1,-1,-1, /* j k l m n */
0x014D, /* SMALL LETTER O WITH MACRON */
-1,-1,-1,-1,-1, /* p q r s t */
0x016B, /* SMALL LETTER U WITH MACRON */
-1,-1,-1, /* v w x */
0x0233, /* SMALL LETTER Y WITH MACRON */
-1, /* z */
0 /* OVER */
},{
/* 10 = DOT */
0x0227, /* SMALL LETTER A WITH DOT ABOVE */
0x1E03, /* SMALL LETTER B WITH DOT ABOVE */
0x010B, /* SMALL LETTER C WITH DOT ABOVE */
0x1E0B, /* SMALL LETTER D WITH DOT ABOVE */
0x0117, /* SMALL LETTER E WITH DOT ABOVE */
0x1E1F, /* SMALL LETTER F WITH DOT ABOVE */
0x0121, /* SMALL LETTER G WITH DOT ABOVE */
0x1E23, /* SMALL LETTER H WITH DOT ABOVE */
-1,-1,-1,-1, /* i j k l */
0x1E41, /* SMALL LETTER M WITH DOT ABOVE */
0x1E45, /* SMALL LETTER N WITH DOT ABOVE */
0x022F, /* SMALL LETTER O WITH DOT ABOVE */
0x1E57, /* SMALL LETTER P WITH DOT ABOVE */
-1, /* q */
0x1E59, /* SMALL LETTER R WITH DOT ABOVE */
0x1E61, /* SMALL LETTER S WITH DOT ABOVE */
0x1E6B, /* SMALL LETTER T WITH DOT ABOVE */
-1, /* u */
-1, /* v */
0x1E87, /* SMALL LETTER W WITH DOT ABOVE */
0x1E8B, /* SMALL LETTER X WITH DOT ABOVE */
0x1E8F, /* SMALL LETTER Y WITH DOT ABOVE */
0x017C, /* SMALL LETTER Z WITH DOT ABOVE */
0 /* OVER */
},{
/* 11 = CEDILLA */
-1, /* a */
-1, /* b */
0x00E7, /* SMALL LETTER C WITH CEDILLA */
0x1E11, /* SMALL LETTER D WITH CEDILLA */
0x0229, /* SMALL LETTER E WITH CEDILLA */
-1, /* f */
0x0123, /* SMALL LETTER G WITH CEDILLA */
0x1E29, /* SMALL LETTER H WITH CEDILLA */
-1, /* i */
-1, /* j */
0x0137, /* SMALL LETTER K WITH CEDILLA */
0x013C, /* SMALL LETTER L WITH CEDILLA */
-1, /* m */
0x0146, /* SMALL LETTER N WITH CEDILLA */
-1, /* o */
-1, /* p */
-1, /* q */
0x0157, /* SMALL LETTER R WITH CEDILLA */
0x015F, /* SMALL LETTER S WITH CEDILLA */
0x0163, /* SMALL LETTER T WITH CEDILLA */
-1, /* u */
-1, -1, -1, -1, -1, /* v w x y z */
1 /* UNDER */
},{
/* 12 = OGONEK */
0x0105, /* SMALL LETTER A WITH OGONEK */
-1,-1,-1, /* b c d */
0x0119, /* SMALL LETTER E WITH OGONEK */
-1,-1,-1, /* f g h */
0x012F, /* SMALL LETTER I WITH OGONEK */
-1,-1,-1,-1,-1, /* j k l m n */
0x01EB, /* SMALL LETTER O WITH OGONEK */
-1,-1,-1,-1,-1, /* p q r s t */
0x0173, /* SMALL LETTER U WITH OGONEK */
-1, -1, -1, -1, -1, /* v w x y z */
1 /* UNDER */
},{
/* 13 = DOUBLE GRAVE */
0x0201, /* SMALL LETTER A WITH DOUBLE GRAVE */
-1,-1,-1, /* b c d */
0x0205, /* SMALL LETTER E WITH DOUBLE GRAVE */
-1,-1,-1, /* f g h */
0x0209, /* SMALL LETTER I WITH DOUBLE GRAVE */
-1,-1,-1,-1,-1, /* j k l m n */
0x020D, /* SMALL LETTER O WITH DOUBLE GRAVE */
-1,-1, /* p q */
0x0211, /* SMALL LETTER R WITH DOUBLE GRAVE */
-1,-1, /* s t */
0x0215, /* SMALL LETTER U WITH DOUBLE GRAVE */
-1, -1, -1, -1, -1, /* v w x y z */
0 /* OVER */
},{
/* 14 = DOT BELOW */
0x1EA1, /* ".a" dot under a: SMALL LETTER A WITH DOT BELOW */
0x1E05, /* ".b" dot under b: SMALL LETTER B WITH DOT BELOW */
-1, /* c */
0x1E0D, /* ".d" dot under d: SMALL LETTER D WITH DOT BELOW */
0x1EB9, /* ".e" dot under e: SMALL LETTER E WITH DOT BELOW */
-1, /* f */
-1, /* g */
0x1E25, /* ".h" dot under h: SMALL LETTER H WITH DOT BELOW */
0x1ECB, /* ".i" dot under i: SMALL LETTER I WITH DOT BELOW */
-1, /* j */
0x1E33, /* ".k" dot under k: SMALL LETTER K WITH DOT BELOW */
0x1E37, /* ".l" dot under l: SMALL LETTER L WITH DOT BELOW */
0x1E43, /* ".m" dot under m: SMALL LETTER M WITH DOT BELOW */
0x1E47, /* ".n" dot under n: SMALL LETTER N WITH DOT BELOW */
0x1ECD, /* ".o" dot under o: SMALL LETTER O WITH DOT BELOW */
-1, /* p */
-1, /* q */
0x1E5B, /* ".r" dot under r: SMALL LETTER R WITH DOT BELOW */
0x1E63, /* ".s" dot under s: SMALL LETTER S WITH DOT BELOW */
0x1E6D, /* ".t" dot under t: SMALL LETTER T WITH DOT BELOW */
0x1EE5, /* ".u" dot under u: SMALL LETTER U WITH DOT BELOW */
0x1E7F, /* ".v" dot under v: SMALL LETTER V WITH DOT BELOW */
0x1E89, /* ".w" dot under w: SMALL LETTER W WITH DOT BELOW */
-1, /* x */
0x1EF5, /* ".y" dot under y: SMALL LETTER Y WITH DOT BELOW */
0x1E93, /* ".z" dot under z: SMALL LETTER Z WITH DOT BELOW */
1 /* UNDER */
}
};

static int uni_symbol[256] = {
-1,-1,-1,-1,-1,-1,-1,-1,
-1,-1,-1,-1,-1,-1,-1,-1,
-1,-1,-1,-1,-1,-1,-1,-1,
-1,-1,-1,-1,-1,-1,-1,-1,
0x0020,	/* 20	# SPACE	# space */
0x0021,	/* 21	# EXCLAMATION MARK	# exclam */
0x2200,	/* 22	# FOR ALL	# universal */
0x0023,	/* 23	# NUMBER SIGN	# numbersign */
0x2203,	/* 24	# THERE EXISTS	# existential */
0x0025,	/* 25	# PERCENT SIGN	# percent */
0x0026,	/* 26	# AMPERSAND	# ampersand */
0x220B,	/* 27	# CONTAINS AS MEMBER	# suchthat */
0x0028,	/* 28	# LEFT PARENTHESIS	# parenleft */
0x0029,	/* 29	# RIGHT PARENTHESIS	# parenright */
0x2217,	/* 2A	# ASTERISK OPERATOR	# asteriskmath */
0x002B,	/* 2B	# PLUS SIGN	# plus */
0x002C,	/* 2C	# COMMA	# comma */
0x2212,	/* 2D	# MINUS SIGN	# minus */
0x002E,	/* 2E	# FULL STOP	# period */
0x002F,	/* 2F	# SOLIDUS	# slash */
0x0030,	/* 30	# DIGIT ZERO	# zero */
0x0031,	/* 31	# DIGIT ONE	# one */
0x0032,	/* 32	# DIGIT TWO	# two */
0x0033,	/* 33	# DIGIT THREE	# three */
0x0034,	/* 34	# DIGIT FOUR	# four */
0x0035,	/* 35	# DIGIT FIVE	# five */
0x0036,	/* 36	# DIGIT SIX	# six */
0x0037,	/* 37	# DIGIT SEVEN	# seven */
0x0038,	/* 38	# DIGIT EIGHT	# eight */
0x0039,	/* 39	# DIGIT NINE	# nine */
0x003A,	/* 3A	# COLON	# colon */
0x003B,	/* 3B	# SEMICOLON	# semicolon */
0x003C,	/* 3C	# LESS-THAN SIGN	# less */
0x003D,	/* 3D	# EQUALS SIGN	# equal */
0x003E,	/* 3E	# GREATER-THAN SIGN	# greater */
0x003F,	/* 3F	# QUESTION MARK	# question */
0x2245,	/* 40	# APPROXIMATELY EQUAL TO	# congruent */
0x0391,	/* 41	# GREEK CAPITAL LETTER ALPHA	# Alpha */
0x0392,	/* 42	# GREEK CAPITAL LETTER BETA	# Beta */
0x03A7,	/* 43	# GREEK CAPITAL LETTER CHI	# Chi */
0x0394,	/* 44	# GREEK CAPITAL LETTER DELTA	# Delta cf. 0x2206, 44	# INCREMENT	# Delta */
0x0395,	/* 45	# GREEK CAPITAL LETTER EPSILON	# Epsilon */
0x03A6,	/* 46	# GREEK CAPITAL LETTER PHI	# Phi */
0x0393,	/* 47	# GREEK CAPITAL LETTER GAMMA	# Gamma */
0x0397,	/* 48	# GREEK CAPITAL LETTER ETA	# Eta */
0x0399,	/* 49	# GREEK CAPITAL LETTER IOTA	# Iota */
0x03D1,	/* 4A	# GREEK THETA SYMBOL	# theta1 */
0x039A,	/* 4B	# GREEK CAPITAL LETTER KAPPA	# Kappa */
0x039B,	/* 4C	# GREEK CAPITAL LETTER LAMDA	# Lambda */
0x039C,	/* 4D	# GREEK CAPITAL LETTER MU	# Mu */
0x039D,	/* 4E	# GREEK CAPITAL LETTER NU	# Nu */
0x039F,	/* 4F	# GREEK CAPITAL LETTER OMICRON	# Omicron */
0x03A0,	/* 50	# GREEK CAPITAL LETTER PI	# Pi */
0x0398,	/* 51	# GREEK CAPITAL LETTER THETA	# Theta */
0x03A1,	/* 52	# GREEK CAPITAL LETTER RHO	# Rho */
0x03A3,	/* 53	# GREEK CAPITAL LETTER SIGMA	# Sigma */
0x03A4,	/* 54	# GREEK CAPITAL LETTER TAU	# Tau */
0x03A5,	/* 55	# GREEK CAPITAL LETTER UPSILON	# Upsilon */
0x03C2,	/* 56	# GREEK SMALL LETTER FINAL SIGMA	# sigma1 */
0x03A9,	/* 57	# GREEK CAPITAL LETTER OMEGA	# Omega */
/* 0x2126, 57	# OHM SIGN	# Omega */
0x039E,	/* 58	# GREEK CAPITAL LETTER XI	# Xi */
0x03A8,	/* 59	# GREEK CAPITAL LETTER PSI	# Psi */
0x0396,	/* 5A	# GREEK CAPITAL LETTER ZETA	# Zeta */
0x005B,	/* 5B	# LEFT SQUARE BRACKET	# bracketleft */
0x2234,	/* 5C	# THEREFORE	# therefore */
0x005D,	/* 5D	# RIGHT SQUARE BRACKET	# bracketright */
0x22A5,	/* 5E	# UP TACK	# perpendicular */
0x005F,	/* 5F	# LOW LINE	# underscore */
0xF8E5,	/* 60	# RADICAL EXTENDER	# radicalex (CUS) */
0x03B1,	/* 61	# GREEK SMALL LETTER ALPHA	# alpha */
0x03B2,	/* 62	# GREEK SMALL LETTER BETA	# beta */
0x03C7,	/* 63	# GREEK SMALL LETTER CHI	# chi */
0x03B4,	/* 64	# GREEK SMALL LETTER DELTA	# delta */
0x03B5,	/* 65	# GREEK SMALL LETTER EPSILON	# epsilon */
0x03C6,	/* 66	# GREEK SMALL LETTER PHI	# phi */
0x03B3,	/* 67	# GREEK SMALL LETTER GAMMA	# gamma */
0x03B7,	/* 68	# GREEK SMALL LETTER ETA	# eta */
0x03B9,	/* 69	# GREEK SMALL LETTER IOTA	# iota */
0x03D5,	/* 6A	# GREEK PHI SYMBOL	# phi1 */
0x03BA,	/* 6B	# GREEK SMALL LETTER KAPPA	# kappa */
0x03BB,	/* 6C	# GREEK SMALL LETTER LAMDA	# lambda */
/* 0x00B5,	6D	# MICRO SIGN	# mu */
0x03BC,	/* 6D	# GREEK SMALL LETTER MU	# mu */
0x03BD,	/* 6E	# GREEK SMALL LETTER NU	# nu */
0x03BF,	/* 6F	# GREEK SMALL LETTER OMICRON	# omicron */
0x03C0,	/* 70	# GREEK SMALL LETTER PI	# pi */
0x03B8,	/* 71	# GREEK SMALL LETTER THETA	# theta */
0x03C1,	/* 72	# GREEK SMALL LETTER RHO	# rho */
0x03C3,	/* 73	# GREEK SMALL LETTER SIGMA	# sigma */
0x03C4,	/* 74	# GREEK SMALL LETTER TAU	# tau */
0x03C5,	/* 75	# GREEK SMALL LETTER UPSILON	# upsilon */
0x03D6,	/* 76	# GREEK PI SYMBOL	# omega1 */
0x03C9,	/* 77	# GREEK SMALL LETTER OMEGA	# omega */
0x03BE,	/* 78	# GREEK SMALL LETTER XI	# xi */
0x03C8,	/* 79	# GREEK SMALL LETTER PSI	# psi */
0x03B6,	/* 7A	# GREEK SMALL LETTER ZETA	# zeta */
0x007B,	/* 7B	# LEFT CURLY BRACKET	# braceleft */
0x007C,	/* 7C	# VERTICAL LINE	# bar */
0x007D,	/* 7D	# RIGHT CURLY BRACKET	# braceright */
0x223C,	/* 7E	# TILDE OPERATOR	# similar */
-1,
-1,-1,-1,-1,-1,-1,-1,-1,
-1,-1,-1,-1,-1,-1,-1,-1,
-1,-1,-1,-1,-1,-1,-1,-1,
-1,-1,-1,-1,-1,-1,-1,-1,
0x20AC,	/* A0	# EURO SIGN	# Euro */
0x03D2,	/* A1	# GREEK UPSILON WITH HOOK SYMBOL	# Upsilon1 */
0x2032,	/* A2	# PRIME	# minute */
0x2264,	/* A3	# LESS-THAN OR EQUAL TO	# lessequal */
0x2044,	/* A4	# FRACTION SLASH	# fraction */
0x2215,	/* A4	# DIVISION SLASH	# fraction */
0x221E,	/* A5	# INFINITY	# infinity */
0x0192,	/* A6	# LATIN SMALL LETTER F WITH HOOK	# florin */
0x2663,	/* A7	# BLACK CLUB SUIT	# club */
0x2666,	/* A8	# BLACK DIAMOND SUIT	# diamond */
0x2665,	/* A9	# BLACK HEART SUIT	# heart */
0x2660,	/* AA	# BLACK SPADE SUIT	# spade */
0x2194,	/* AB	# LEFT RIGHT ARROW	# arrowboth */
0x2190,	/* AC	# LEFTWARDS ARROW	# arrowleft */
0x2191,	/* AD	# UPWARDS ARROW	# arrowup */
0x2192,	/* AE	# RIGHTWARDS ARROW	# arrowright */
0x2193,	/* AF	# DOWNWARDS ARROW	# arrowdown */
0x00B0,	/* B0	# DEGREE SIGN	# degree */
0x00B1,	/* B1	# PLUS-MINUS SIGN	# plusminus */
0x2033,	/* B2	# DOUBLE PRIME	# second */
0x2265,	/* B3	# GREATER-THAN OR EQUAL TO	# greaterequal */
0x00D7,	/* B4	# MULTIPLICATION SIGN	# multiply */
0x221D,	/* B5	# PROPORTIONAL TO	# proportional */
0x2202,	/* B6	# PARTIAL DIFFERENTIAL	# partialdiff */
0x2022,	/* B7	# BULLET	# bullet */
0x00F7,	/* B8	# DIVISION SIGN	# divide */
0x2260,	/* B9	# NOT EQUAL TO	# notequal */
0x2261,	/* BA	# IDENTICAL TO	# equivalence */
0x2248,	/* BB	# ALMOST EQUAL TO	# approxequal */
0x2026,	/* BC	# HORIZONTAL ELLIPSIS	# ellipsis */
0xF8E6,	/* BD	# VERTICAL ARROW EXTENDER	# arrowvertex (CUS) */
0xF8E7,	/* BE	# HORIZONTAL ARROW EXTENDER	# arrowhorizex (CUS) */
0x21B5,	/* BF	# DOWNWARDS ARROW WITH CORNER LEFTWARDS	# carriagereturn */
0x2135,	/* C0	# ALEF SYMBOL	# aleph */
0x2111,	/* C1	# BLACK-LETTER CAPITAL I	# Ifraktur */
0x211C,	/* C2	# BLACK-LETTER CAPITAL R	# Rfraktur */
0x2118,	/* C3	# SCRIPT CAPITAL P	# weierstrass */
0x2297,	/* C4	# CIRCLED TIMES	# circlemultiply */
0x2295,	/* C5	# CIRCLED PLUS	# circleplus */
0x2205,	/* C6	# EMPTY SET	# emptyset */
0x2229,	/* C7	# INTERSECTION	# intersection */
0x222A,	/* C8	# UNION	# union */
0x2283,	/* C9	# SUPERSET OF	# propersuperset */
0x2287,	/* CA	# SUPERSET OF OR EQUAL TO	# reflexsuperset */
0x2284,	/* CB	# NOT A SUBSET OF	# notsubset */
0x2282,	/* CC	# SUBSET OF	# propersubset */
0x2286,	/* CD	# SUBSET OF OR EQUAL TO	# reflexsubset */
0x2208,	/* CE	# ELEMENT OF	# element */
0x2209,	/* CF	# NOT AN ELEMENT OF	# notelement */
0x2220,	/* D0	# ANGLE	# angle */
0x2207,	/* D1	# NABLA	# gradient */
0xF6DA,	/* D2	# REGISTERED SIGN SERIF	# registerserif (CUS) */
0xF6D9,	/* D3	# COPYRIGHT SIGN SERIF	# copyrightserif (CUS) */
0xF6DB,	/* D4	# TRADE MARK SIGN SERIF	# trademarkserif (CUS) */
0x220F,	/* D5	# N-ARY PRODUCT	# product */
0x221A,	/* D6	# SQUARE ROOT	# radical */
0x22C5,	/* D7	# DOT OPERATOR	# dotmath */
0x00AC,	/* D8	# NOT SIGN	# logicalnot */
0x2227,	/* D9	# LOGICAL AND	# logicaland */
0x2228,	/* DA	# LOGICAL OR	# logicalor */
0x21D4,	/* DB	# LEFT RIGHT DOUBLE ARROW	# arrowdblboth */
0x21D0,	/* DC	# LEFTWARDS DOUBLE ARROW	# arrowdblleft */
0x21D1,	/* DD	# UPWARDS DOUBLE ARROW	# arrowdblup */
0x21D2,	/* DE	# RIGHTWARDS DOUBLE ARROW	# arrowdblright */
0x21D3,	/* DF	# DOWNWARDS DOUBLE ARROW	# arrowdbldown */
0x25CA,	/* E0	# LOZENGE	# lozenge */
0x2329,	/* E1	# LEFT-POINTING ANGLE BRACKET	# angleleft */
0xF8E8,	/* E2	# REGISTERED SIGN SANS SERIF	# registersans (CUS) */
0xF8E9,	/* E3	# COPYRIGHT SIGN SANS SERIF	# copyrightsans (CUS) */
0xF8EA,	/* E4	# TRADE MARK SIGN SANS SERIF	# trademarksans (CUS) */
0x2211,	/* E5	# N-ARY SUMMATION	# summation */
0xF8EB,	/* E6	# LEFT PAREN TOP	# parenlefttp (CUS) */
0xF8EC,	/* E7	# LEFT PAREN EXTENDER	# parenleftex (CUS) */
0xF8ED,	/* E8	# LEFT PAREN BOTTOM	# parenleftbt (CUS) */
0xF8EE,	/* E9	# LEFT SQUARE BRACKET TOP	# bracketlefttp (CUS) */
0xF8EF,	/* EA	# LEFT SQUARE BRACKET EXTENDER	# bracketleftex (CUS) */
0xF8F0,	/* EB	# LEFT SQUARE BRACKET BOTTOM	# bracketleftbt (CUS) */
0xF8F1,	/* EC	# LEFT CURLY BRACKET TOP	# bracelefttp (CUS) */
0xF8F2,	/* ED	# LEFT CURLY BRACKET MID	# braceleftmid (CUS) */
0xF8F3,	/* EE	# LEFT CURLY BRACKET BOTTOM	# braceleftbt (CUS) */
0xF8F4,	/* EF	# CURLY BRACKET EXTENDER	# braceex (CUS) */
0x232A,	/* F1	# RIGHT-POINTING ANGLE BRACKET	# angleright */
0x222B,	/* F2	# INTEGRAL	# integral */
0x2320,	/* F3	# TOP HALF INTEGRAL	# integraltp */
0xF8F5,	/* F4	# INTEGRAL EXTENDER	# integralex (CUS) */
0x2321,	/* F5	# BOTTOM HALF INTEGRAL	# integralbt */
0xF8F6,	/* F6	# RIGHT PAREN TOP	# parenrighttp (CUS) */
0xF8F7,	/* F7	# RIGHT PAREN EXTENDER	# parenrightex (CUS) */
0xF8F8,	/* F8	# RIGHT PAREN BOTTOM	# parenrightbt (CUS) */
0xF8F9,	/* F9	# RIGHT SQUARE BRACKET TOP	# bracketrighttp (CUS) */
0xF8FA,	/* FA	# RIGHT SQUARE BRACKET EXTENDER	# bracketrightex (CUS) */
0xF8FB,	/* FB	# RIGHT SQUARE BRACKET BOTTOM	# bracketrightbt (CUS) */
0xF8FC,	/* FC	# RIGHT CURLY BRACKET TOP	# bracerighttp (CUS) */
0xF8FD,	/* FD	# RIGHT CURLY BRACKET MID	# bracerightmid (CUS) */
0xF8FE,	/* FE	# RIGHT CURLY BRACKET BOTTOM	# bracerightbt (CUS) */
-1
};

int uni_zd[256] = {
-1,-1,-1,-1,-1,-1,-1,-1,
-1,-1,-1,-1,-1,-1,-1,-1,
-1,-1,-1,-1,-1,-1,-1,-1,
-1,-1,-1,-1,-1,-1,-1,-1,
0x0020, /* 20	SPACE */
0x2701, /* 21	UPPER BLADE SCISSORS */
0x2702, /* 22	BLACK SCISSORS */
0x2703, /* 23	LOWER BLADE SCISSORS */
0x2704, /* 24	WHITE SCISSORS */
0x260E, /* 25	BLACK TELEPHONE */
0x2706, /* 26	TELEPHONE LOCATION SIGN */
0x2707, /* 27	TAPE DRIVE */
0x2708, /* 28	AIRPLANE */
0x2709, /* 29	ENVELOPE */
0x261B, /* 2A	BLACK RIGHT POINTING INDEX */
0x261E, /* 2B	WHITE RIGHT POINTING INDEX */
0x270C, /* 2C	VICTORY HAND */
0x270D, /* 2D	WRITING HAND */
0x270E, /* 2E	LOWER RIGHT PENCIL */
0x270F, /* 2F	PENCIL */
0x2710, /* 30	UPPER RIGHT PENCIL */
0x2711, /* 31	WHITE NIB */
0x2712, /* 32	BLACK NIB */
0x2713, /* 33	CHECK MARK */
0x2714, /* 34	HEAVY CHECK MARK */
0x2715, /* 35	MULTIPLICATION X */
0x2716, /* 36	HEAVY MULTIPLICATION X */
0x2717, /* 37	BALLOT X */
0x2718, /* 38	HEAVY BALLOT X */
0x2719, /* 39	OUTLINED GREEK CROSS */
0x271A, /* 3A	HEAVY GREEK CROSS */
0x271B, /* 3B	OPEN CENTRE CROSS */
0x271C, /* 3C	HEAVY OPEN CENTRE CROSS */
0x271D, /* 3D	LATIN CROSS */
0x271E, /* 3E	SHADOWED WHITE LATIN CROSS */
0x271F, /* 3F	OUTLINED LATIN CROSS */
0x2720, /* 40	MALTESE CROSS */
0x2721, /* 41	STAR OF DAVID */
0x2722, /* 42	FOUR TEARDROP-SPOKED ASTERISK */
0x2723, /* 43	FOUR BALLOON-SPOKED ASTERISK */
0x2724, /* 44	HEAVY FOUR BALLOON-SPOKED ASTERISK */
0x2725, /* 45	FOUR CLUB-SPOKED ASTERISK */
0x2726, /* 46	BLACK FOUR POINTED STAR */
0x2727, /* 47	WHITE FOUR POINTED STAR */
0x2605, /* 48	BLACK STAR */
0x2729, /* 49	STRESS OUTLINED WHITE STAR */
0x272A, /* 4A	CIRCLED WHITE STAR */
0x272B, /* 4B	OPEN CENTRE BLACK STAR */
0x272C, /* 4C	BLACK CENTRE WHITE STAR */
0x272D, /* 4D	OUTLINED BLACK STAR */
0x272E, /* 4E	HEAVY OUTLINED BLACK STAR */
0x272F, /* 4F	PINWHEEL STAR */
0x2730, /* 50	SHADOWED WHITE STAR */
0x2731, /* 51	HEAVY ASTERISK */
0x2732, /* 52	OPEN CENTRE ASTERISK */
0x2733, /* 53	EIGHT SPOKED ASTERISK */
0x2734, /* 54	EIGHT POINTED BLACK STAR */
0x2735, /* 55	EIGHT POINTED PINWHEEL STAR */
0x2736, /* 56	SIX POINTED BLACK STAR */
0x2737, /* 57	EIGHT POINTED RECTILINEAR BLACK STAR */
0x2738, /* 58	HEAVY EIGHT POINTED RECTILINEAR BLACK STAR */
0x2739, /* 59	TWELVE POINTED BLACK STAR */
0x273A, /* 5A	SIXTEEN POINTED ASTERISK */
0x273B, /* 5B	TEARDROP-SPOKED ASTERISK */
0x273C, /* 5C	OPEN CENTRE TEARDROP-SPOKED ASTERISK */
0x273D, /* 5D	HEAVY TEARDROP-SPOKED ASTERISK */
0x273E, /* 5E	SIX PETALLED BLACK AND WHITE FLORETTE */
0x273F, /* 5F	BLACK FLORETTE */
0x2740, /* 60	WHITE FLORETTE */
0x2741, /* 61	EIGHT PETALLED OUTLINED BLACK FLORETTE */
0x2742, /* 62	CIRCLED OPEN CENTRE EIGHT POINTED STAR */
0x2743, /* 63	HEAVY TEARDROP-SPOKED PINWHEEL ASTERISK */
0x2744, /* 64	SNOWFLAKE */
0x2745, /* 65	TIGHT TRIFOLIATE SNOWFLAKE */
0x2746, /* 66	HEAVY CHEVRON SNOWFLAKE */
0x2747, /* 67	SPARKLE */
0x2748, /* 68	HEAVY SPARKLE */
0x2749, /* 69	BALLOON-SPOKED ASTERISK */
0x274A, /* 6A	EIGHT TEARDROP-SPOKED PROPELLER ASTERISK */
0x274B, /* 6B	HEAVY EIGHT TEARDROP-SPOKED PROPELLER ASTERISK */
0x25CF, /* 6C	BLACK CIRCLE */
0x274D, /* 6D	SHADOWED WHITE CIRCLE */
0x25A0, /* 6E	BLACK SQUARE */
0x274F, /* 6F	LOWER RIGHT DROP-SHADOWED WHITE SQUARE */
0x2750, /* 70	UPPER RIGHT DROP-SHADOWED WHITE SQUARE */
0x2751, /* 71	LOWER RIGHT SHADOWED WHITE SQUARE */
0x2752, /* 72	UPPER RIGHT SHADOWED WHITE SQUARE */
0x25B2, /* 73	BLACK UP-POINTING TRIANGLE */
0x25BC, /* 74	BLACK DOWN-POINTING TRIANGLE */
0x25C6, /* 75	BLACK DIAMOND */
0x2756, /* 76	BLACK DIAMOND MINUS WHITE X */
0x25D7, /* 77	RIGHT HALF BLACK CIRCLE */
0x2758, /* 78	LIGHT VERTICAL BAR */
0x2759, /* 79	MEDIUM VERTICAL BAR */
0x275A, /* 7A	HEAVY VERTICAL BAR */
0x275B, /* 7B	HEAVY SINGLE TURNED COMMA QUOTATION MARK ORNAMENT */
0x275C, /* 7C	HEAVY SINGLE COMMA QUOTATION MARK ORNAMENT */
0x275D, /* 7D	HEAVY DOUBLE TURNED COMMA QUOTATION MARK ORNAMENT */
0x275E, /* 7E	HEAVY DOUBLE COMMA QUOTATION MARK ORNAMENT */
-1,
0xF8D7, /* 80	MEDIUM LEFT PARENTHESIS ORNAMENT */
0xF8D8, /* 81	MEDIUM RIGHT PARENTHESIS ORNAMENT */
0xF8D9, /* 82	MEDIUM FLATTENED LEFT PARENTHESIS ORNAMENT */
0xF8DA, /* 83	MEDIUM FLATTENED RIGHT PARENTHESIS ORNAMENT */
0xF8DB, /* 84	MEDIUM LEFT-POINTING ANGLE BRACKET ORNAMENT */
0xF8DC, /* 85	MEDIUM RIGHT-POINTING ANGLE BRACKET ORNAMENT */
0xF8DD, /* 86	HEAVY LEFT-POINTING ANGLE QUOTATION MARK ORNAMENT */
0xF8DE, /* 87	HEAVY RIGHT-POINTING ANGLE QUOTATION MARK ORNAMENT */
0xF8DF, /* 88	HEAVY LEFT-POINTING ANGLE BRACKET ORNAMENT */
0xF8E0, /* 89	HEAVY RIGHT-POINTING ANGLE BRACKET ORNAMENT */
0xF8E1, /* 8A	LIGHT LEFT TORTOISE SHELL BRACKET ORNAMENT */
0xF8E2, /* 8B	LIGHT RIGHT TORTOISE SHELL BRACKET ORNAMENT */
0xF8E3, /* 8C	MEDIUM LEFT CURLY BRACKET ORNAMENT */
0xF8E4, /* 8D	MEDIUM RIGHT CURLY BRACKET ORNAMENT */
-1,-1,
-1,-1,-1,-1,-1,-1,-1,-1,
-1,-1,-1,-1,-1,-1,-1,-1,
-1,
0x2761, /* A1	CURVED STEM PARAGRAPH SIGN ORNAMENT */
0x2762, /* A2	HEAVY EXCLAMATION MARK ORNAMENT */
0x2763, /* A3	HEAVY HEART EXCLAMATION MARK ORNAMENT */
0x2764, /* A4	HEAVY BLACK HEART */
0x2765, /* A5	ROTATED HEAVY BLACK HEART BULLET */
0x2766, /* A6	FLORAL HEART */
0x2767, /* A7	ROTATED FLORAL HEART BULLET */
0x2663, /* A8	BLACK CLUB SUIT */
0x2666, /* A9	BLACK DIAMOND SUIT */
0x2665, /* AA	BLACK HEART SUIT */
0x2660, /* AB	BLACK SPADE SUIT */
0x2460, /* AC	CIRCLED DIGIT ONE */
0x2461, /* AD	CIRCLED DIGIT TWO */
0x2462, /* AE	CIRCLED DIGIT THREE */
0x2463, /* AF	CIRCLED DIGIT FOUR */
0x2464, /* B0	CIRCLED DIGIT FIVE */
0x2465, /* B1	CIRCLED DIGIT SIX */
0x2466, /* B2	CIRCLED DIGIT SEVEN */
0x2467, /* B3	CIRCLED DIGIT EIGHT */
0x2468, /* B4	CIRCLED DIGIT NINE */
0x2469, /* B5	CIRCLED NUMBER TEN */
0x2776, /* B6	DINGBAT NEGATIVE CIRCLED DIGIT ONE */
0x2777, /* B7	DINGBAT NEGATIVE CIRCLED DIGIT TWO */
0x2778, /* B8	DINGBAT NEGATIVE CIRCLED DIGIT THREE */
0x2779, /* B9	DINGBAT NEGATIVE CIRCLED DIGIT FOUR */
0x277A, /* BA	DINGBAT NEGATIVE CIRCLED DIGIT FIVE */
0x277B, /* BB	DINGBAT NEGATIVE CIRCLED DIGIT SIX */
0x277C, /* BC	DINGBAT NEGATIVE CIRCLED DIGIT SEVEN */
0x277D, /* BD	DINGBAT NEGATIVE CIRCLED DIGIT EIGHT */
0x277E, /* BE	DINGBAT NEGATIVE CIRCLED DIGIT NINE */
0x277F, /* BF	DINGBAT NEGATIVE CIRCLED NUMBER TEN */
0x2780, /* C0	DINGBAT CIRCLED SANS-SERIF DIGIT ONE */
0x2781, /* C1	DINGBAT CIRCLED SANS-SERIF DIGIT TWO */
0x2782, /* C2	DINGBAT CIRCLED SANS-SERIF DIGIT THREE */
0x2783, /* C3	DINGBAT CIRCLED SANS-SERIF DIGIT FOUR */
0x2784, /* C4	DINGBAT CIRCLED SANS-SERIF DIGIT FIVE */
0x2785, /* C5	DINGBAT CIRCLED SANS-SERIF DIGIT SIX */
0x2786, /* C6	DINGBAT CIRCLED SANS-SERIF DIGIT SEVEN */
0x2787, /* C7	DINGBAT CIRCLED SANS-SERIF DIGIT EIGHT */
0x2788, /* C8	DINGBAT CIRCLED SANS-SERIF DIGIT NINE */
0x2789, /* C9	DINGBAT CIRCLED SANS-SERIF NUMBER TEN */
0x278A, /* CA	DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT ONE */
0x278B, /* CB	DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT TWO */
0x278C, /* CC	DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT THREE */
0x278D, /* CD	DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT FOUR */
0x278E, /* CE	DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT FIVE */
0x278F, /* CF	DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT SIX */
0x2790, /* D0	DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT SEVEN */
0x2791, /* D1	DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT EIGHT */
0x2792, /* D2	DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT NINE */
0x2793, /* D3	DINGBAT NEGATIVE CIRCLED SANS-SERIF NUMBER TEN */
0x2794, /* D4	HEAVY WIDE-HEADED RIGHTWARDS ARROW */
0x2192, /* D5	RIGHTWARDS ARROW */
0x2194, /* D6	LEFT RIGHT ARROW */
0x2195, /* D7	UP DOWN ARROW */
0x2798, /* D8	HEAVY SOUTH EAST ARROW */
0x2799, /* D9	HEAVY RIGHTWARDS ARROW */
0x279A, /* DA	HEAVY NORTH EAST ARROW */
0x279B, /* DB	DRAFTING POINT RIGHTWARDS ARROW */
0x279C, /* DC	HEAVY ROUND-TIPPED RIGHTWARDS ARROW */
0x279D, /* DD	TRIANGLE-HEADED RIGHTWARDS ARROW */
0x279E, /* DE	HEAVY TRIANGLE-HEADED RIGHTWARDS ARROW */
0x279F, /* DF	DASHED TRIANGLE-HEADED RIGHTWARDS ARROW */
0x27A0, /* E0	HEAVY DASHED TRIANGLE-HEADED RIGHTWARDS ARROW */
0x27A1, /* E1	BLACK RIGHTWARDS ARROW */
0x27A2, /* E2	THREE-D TOP-LIGHTED RIGHTWARDS ARROWHEAD */
0x27A3, /* E3	THREE-D BOTTOM-LIGHTED RIGHTWARDS ARROWHEAD */
0x27A4, /* E4	BLACK RIGHTWARDS ARROWHEAD */
0x27A5, /* E5	HEAVY BLACK CURVED DOWNWARDS AND RIGHTWARDS ARROW */
0x27A6, /* E6	HEAVY BLACK CURVED UPWARDS AND RIGHTWARDS ARROW */
0x27A7, /* E7	SQUAT BLACK RIGHTWARDS ARROW */
0x27A8, /* E8	HEAVY CONCAVE-POINTED BLACK RIGHTWARDS ARROW */
0x27A9, /* E9	RIGHT-SHADED WHITE RIGHTWARDS ARROW */
0x27AA, /* EA	LEFT-SHADED WHITE RIGHTWARDS ARROW */
0x27AB, /* EB	BACK-TILTED SHADOWED WHITE RIGHTWARDS ARROW */
0x27AC, /* EC	FRONT-TILTED SHADOWED WHITE RIGHTWARDS ARROW */
0x27AD, /* ED	HEAVY LOWER RIGHT-SHADOWED WHITE RIGHTWARDS ARROW */
0x27AE, /* EE	HEAVY UPPER RIGHT-SHADOWED WHITE RIGHTWARDS ARROW */
0x27AF, /* EF	NOTCHED LOWER RIGHT-SHADOWED WHITE RIGHTWARDS ARROW */
-1,
0x27B1, /* F1	NOTCHED UPPER RIGHT-SHADOWED WHITE RIGHTWARDS ARROW */
0x27B2, /* F2	CIRCLED HEAVY WHITE RIGHTWARDS ARROW */
0x27B3, /* F3	WHITE-FEATHERED RIGHTWARDS ARROW */
0x27B4, /* F4	BLACK-FEATHERED SOUTH EAST ARROW */
0x27B5, /* F5	BLACK-FEATHERED RIGHTWARDS ARROW */
0x27B6, /* F6	BLACK-FEATHERED NORTH EAST ARROW */
0x27B7, /* F7	HEAVY BLACK-FEATHERED SOUTH EAST ARROW */
0x27B8, /* F8	HEAVY BLACK-FEATHERED RIGHTWARDS ARROW */
0x27B9, /* F9	HEAVY BLACK-FEATHERED NORTH EAST ARROW */
0x27BA, /* FA	TEARDROP-BARBED RIGHTWARDS ARROW */
0x27BB, /* FB	HEAVY TEARDROP-SHANKED RIGHTWARDS ARROW */
0x27BC, /* FC	WEDGE-TAILED RIGHTWARDS ARROW */
0x27BD, /* FD	HEAVY WEDGE-TAILED RIGHTWARDS ARROW */
0x27BE, /* FE	OPEN-OUTLINED RIGHTWARDS ARROW */
-1
};

static int combining[114][3] = {
{ 0x0300, /* COMBINING GRAVE ACCENT */				1, 0x0316 },
{ 0x0301, /* COMBINING ACUTE ACCENT */				1, 0x0317 },
{ 0x0302, /* COMBINING CIRCUMFLEX ACCENT */			1, 0x032D },
{ 0x0303, /* COMBINING TILDE */					1, 0x0330 },
{ 0x0304, /* COMBINING MACRON */				1, 0x0331 },
{ 0x0305, /* COMBINING OVERLINE */				1, 0x0332 },
{ 0x0306, /* COMBINING BREVE */					1, 0x032E },
{ 0x0307, /* COMBINING DOT ABOVE */				1, 0x0323 },
{ 0x0308, /* COMBINING DIAERESIS */				1, 0x0324 },
{ 0x0309, /* COMBINING HOOK ABOVE */				1, 0 },
{ 0x030A, /* COMBINING RING ABOVE */				1, 0x0325 },
{ 0x030B, /* COMBINING DOUBLE ACUTE ACCENT */			1, 0 },
{ 0x030C, /* COMBINING CARON */					1, 0x032C },
{ 0x030D, /* COMBINING VERTICAL LINE ABOVE */			1, 0x0329 },
{ 0x030E, /* COMBINING DOUBLE VERTICAL LINE ABOVE */		1, 0x0348 },
{ 0x030F, /* COMBINING DOUBLE GRAVE ACCENT */			1, 0 },
{ 0x0310, /* COMBINING CANDRABINDU */				1, 0 },
{ 0x0311, /* COMBINING INVERTED BREVE */			1, 0x032F },
{ 0x0312, /* COMBINING TURNED COMMA ABOVE */			1, 0 },
{ 0x0313, /* COMBINING COMMA ABOVE */				1, 0 },
{ 0x0314, /* COMBINING REVERSED COMMA ABOVE */			1, 0 },
{ 0x0315, /* COMBINING COMMA ABOVE RIGHT */			1, 0x0326 },
{ 0x0316, /* COMBINING GRAVE ACCENT BELOW */			-1, 0x0300 },
{ 0x0317, /* COMBINING ACUTE ACCENT BELOW */			-1, 0x0301 },
{ 0x0318, /* COMBINING LEFT TACK BELOW */			-1, 0 },
{ 0x0319, /* COMBINING RIGHT TACK BELOW */			-1, 0 },
{ 0x031A, /* COMBINING LEFT ANGLE ABOVE */			1, 0 },
{ 0x031B, /* COMBINING HORN */					1, 0 },
{ 0x031C, /* COMBINING LEFT HALF RING BELOW */			-1, 0 },
{ 0x031D, /* COMBINING UP TACK BELOW */				-1, 0 },
{ 0x031E, /* COMBINING DOWN TACK BELOW */			-1, 0 },
{ 0x031F, /* COMBINING PLUS SIGN BELOW */			-1, 0 },
{ 0x0320, /* COMBINING MINUS SIGN BELOW */			-1, 0 },
{ 0x0321, /* COMBINING PALATALIZED HOOK BELOW */		-1, 0 },
{ 0x0322, /* COMBINING RETROFLEX HOOK BELOW */			-1, 0 },
{ 0x0323, /* COMBINING DOT BELOW */				-1, 0x0307 },
{ 0x0324, /* COMBINING DIAERESIS BELOW */			-1, 0x0308 },
{ 0x0325, /* COMBINING RING BELOW */				-1, 0x030A },
{ 0x0326, /* COMBINING COMMA BELOW */				-1, 0x0315 },
{ 0x0327, /* COMBINING CEDILLA */				-1, 0 },
{ 0x0328, /* COMBINING OGONEK */				-1, 0 },
{ 0x0329, /* COMBINING VERTICAL LINE BELOW */			-1, 0x030D },
{ 0x032A, /* COMBINING BRIDGE BELOW */				-1, 0x0346 },
{ 0x032B, /* COMBINING INVERTED DOUBLE ARCH BELOW */		-1, 0 },
{ 0x032C, /* COMBINING CARON BELOW */				-1, 0x030C },
{ 0x032D, /* COMBINING CIRCUMFLEX ACCENT BELOW */		-1, 0x0302 },
{ 0x032E, /* COMBINING BREVE BELOW */				-1, 0x0306 },
{ 0x032F, /* COMBINING INVERTED BREVE BELOW */			-1, 0x0311 },
{ 0x0330, /* COMBINING TILDE BELOW */				-1, 0x0303 },
{ 0x0331, /* COMBINING MACRON BELOW */				-1, 0x0304 },
{ 0x0332, /* COMBINING LOW LINE */				-1, 0x0305 },
{ 0x0333, /* COMBINING DOUBLE LOW LINE */			-1, 0 },
{ 0x0334, /* COMBINING TILDE OVERLAY */				0, 0 },
{ 0x0335, /* COMBINING SHORT STROKE OVERLAY */			0, 0 },
{ 0x0336, /* COMBINING LONG STROKE OVERLAY */			0, 0 },
{ 0x0337, /* COMBINING SHORT SOLIDUS OVERLAY */			0, 0 },
{ 0x0338, /* COMBINING LONG SOLIDUS OVERLAY */			0, 0 },
{ 0x0339, /* COMBINING RIGHT HALF RING BELOW */			-1, 0 },
{ 0x033A, /* COMBINING INVERTED BRIDGE BELOW */			-1, 0 },
{ 0x033B, /* COMBINING SQUARE BELOW */				-1, 0 },
{ 0x033C, /* COMBINING SEAGULL BELOW */				-1, 0 },
{ 0x033D, /* COMBINING X ABOVE */				1, 0 },
{ 0x033E, /* COMBINING VERTICAL TILDE */			1, 0 },
{ 0x033F, /* COMBINING DOUBLE OVERLINE */			1, 0 },
{ 0x0340, /* COMBINING GRAVE TONE MARK */			1, 0 },
{ 0x0341, /* COMBINING ACUTE TONE MARK */			1, 0 },
{ 0x0342, /* COMBINING GREEK PERISPOMENI */			1, 0 },
{ 0x0343, /* COMBINING GREEK KORONIS */				1, 0 },
{ 0x0344, /* COMBINING GREEK DIALYTIKA TONOS */			1, 0 },
{ 0x0345, /* COMBINING GREEK YPOGEGRAMMENI */			1, 0 },
{ 0x0346, /* COMBINING BRIDGE ABOVE */				1, 0x032A },
{ 0x0347, /* COMBINING EQUALS SIGN BELOW */			-1, 0 },
{ 0x0348, /* COMBINING DOUBLE VERTICAL LINE BELOW */		-1, 0x030E },
{ 0x0349, /* COMBINING LEFT ANGLE BELOW */			-1, 0 },
{ 0x034A, /* COMBINING NOT TILDE ABOVE */			1, 0 },
{ 0x034B, /* COMBINING HOMOTHETIC ABOVE */			1, 0 },
{ 0x034C, /* COMBINING ALMOST EQUAL TO ABOVE */			1, 0 },
{ 0x034D, /* COMBINING LEFT RIGHT ARROW BELOW */		-1, 0 },
{ 0x034E, /* COMBINING UPWARDS ARROW BELOW */			-1, 0 },
{ 0x0360, /* COMBINING DOUBLE TILDE */				1, 0 },
{ 0x0361, /* COMBINING DOUBLE INVERTED BREVE */			1, 0 },
{ 0x0362, /* COMBINING DOUBLE RIGHTWARDS ARROW BELOW */		-1, 0 },
{ 0x0483, /* COMBINING CYRILLIC TITLO */			1, 0 },
{ 0x0484, /* COMBINING CYRILLIC PALATALIZATION */		1, 0 },
{ 0x0485, /* COMBINING CYRILLIC DASIA PNEUMATA */		1, 0 },
{ 0x0486, /* COMBINING CYRILLIC PSILI PNEUMATA */		1, 0 },
{ 0x0488, /* COMBINING CYRILLIC HUNDRED THOUSANDS SIGN */	1, 0 },
{ 0x0489, /* COMBINING CYRILLIC MILLIONS SIGN */		1, 0 },
{ 0x20D0, /* COMBINING LEFT HARPOON ABOVE */			1, 0 },
{ 0x20D1, /* COMBINING RIGHT HARPOON ABOVE */			1, 0 },
{ 0x20D2, /* COMBINING LONG VERTICAL LINE OVERLAY */		0, 0 },
{ 0x20D3, /* COMBINING SHORT VERTICAL LINE OVERLAY */		0, 0 },
{ 0x20D4, /* COMBINING ANTICLOCKWISE ARROW ABOVE */		1, 0 },
{ 0x20D5, /* COMBINING CLOCKWISE ARROW ABOVE */			1, 0 },
{ 0x20D6, /* COMBINING LEFT ARROW ABOVE */			1, 0 },
{ 0x20D7, /* COMBINING RIGHT ARROW ABOVE */			1, 0 },
{ 0x20D8, /* COMBINING RING OVERLAY */				0, 0 },
{ 0x20D9, /* COMBINING CLOCKWISE RING OVERLAY */		0, 0 },
{ 0x20DA, /* COMBINING ANTICLOCKWISE RING OVERLAY */		0, 0 },
{ 0x20DB, /* COMBINING THREE DOTS ABOVE */			1, 0 },
{ 0x20DC, /* COMBINING FOUR DOTS ABOVE */			1, 0 },
{ 0x20DD, /* COMBINING ENCLOSING CIRCLE */			0, 0 },
{ 0x20DE, /* COMBINING ENCLOSING SQUARE */			0, 0 },
{ 0x20DF, /* COMBINING ENCLOSING DIAMOND */			0, 0 },
{ 0x20E0, /* COMBINING ENCLOSING CIRCLE BACKSLASH */		0, 0 },
{ 0x20E1, /* COMBINING LEFT RIGHT ARROW ABOVE */		1, 0 },
{ 0x20E2, /* COMBINING ENCLOSING SCREEN */			0, 0 },
{ 0x20E3, /* COMBINING ENCLOSING KEYCAP */			0, 0 },
{ 0x3099, /* COMBINING KATAKANA-HIRAGANA VOICED SOUND MARK */	0, 0 },
{ 0x309A, /* COMBINING KATAKANA-HIRAGANA SEMI-VOICED SOUND MARK */	0, 0 },
{ 0xFE20, /* COMBINING LIGATURE LEFT HALF */			0, 0 },
{ 0xFE21, /* COMBINING LIGATURE RIGHT HALF */			0, 0 },
{ 0xFE22, /* COMBINING DOUBLE TILDE LEFT HALF */		0, 0 },
{ 0xFE23,  /* COMBINING DOUBLE TILDE RIGHT HALF */		0, 0 }
};

int is_combining(unsigned int u) {
	int i;
	for (i = 0; i < 114; i++) if (u == combining[i][0]) return TRUE;
	return FALSE;
}

static int lost_accent = 0;

int uni_out(int ch, unsigned char a, int diacritic, unsigned char da) {
    unsigned int wc;
    int count, ret, t;

    /* if it's out of range, do nothing */
    if (ch < 0 || ch > 255) return 0;

    if (is_ipa(a)) {
	t = -1;

	/* if it's a diacritic, do nothing */
	if (!ipa_width[ch]) {
		if (!lost_accent) return 0;
		diacritic = -1;
	}
	wc = ipa_uni[ch];
	/* if no known unicode translation, just space to fill column */
	if (!wc) {
		putchar(' ');
		return 1;
	}
	/* true diacritics have no width */
	if (diacritic > 255) diacritic = -1;
	if (diacritic >= 0 && ipa_width[diacritic]) diacritic = -1;
	if (a == IPASUPER || a == IPASUB) diacritic = -1;

	/* here if diacritic >= 0, look up combining form */
	if (diacritic >= 0 && diacritic <= 13) {
		int underflag;
		if (diacritic == 10 && da == IPAUNDER) diacritic = 14;

		if (diacritic == 3 && da == IPAUNDER) {
			if (ch == 'e') t = 0x1E1B; /* SMALL LETTER E WITH TILDE BELOW */
			else if (ch == 'i') t = 0x1E2D; /* SMALL LETTER I WITH TILDE BELOW */
			else if (ch == 'u') t = 0x1E75; /* SMALL LETTER U WITH TILDE BELOW */
		}
		else if (diacritic == 2 && da == IPAUNDER) {
			/* 2 = CIRCUMFLEX */
			if (ch == 'd') t = 0x1E13; /* SMALL LETTER D WITH CIRCUMFLEX BELOW */
			else if (ch == 'e') t = 0x1E19; /* SMALL LETTER E WITH CIRCUMFLEX BELOW */
			else if (ch == 'l') t = 0x1E3D; /* SMALL LETTER L WITH CIRCUMFLEX BELOW */
			else if (ch == 'n') t = 0x1E4B; /* SMALL LETTER N WITH CIRCUMFLEX BELOW */
			else if (ch == 't') t = 0x1E71; /* SMALL LETTER T WITH CIRCUMFLEX BELOW */
			else if (ch == 'u') t = 0x1E77; /* SMALL LETTER U WITH CIRCUMFLEX BELOW */
		}
		else if (diacritic == 6 && da == IPAUNDER) {
			if (ch == 'a')
				t = 0x1E01; /* SMALL LETTER A WITH RING BELOW */
		}
		else if (ch >= 'a' && ch <= 'z') {
			t = accents[diacritic][ch - 'a'];
			underflag = accents[diacritic][26];
			if (underflag && da == IPAOVER) t = -1;
			if (!underflag && da == IPAUNDER) t = -1;
		}
		else if (ch == '@' && da != IPAUNDER) {
		    if (diacritic == 1)
			t = 0x01FD; /* SMALL LETTER AE WITH ACUTE */
		    else if (diacritic == 9)
			t = 0x01E3; /* SMALL LETTER AE WITH MACRON */
		}
		if (t != -1) wc = (unsigned int)t;
	}
	/* ... handle combined forms for other diacritics */
	else if (ch == 'e' && diacritic == 152 && da != IPAUNDER) t = 0x1e17;
	else if (ch == 'e' && diacritic == 153 && da != IPAUNDER) t = 0x1e15;
	else if (ch == 'o' && diacritic == 152 && da != IPAUNDER) t = 0x1e53;
	else if (ch == 'o' && diacritic == 153 && da != IPAUNDER) t = 0x1e51;

	if (lost_accent && a == IPAUNDER) {
	    if (ch == 10) t = 0x0323; /* COMBINING DOT BELOW */
	    else if (ch == 4) t = 0x0324; /* COMBINING DIAERESIS BELOW */
	    else if (ch == 6) t = 0x0325; /* COMBINING RING BELOW */
	    else if (ch == 7) t = 0x032C; /* COMBINING CARON BELOW */
	    else if (ch == 2) t = 0x032D; /* COMBINING CIRCUMFLEX ACCENT BELOW */
	    else if (ch == 8) t = 0x032E; /* COMBINING BREVE BELOW */
	    else if (ch == 3) t = 0x0330; /* COMBINING TILDE BELOW */
	    else if (ch == 9) t = 0x0331; /* COMBINING MACRON BELOW */
	}

	/* if no combined form was found, try searching revtab */
	if (diacritic >= 0 && t == -1) {
		int i;
		for (i = 0; ; i++) {
			if (revtab[i][0] == -1) break;
			if (!is_ipa(revtab[i][1])) continue;
			if (revtab[i][2] == ch && revtab[i][3] == diacritic)
				break; 
		}
		if (revtab[i][0] != -1 && revtab[i][1] == da) {
			t = revtab[i][0];
		}
	}
	if (lost_accent) {
		int i;
		for (i = 0; i < 114; i++) if (wc == combining[i][0]) break;
		if (i < 114) {
/* printf("[%d,%d]", a, combining[i][1]); */
			if (a == IPAUNDER && combining[i][1] != -1) {
				if ( !(wc = combining[i][2]) ) return 0;
			}
			if (a == IPAOVER && combining[i][1] != 1) {
				if ( !(wc = combining[i][2]) ) return 0;
			}
		}
	}

	lost_accent = 0;
	/* there is a diacritic, but no combined form was found ... */
	if (diacritic >= 0 && t == -1) lost_accent = 1;

/*
0x02B0 "^h"	superscript h
0x02B1 "^H"	superscript hook h
0x02B2 "^j"	superscript j
0x02B3 "^r"	superscript r
0x02B4 "^R"	superscript turned r
0x02B6 "^K"	superscript small cap turned r
0x02B7 "^w"	superscript w
0x02B8 "^y"	superscript y
0x02E0 "^G"	superscript gamma
0x02E1 "^l"	superscript l
0x02E2 "^s"	superscript s
0x02E3 "^x"	superscript x
0x02E4 "^Q"	superscript reverse glottal stop
*/
	/* if no combined form was found, try superscript */
	if (t == -1) {
		if (a == IPAOVER || a == IPASUPER) {
		    if (ch == 'h') t = 0x02B0; /* MODIFIER LETTER SMALL H */
		    else if (ch == 72) t = 0x02B1; /* MODIFIER LETTER SMALL H WITH HOOK */
		    else if (ch == 'j') t = 0x02B2; /* MODIFIER LETTER SMALL J */
		    else if (ch == 'r') t = 0x02B3; /* MODIFIER LETTER SMALL R */
		    else if (ch == 244) t = 0x02B4; /* MODIFIER LETTER SMALL TURNED R */
		    else if (ch == 'K') t = 0x02B6; /* MODIFIER LETTER SMALL CAPITAL INVERTED R */
		    else if (ch == 'w') t = 0x02B7; /* MODIFIER LETTER SMALL W */
		    else if (ch == 'y') t = 0x02B8; /* MODIFIER LETTER SMALL Y */
		    else if (ch == 71 || ch == 200) /* 71 is gamma, 200 is baby gamma */
			t = 0x02E0; /* MODIFIER LETTER SMALL GAMMA */
		    else if (ch == 'l') t = 0x02E1; /* MODIFIER LETTER SMALL L */
		    else if (ch == 's') t = 0x02E2; /* MODIFIER LETTER SMALL S */
		    else if (ch == 'x') t = 0x02E3; /* MODIFIER LETTER SMALL X */
		    else if (ch == 81) t = 0x02E4; /* MODIFIER LETTER SMALL REVERSED GLOTTAL STOP */
		}
	}

	if (t != -1) wc = (unsigned int)t;
	if (diacritic >= 0 && t == -1) lost_accent = 1;

    }
    else if (a == MATH) {
	t = uni_symbol[ch];
	if (t == -1) {
		putchar(' ');
		return 1;
	}
	wc = (unsigned int)t;
    }
    else if (a == ZD) {
	t = uni_zd[ch];
	if (t == -1) {
		putchar(' ');
		return 1;
	}
	wc = (unsigned int)t;
    }
    else if (a == SUPER) {
      switch (ch) {
	case '0': wc = 0x2070; break;
	case '1': wc = 0x00b9; break;
	case '2': wc = 0x00b2; break;
	case '3': wc = 0x00b3; break;
	case '4': wc = 0x2074; break;
	case '5': wc = 0x2075; break;
	case '6': wc = 0x2076; break;
	case '7': wc = 0x2077; break;
	case '8': wc = 0x2078; break;
	case '9': wc = 0x2079; break;
	case '+': wc = 0x207a; break;
	case '-': wc = 0x207b; break;
	case '=': wc = 0x207c; break;
	case '(': wc = 0x207d; break;
	case ')': wc = 0x207e; break;
	case 'n': wc = 0x207f; break;
	default: wc = (unsigned int)ch; break;
      }
    }
    else if (a == SUB) {
      switch (ch) {
	case '0': wc = 0x2080; break;
	case '1': wc = 0x2081; break;
	case '2': wc = 0x2082; break;
	case '3': wc = 0x2083; break;
	case '4': wc = 0x2084; break;
	case '5': wc = 0x2085; break;
	case '6': wc = 0x2086; break;
	case '7': wc = 0x2087; break;
	case '8': wc = 0x2088; break;
	case '9': wc = 0x2089; break;
	case '+': wc = 0x208a; break;
	case '-': wc = 0x208b; break;
	case '=': wc = 0x208c; break;
	case '(': wc = 0x208d; break;
	case ')': wc = 0x208e; break;
	default: wc = (unsigned int)ch; break;
      }
    }
    else if (a > 128 || a == BOX) {
      switch (ch) {
	case 'q': wc = 0x2500; break;
	case 'Q': wc = 0x2501; break;
	case 'x': wc = 0x2502; break;
	case 'X': wc = 0x2503; break;
	case 'l': wc = 0x250C; break;
	case 'L': wc = 0x250F; break;
	case 'k': wc = 0x2510; break;
	case 'K': wc = 0x2513; break;
	case 'm': wc = 0x2514; break;
	case 'M': wc = 0x2517; break;
	case 'j': wc = 0x2518; break;
	case 'J': wc = 0x251B; break;
	case 't': wc = 0x251C; break;
	case 'T': wc = 0x2523; break;
	case 'u': wc = 0x2524; break;
	case 'U': wc = 0x252B; break;
	case 'w': wc = 0x252C; break;
	case 'W': wc = 0x2533; break;
	case 'v': wc = 0x2534; break;
	case 'V': wc = 0x253B; break;
	case 'n': wc = 0x253C; break;
	case 'N': wc = 0x254B; break;
	case 'h': wc = 0x2591; break;
	case 'H': wc = 0x2592; break;
	case 'd': wc = 0x2504; break;
	case 'D': wc = 0x2505; break;
	case 'e': wc = 0x2506; break;
	case 'E': wc = 0x2507; break;
	default: putchar(' '); return 1;
      }
    }
    else if (!a) {
	    wc = (unsigned int)((diacritic << 8) | ch);
    }
    else wc = (unsigned int) ch;


    if (wc < 0x80) {
/*printf("[wc%d,a%d,ch%d,diac%d]", wc,a,ch,diacritic);*/
        /* this should not be possible */
	if (wc < ' ') wc += 64;
	putchar(wc);
	return 1;
    }

    if (is_combining(wc)) ret = 0;
    else ret = 1;

    if (wc < 0x800) {
      count = 1;
      putchar( (unsigned char) ((wc >> 6) | 0xC0) );
    } else if (wc < 0x10000) {
      count = 2;
      putchar( (unsigned char) ((wc >> 12) | 0xE0) );
    } else if (wc < 0x200000) {
      count = 3;
      putchar( (unsigned char) ((wc >> 18) | 0xF0) );
    } else if (wc < 0x4000000) {
      count = 4;
      putchar( (unsigned char) ((wc >> 24) | 0xF8) );
    } else if (wc < 0x80000000) {
      count = 5;
      putchar( (unsigned char) ((wc >> 30) | 0xFC) );
    } else {
      return 0;
    }

    do {
      putchar ( (unsigned char)(((wc >> (6 * --count)) & 0x3F) | 0x80) );
    } while (count > 0);
    return ret;
}
